/*******************************************************************
  uLan Communication - low level and link driver

  ul_hdep.h	- host system dependant driver routines

  (C) Copyright 1996,1997 by Pavel Pisa 

  The uLan driver is distributed under the Gnu General Public Licence. 
  See file COPYING for details.

  Author reserves right to use and publish sources for embedded 
  applications under different conditions too.
 *******************************************************************/

#ifndef _LINUX_UL_HDEP_H
#define _LINUX_UL_HDEP_H

#include "ul_drv.h"

#define  ULAN_MAJOR 248

#define  ENABLE_UL_MEM_CHECK

#define  ENABLE_UL_IRQ_STALE_WDG

#ifdef KERNEL
 #ifndef ENABLE_UL_MEM_CHECK
  #define MALLOC(x) kmalloc(x,GFP_KERNEL)
  #define FREE(x)   kfree(x)
 #else
  #define MALLOC ul_mem_check_malloc
  #define FREE   ul_mem_check_free
  void * ul_mem_check_malloc(size_t size);
  void ul_mem_check_free(void *ptr);
 #endif
 #define UL_CHK_IOS(port,cnt)      check_region(port, cnt)
 #define UL_REQ_IOS(port,cnt,name) request_region(port, cnt, name)
 #define UL_REL_IOS(port,cnt)      release_region(port, cnt)
 #ifndef FOR_LINUX_KERNEL
  #define FOR_LINUX_KERNEL
 #endif
 #define ULD_HZ HZ
#elif defined(_WIN32)
 #ifndef ENABLE_UL_MEM_CHECK
  #define MALLOC(x) ExAllocatePool(NonPagedPool,x)
  #define FREE(x)   ExFreePool(x)
 #else
  #define MALLOC ul_mem_check_malloc
  #define FREE   ul_mem_check_free
  void * ul_mem_check_malloc(size_t size);
  void ul_mem_check_free(void *ptr);
 #endif
 #define UL_CHK_IOS(port,cnt)      (0)
 #define UL_REQ_IOS(port,cnt,name) (0)
 #define UL_REL_IOS(port,cnt)      (0)
#elif defined(__DJGPP__)
 #include <malloc.h>
 #include <dpmi.h>
 #ifndef ENABLE_UL_MEM_CHECK
  #define MALLOC(_size) ({void *_p=malloc(_size);\
                        _go32_dpmi_lock_data(_p,_size);\
                        _p;})
  #define FREE   free
 #else
  #define MALLOC ul_mem_check_malloc
  #define FREE   ul_mem_check_free
  void * ul_mem_check_malloc(size_t size);
  void ul_mem_check_free(void *ptr);
 #endif
 #define UL_CHK_IOS(port,cnt)      ({0;})
 #define UL_REQ_IOS(port,cnt,name) ({0;})
 #define UL_REL_IOS(port,cnt)      ({0;})
 #define UL_DRV_IN_LIB
 #define error perror
 #define ULD_HZ UCLOCKS_PER_SEC
 long int uld_jiffies;
 #define jiffies uld_jiffies
 #define JIFFIES_FROM_GETTIMEOFDAY
#else
 #include <malloc.h>
 #include <sys/time.h>
 #define MALLOC malloc
 #define FREE   free
 #define UL_CHK_IOS(port,cnt) \
	({int i=ioperm((long)port, (long)cnt, 1);\
 	  ioperm((long)port, (long)cnt, 0);\
 	  i<0?i:0;\
	})
 #define UL_REQ_IOS(port,cnt,name) ioperm((long)port, (long)cnt, 1)
 #define UL_REL_IOS(port,cnt)      ioperm((long)port, (long)cnt, 0)
 #define error perror
 #define ULD_HZ 1000000
 #define jiffies ({struct timeval tv;gettimeofday(&tv,NULL);\
                  tv.tv_usec+tv.tv_sec*1000000;})
#endif


#ifndef INLINE
  #define INLINE extern inline
#endif

#ifndef uchar
 #define uchar unsigned char
#endif

#if defined(__DJGPP__)
  /* Taken from Linux kernel - Copyright 1992, Linus Torvalds. */
  #define __sti() __asm__ __volatile__ ("sti": : :"memory")
  #define __cli() __asm__ __volatile__ ("cli": : :"memory")
  #define __save_flags(x) \
  __asm__ __volatile__("pushfl ; popl %0":"=g" (x): /* no input */ :"memory")
  #define __restore_flags(x) \
  __asm__ __volatile__("pushl %0 ; popfl": /* no output */ :"g" (x):"memory")
  INLINE void set_bit(int nr, volatile int * addr)
    { __asm__ __volatile__("btsl %1,%0":"=m" (*addr):"Ir" (nr));}
  INLINE void clear_bit(int nr, volatile int * addr)
    { __asm__ __volatile__("btrl %1,%0":"=m" (*addr):"Ir" (nr));}
  INLINE int test_and_set_bit(int nr, volatile int * addr)
    {int oldbit;__asm__ __volatile__("btsl %2,%1\n\tsbbl %0,%0"
      :"=r" (oldbit),"=m" (*addr):"Ir" (nr));return oldbit;}
  INLINE int test_bit(int nr, volatile int * addr)
    {int oldbit;__asm__ __volatile__("btl %2,%1\n\tsbbl %0,%0"
      :"=r" (oldbit):"m" (*addr),"Ir" (nr));return oldbit;}
  #define udelay usleep
  INLINE void schedule(void) {;}
  int request_irq(int irq,void *handler,int flags,char *name,void *dev_id);
  int free_irq(int irq,void *ctx);
  INLINE void synchronize_irq() {asm volatile ("":::"memory");}
#elif defined(USER_SPACE_EMU_IRQ)
  INLINE void schedule(void) {;}
  #define request_irq request_emu_irq
  #define free_irq release_emu_irq
  INLINE void synchronize_irq() {asm volatile ("":::"memory");}
#endif /* USER_SPACE_EMU_IRQ */

/*******************************************************************/
/* locking, irq and memory management */

/* disables another IRQ on same CPU when uLan IRQ is handled */
#ifdef FOR_LINUX_KERNEL
 #if 1
  #define UL_SA_IRQ_FLAGS (SA_SHIRQ|SA_INTERRUPT)
 #else
  #define UL_SA_IRQ_FLAGS SA_SHIRQ
 #endif
#else
 #define UL_SA_IRQ_FLAGS 0
#endif

#ifdef USER_SPACE_EMU_IRQ
 #define UL_IRQ_LOCK_FINI
 #define ul_synchronize_irq(irqnum) synchronize_irq()
 int INLINE irq_lock_cnt(int i)
 {static int lock_counter=0;
  return lock_counter+=i;
 };
 #define UL_IRQ_LOCK   {if(irq_lock_cnt(1)==1) emu_irq_cli();}
 #define UL_IRQ_UNLOCK {if(irq_lock_cnt(-1)==0)emu_irq_sti();}
 #define UL_MB()       {asm volatile ("":::"memory");}
#elif defined(FOR_LINUX_KERNEL)
 #define ul_synchronize_irq kc_synchronize_irq
 #if 1
  spinlock_t ul_irq_lock_spinlock;
  #define UL_IRQ_LOCK_GINI \
    spinlock_t ul_irq_lock_spinlock=SPIN_LOCK_UNLOCKED;
  #define UL_IRQ_LOCK_FINI unsigned long ul_irq_lock_flags;
  #define UL_IRQ_LOCK \
    {spin_lock_irqsave(&ul_irq_lock_spinlock,ul_irq_lock_flags);}
  #define UL_IRQ_UNLOCK \
    {spin_unlock_irqrestore(&ul_irq_lock_spinlock,ul_irq_lock_flags);}
 #else
  #define UL_IRQ_LOCK_FINI
  #define UL_IRQ_LOCK   {cli();}
  #define UL_IRQ_UNLOCK {sti();}
 #endif
 #define UL_MB()       mb()
#elif defined(_WIN32)
 typedef LONG atomic_t;
 KIRQL  uL_SpinLock_Irql;
 LONG   uL_SpinLock_Flag;
 #define UL_IRQ_LOCK_GINI \
     KIRQL  uL_SpinLock_Irql=0; \
     LONG   uL_SpinLock_Flag=0;
 #define UL_IRQ_LOCK_FINI \
	 KIRQL uL_IRQ_OldIrql;
 #define UL_IRQ_LOCK \
    {KeRaiseIrql (uL_SpinLock_Irql,&uL_IRQ_OldIrql); \
     while(InterlockedExchange(&uL_SpinLock_Flag,1)); \
    }
 #define UL_IRQ_UNLOCK \
    {InterlockedExchange(&uL_SpinLock_Flag,0); \
     KeLowerIrql(uL_IRQ_OldIrql); \
    }
 #define UL_MB()       {/* Flush cache here */;}
#elif defined(__DJGPP__)
  typedef long atomic_t;
  #define UL_IRQ_LOCK_FINI unsigned long ul_irq_lock_flags;
  #define UL_IRQ_LOCK \
    {__save_flags(ul_irq_lock_flags);__cli();}
  #define UL_IRQ_UNLOCK \
    {__restore_flags(ul_irq_lock_flags);}
  #define UL_MB()       {asm volatile ("":::"memory");}
  #define ul_synchronize_irq(irqnum) synchronize_irq()
#endif

#ifndef ATOMIC_INIT
  #define ATOMIC_INIT(i) (i)
#endif
#ifndef atomic_read
  #define atomic_read(v) (*(v))
#endif
#ifndef atomic_set
  #define atomic_set(v,i) (*(v)=(i))
#endif

#define UL_DRV_LOCK_FINI UL_IRQ_LOCK_FINI
#define UL_DRV_LOCK {\
  UL_IRQ_LOCK;\
 }
#define UL_DRV_UNLOCK {\
  UL_IRQ_UNLOCK;\
 }

/* enable to globaly lock CPU through uLan events handling, bad
   use only for special test purposes */
#if 0
  spinlock_t ul_global_irq_lock_spinlock;
  #define UL_GLOBAL_IRQ_LOCK_GINI \
    spinlock_t ul_global_irq_lock_spinlock=SPIN_LOCK_UNLOCKED;
  #define UL_GLOBAL_IRQ_LOCK_FINI unsigned long ul_global_irq_lock_flags;
  #define UL_GLOBAL_IRQ_LOCK \
    {spin_lock_irqsave(&ul_global_irq_lock_spinlock,ul_global_irq_lock_flags);}
  #define UL_GLOBAL_IRQ_UNLOCK \
    {spin_unlock_irqrestore(&ul_global_irq_lock_spinlock,ul_global_irq_lock_flags);}
#endif


/* get one free block or NULL */
INLINE ul_mem_blk *ul_alloc_blk(ul_drv *udrv)
{
  UL_DRV_LOCK_FINI
  ul_mem_blk *p;
  UL_DRV_LOCK;
  if((p=udrv->free_blk)!=NULL)
  {
    udrv->free_blk_cnt--;
    udrv->free_blk=p->next;
    p->next=NULL;
  };
  UL_DRV_UNLOCK;
  return p;
};

/* free one block */
INLINE void ul_free_blk(ul_drv *udrv, ul_mem_blk *blk)
{
  UL_DRV_LOCK_FINI
  UL_DRV_LOCK;
  blk->next=udrv->free_blk;
  udrv->free_blk=blk;
  udrv->free_blk_cnt++;
  UL_DRV_UNLOCK;
};

/* bottom half definitions */
#ifdef FOR_LINUX_KERNEL
  void ulan_do_bh(kc_tasklet_data_type data);
  typedef kc_tasklet_data_type ulan_do_bh_data_type;
  #define SCHEDULE_BH(udrv) \
        kc_tasklet_schedule(&(udrv)->bottom)
  #define INIT_UDRV_BH(udrv) \
	kc_tasklet_init(&(udrv)->bottom,ulan_do_bh,(long)udrv)
  #define KILL_UDRV_BH(udrv) \
	kc_tasklet_kill(&(udrv)->bottom)
  #define DECLARE_UDRV_BH struct kc_tasklet_struct bottom;
  #define uld_test_dfl(udrv,_DFL_flag) \
    test_bit(UL_DFLB_##_DFL_flag,&(udrv)->flags)
  #define uld_set_dfl(udrv,_DFL_flag) \
    set_bit(UL_DFLB_##_DFL_flag,&(udrv)->flags)
  #define uld_clear_dfl(udrv,_DFL_flag) \
    clear_bit(UL_DFLB_##_DFL_flag,&(udrv)->flags)
  #define uld_test_and_set_dfl(udrv,_DFL_flag) \
    test_and_set_bit(UL_DFLB_##_DFL_flag,&(udrv)->flags)
  #define uld_atomic_test_dfl(udrv,_DFL_flag) \
    test_bit(UL_DFLB_##_DFL_flag,&(udrv)->flags)
  #define uld_atomic_clear_dfl(udrv,_DFL_flag) \
    clear_bit(UL_DFLB_##_DFL_flag,&(udrv)->flags)
  #define uld_atomic_set_dfl(udrv,_DFL_flag) \
    set_bit(UL_DFLB_##_DFL_flag,&(udrv)->flags)

#elif defined(_WIN32)
			/* must be used only at IRQL >= DISPATCH_LEVEL */
  #define SCHEDULE_BH(udrv) \
    do{ \
	(udrv)->flag_ASK_BOTTOM=1; \
	KeInsertQueueDpc(&udrv->bottom_dpc,NULL,NULL); \
    }while(0)
  #define INIT_UDRV_BH(udrv) /* not used search Dpc */
  #define KILL_UDRV_BH(udrv)
  #define DECLARE_UDRV_BH	/* not used search for Dpc */
  PIRP xch_pending_irp(PIRP *irp_ptr, PIRP new_irp); /* for user app waitting */
  void wake_and_xch_pending_irp(PIRP *irp_ptr, PIRP new_irp, NTSTATUS status);
  #define uld_test_dfl(udrv,_DFL_flag) \
    ((udrv)->flag_##_DFL_flag)
  #define uld_set_dfl(udrv,_DFL_flag) \
    ((udrv)->flag_##_DFL_flag=1)
  #define uld_clear_dfl(udrv,_DFL_flag) \
    ((udrv)->flag_##_DFL_flag=0)
  #define uld_test_and_set_dfl(udrv,_DFL_flag) \
    (InterlockedExchange(&(udrv)->flag_##_DFL_flag,1))
  #define uld_atomic_test_dfl(udrv,_DFL_flag) \
    ((udrv)->flag_##_DFL_flag)
  #define uld_atomic_clear_dfl(udrv,_DFL_flag) \
    (InterlockedExchange(&(udrv)->flag_##_DFL_flag,0))
  #define uld_atomic_set_dfl(udrv,_DFL_flag) \
    (InterlockedExchange(&(udrv)->flag_##_DFL_flag,1))
    
#else
  typedef void * ulan_do_bh_data_type;
  #define SCHEDULE_BH(udrv) \
    do{set_bit(UL_DFLB_ASK_BOTTOM,&(udrv->flags));}while(0)
  #define INIT_UDRV_BH(udrv)
  #define KILL_UDRV_BH(udrv)
  #define DECLARE_UDRV_BH
  #define uld_test_dfl(udrv,_DFL_flag) \
    test_bit(UL_DFLB_##_DFL_flag,&(udrv)->flags)
  #define uld_set_dfl(udrv,_DFL_flag) \
    set_bit(UL_DFLB_##_DFL_flag,&(udrv)->flags)
  #define uld_clear_dfl(udrv,_DFL_flag) \
    clear_bit(UL_DFLB_##_DFL_flag,&(udrv)->flags)
  #define uld_test_and_set_dfl(udrv,_DFL_flag) \
    test_and_set_bit(UL_DFLB_##_DFL_flag,&(udrv)->flags)
  #define uld_atomic_test_dfl(udrv,_DFL_flag) \
    test_bit(UL_DFLB_##_DFL_flag,&(udrv)->flags)
  #define uld_atomic_clear_dfl(udrv,_DFL_flag) \
    clear_bit(UL_DFLB_##_DFL_flag,&(udrv)->flags)
  #define uld_atomic_set_dfl(udrv,_DFL_flag) \
    set_bit(UL_DFLB_##_DFL_flag,&(udrv)->flags)
#endif

/* link watchdog timer definitions */
#ifdef FOR_LINUX_KERNEL
  void ulan_do_wd_timer(unsigned long data);
  #define SCHEDULE_UDRV_WDTIM(udrv,expires) \
       mod_timer(&(udrv)->wd_timer,expires);
  #define INIT_UDRV_WDTIM(udrv) \
    do{ \
       init_timer(&(udrv)->wd_timer); \
       (udrv)->wd_timer.function=ulan_do_wd_timer; \
       (udrv)->wd_timer.data=(unsigned long)udrv; \
    }while(0)
  #define DONE_UDRV_WDTIM(udrv) \
       del_timer(&(udrv)->wd_timer)
  #define DECLARE_UDRV_WDTIM struct timer_list wd_timer
#elif defined(_WIN32)
  #define SCHEDULE_UDRV_WDTIM(udrv,expires)
  #define INIT_UDRV_WDTIM(udrv)
  #define DONE_UDRV_WDTIM(udrv)
  #define DECLARE_UDRV_WDTIM
#else
  #define SCHEDULE_UDRV_WDTIM(udrv,expires) \
    (udrv->wd_timer_expires=expires,\
     set_bit(UL_DFLB_WDSCHED,&(udrv->flags)))
  #define INIT_UDRV_WDTIM(udrv) \
    (clear_bit(UL_DFLB_WDSCHED,&(udrv->flags)))
  #define DONE_UDRV_WDTIM(udrv) \
    (clear_bit(UL_DFLB_WDSCHED,&(udrv->flags)))
  #define DECLARE_UDRV_WDTIM
#endif

/* compatible generic io-ports access */
#ifdef _WIN32
 #define ul_inb(port)	 READ_PORT_UCHAR((PVOID)(port))
 #define ul_outb(port,val) WRITE_PORT_UCHAR((PVOID)(port),val)
#elif defined(__DJGPP__)
 #include <pc.h>
 #define ul_inb(port)	 inportb(port)
 #define ul_outb(port,val) outportb(port,val)
#else
 #define ul_inb(port)	 inb(port)
 #define ul_outb(port,val) outb(val,port)
#endif

/*******************************************************************/
/* ulan file access structure */

#define ULOP_MAGIC 0x754c2233

struct	ul_opchain;

/* uLan driver operator/user structure pointed by file private/extension */
typedef struct ul_opdata {
	int	magic;		/* magic number */
	ul_drv	*udrv;		/* uLan driver pointer */
	struct ul_opdata *opnext, *opprew; /* per udrv clients bll */
  #ifdef FOR_LINUX_KERNEL
	struct file *file;	/* back ptr to file */
  #elif defined(_WIN32)
	PFILE_OBJECT  file;
    PIRP   wait_irp;		/* IRP waitting for nonempty recchain */ 
  #endif /* _WIN32 */
	ul_mem_blk *message;	/* first block of prepared message */ 
	ul_data_it data;	/* ul_mem_blk data iterator */
  #ifdef FOR_LINUX_KERNEL
	wait_queue_head_t wqrec; /* wait queue for received packet */
  #endif
	struct ul_opchain *recchain; /* received messages */
	struct ul_opchain *filtchain; /* filter for incoming messages */
} ul_opdata;

/* list of filtered mesage types also used for per operator receive queue */
typedef struct	ul_opchain {
	struct	ul_opchain *next; /* Cyklic BLL of received mesages */
	struct	ul_opchain *prev; /* or input filters */
	ul_mem_blk *message;	/* first block of waiting message */ 
	unsigned stamp;		/* unigue message number */
	uchar    cmd;		/* command or service */
	uchar    dadr;		/* destination address */
	uchar    sadr;		/* source address */
	uchar    state;		/* state of filtering */
} ul_opchain;

#define UL_OPST_FREE	0	/* free ul_opchain block */
#define UL_OPST_ONCE	1	/* wait for message only once */
#define UL_OPST_FILT	2	/* filter incomming messages */
#define UL_OPST_MES	3	/* message on recchain list */
#define UL_OPST_FILTNEW	4	/* new filter check for repeat */

typedef int ul_chip_init_fnc(ul_drv *udrv, int port, int irq, int baud, int options);

typedef struct ul_chip_type_ent {
	char *chip_name;
	ul_chip_init_fnc *chip_init;
	int chip_flags;
	int chip_chans;
	int chip_options;
}ul_chip_type_ent;

#define UL_CHIPT_RQPORT	1	/* port number requested */
#define UL_CHIPT_RQIRQ	2	/* irq umber requested */
#define UL_CHIPT_PCI	4	/* pci type */
#define UL_CHIPT_NOPROBE 8	/* do not probe if type undefined */


/*******************************************************************/
/* user space visible structures */

typedef struct ul_msginfo {
	int	dadr;		/* destignation address */
	int	sadr;		/* source address */
	int	cmd;		/* command/socket number */
	int	flg;		/* message flags */
	int	len;		/* length of frame */
	unsigned stamp;		/* unigue message number */
} ul_msginfo;

/*******************************************************************/
/* ioctl definitions */

#ifdef _WIN32

/* 32768-65535 are reserved for use by customers.*/
#define FILE_DEVICE_ULAN  0x0000A000

/* function codes 2048-4095 are reserved for customers. */
#define UL_IOCTL_INDEX  0xA00

#define UL_IO(function) \
	CTL_CODE(FILE_DEVICE_ULAN,UL_IOCTL_INDEX+function,METHOD_BUFFERED,FILE_ANY_ACCESS)
#define UL_IOR(function,size) \
	CTL_CODE(FILE_DEVICE_ULAN,UL_IOCTL_INDEX+function,METHOD_BUFFERED,FILE_ANY_ACCESS)
#define UL_IOW(function,size) \
	CTL_CODE(FILE_DEVICE_ULAN,UL_IOCTL_INDEX+function,METHOD_BUFFERED,FILE_ANY_ACCESS)

#else /* _WIN32 */

#define	UL_IOCTL	0x75

#define UL_IO(function)        _IO(UL_IOCTL,function)
#define UL_IOR(function,size)  _IOR(UL_IOCTL,function,size)
#define UL_IOW(function,size)  _IOW(UL_IOCTL,function,size)

#endif /* _WIN32 */

#define UL_DRV_VER	UL_IO(0x10)
#define UL_NEWMSG	UL_IOW(0x11,ul_msginfo)
#define UL_TAILMSG	UL_IOW(0x12,ul_msginfo)
#define UL_FREEMSG	UL_IO(0x13)
#define UL_ACCEPTMSG	UL_IOR(0x14,ul_msginfo)
#define UL_ACTAILMSG	UL_IOR(0x15,ul_msginfo)
#define UL_ADDFILT	UL_IOW(0x16,ul_msginfo)
#define UL_ABORTMSG	UL_IO(0x17)
#define UL_REWMSG	UL_IO(0x18)
#define UL_INEPOLL	UL_IO(0x19)
#define UL_WAITREC	UL_IO(0x1A)
#define UL_KLOGBLL	UL_IO(0x41)
#define UL_STROKE	UL_IO(0x42)
#define UL_DEBFLG	UL_IO(0x43)  /* arg = new debug mask */
#define UL_HWTEST	UL_IO(0x44)  /* arg = test subcommand */


/*******************************************************************/
/* define unix error codes */

#ifdef _WIN32

#define ENOMEM (-STATUS_NO_MEMORY)
#define EPERM  (-STATUS_PRIVILEGE_NOT_HELD)
#define ENOMSG (-STATUS_END_OF_FILE)
#define EINVAL (-STATUS_INVALID_PARAMETER)
#define ENODEV (-STATUS_INVALID_PARAMETER)

#endif /* _WIN32 */

#ifdef __DJGPP__

#define ENOMSG ENOENT

#endif /* _WIN32 */

/*******************************************************************/
/* debug support */

/* initial value for uld_debug_flg */
#ifndef ULD_DEBUG_DEFAULT 
  #ifndef UL_DRV_IN_LIB
    #define ULD_DEBUG_DEFAULT (0x01|0x10|0x40)
  #else
    #define ULD_DEBUG_DEFAULT (0x01|0x40)
  #endif
#endif /* ULD_DEBUG_DEFAULT */

#ifdef FOR_LINUX_KERNEL
 #define UL_PRINTF printk
#elif defined(_WIN32)
 #define UL_PRINTF DbgPrint
#elif defined(UL_DRV_IN_LIB)
 int uld_printk(const char *format, ...) __attribute__((format(printf,1,2)));
 void uld_printk_flush(void);
 #define UL_PRINTF uld_printk
#else
 #define UL_PRINTF printf
#endif

#if 1
 #define LOG_FATAL	if(uld_debug_flg&0x01) UL_PRINTF
 #define LOG_CHIO	if(uld_debug_flg&0x02) UL_PRINTF
 #define LOG_IRQ	if(uld_debug_flg&0x04) UL_PRINTF
 #define LOG_MESSAGES	if(uld_debug_flg&0x08) UL_PRINTF
 #define LOG_FAILS	if(uld_debug_flg&0x10) UL_PRINTF
 #define LOG_SEQNUM	if(uld_debug_flg&0x20) UL_PRINTF
 #define LOG_PORTS	if(uld_debug_flg&0x40) UL_PRINTF
 #define LOG_FILEIO	if(uld_debug_flg&0x80) UL_PRINTF
 #define LOG_FILTS	if(uld_debug_flg&0x100) UL_PRINTF
 #define LOG_MSK_DISABLE  0x10000   /* for use with UL_DRV_IN_LIB */
#else
 #define LOG_CHIO(x,args...)
 #define LOG_FATAL(x,args...)
 #define LOG_IRQ(x,args...)
 #define LOG_MESSAGES(x,args...)
 #define LOG_FAILS(x,args...)
 #define LOG_SEQNUM(x,args...)
 #define LOG_PORTS(x,args...)
 #define LOG_FILEIO(x,args...)
 #define LOG_FILTS(x,args...)
#endif

#ifndef KERN_CRIT
  #define KERN_CRIT
#endif
#ifndef KERN_ERR
  #define KERN_ERR
#endif
#ifndef KERN_WARNING
  #define KERN_WARNING
#endif
#ifndef KERN_INFO
  #define KERN_INFO
#endif

#if (defined(CPU)&&(CPU>=586)||\
     defined(CONFIG_X86_TSC))&&!defined(_WIN32)
 #define RDTSC \
   ({unsigned long l,h;\
    asm volatile ("rdtsc" : "=a" (l), "=d" (h));\
    l;\
   })
#else
 #define RDTSC 0
#endif

#endif /* _LINUX_UL_HDEP_H */
