/*******************************************************************
  uLan Communication - low level and link driver

  ul_drv.c	- test driver implementation

  (C) Copyright 1996,1997 by Pavel Pisa 

  The uLan driver is distributed under the Gnu General Public Licence. 
  See file COPYING for details.

  Author reserves right to use and publish sources for embedded 
  applications under different conditions too.
 *******************************************************************/

#define UL_WITH_UARTS
#ifdef FOR_USER_SPACE 
  #include <stdio.h>
  #include <unistd.h>
  #include <time.h>
  #include <errno.h>
  #include <asm/io.h>
  #include <asm/bitops.h>
  #include <asm/atomic.h>
  #include <sys/perm.h>
  #include <string.h>
  #include "emu_irq.h"
#elif defined(_WIN32)
  #ifndef FOR_WIN_WDM
    #define FOR_WIN_KMD
    #include "ntddk.h"
  #else 
    #define DBG 1
    #define ULD_DEBUG_DEFAULT (0x01|0x08|0x10|0x40|0x80)
    #define UL_WITH_PCI
    #include "wdm.h"
    #include "ul_wdinc.h"
  #endif
#elif defined(__DJGPP__)
  int uld_djgpp_data_beg;
  void uld_djgpp_code_beg(void) {;};
  #include <unistd.h>
  #include <errno.h>
  #include <stdio.h>
  #include <time.h>
  #include <string.h>
#else
  #include <linux/config.h>
  #include <linux/module.h>
  #include <linux/version.h>
  #include <linux/errno.h>
  #include <linux/kernel.h>
  #include <linux/sched.h>
  #include <linux/ioport.h>
  #include <linux/sys.h>
  #include <linux/fs.h>
  #include <linux/interrupt.h>
  #include <linux/poll.h>
  #include <linux/delay.h>
  #include <asm/bitops.h>
  #include <asm/segment.h>
  #include <asm/system.h>
  #include <asm/io.h>
  #include <asm/irq.h>

  #ifdef __KERNEL__
    #ifndef KERNEL
      #define KERNEL
    #endif
  #endif

  #include "k_compat.h"

  #if (LINUX_VERSION_CODE < VERSION(2,5,50))
    #include <linux/modversions.h>
  #endif /* 2.5.50 */

  #if (LINUX_VERSION_CODE < VERSION(2,5,0))
    #include <linux/malloc.h>
  #endif /* 2.5.0 */

  #if (LINUX_VERSION_CODE < VERSION(2,3,20))
    #include <asm/spinlock.h>
  #else
    #include <linux/spinlock.h>
  #endif

  #if (LINUX_VERSION_CODE >= VERSION(2,4,0))
    #define UL_WITH_DEVFS    
    #define UL_WITH_PCI
    #include <linux/miscdevice.h>
    #include <linux/init.h>
  #endif 

  #if (LINUX_VERSION_CODE < VERSION(2,5,40))
    #include <linux/tqueue.h>
  #else /* 2.5.40 */
    #include <linux/devfs_fs_kernel.h>
  #endif /* 2.5.40 */

  #ifdef UL_WITH_PCI
    #include <linux/pci.h>
  #endif

  #ifdef UL_WITH_USB
    #include <linux/usb.h>
    #include <linux/smp_lock.h>
  #endif

#endif

#include "ul_drv.h"
#include "ul_hdep.h"

#ifndef UL_DRV_IN_LIB
unsigned uld_debug_flg=ULD_DEBUG_DEFAULT;
#else
unsigned uld_debug_flg; /* Left application  set defaults */
#endif
int ulbuffer=0x8000;	/* default communication buffer size */

#ifdef UL_IRQ_LOCK_GINI
 UL_IRQ_LOCK_GINI	/* Inicialize global uLan spinlock */
#endif

#ifdef UL_GLOBAL_IRQ_LOCK_GINI
 UL_GLOBAL_IRQ_LOCK_GINI /* Initialize global IRQ lock,
 			    very bad if needed */
#endif

#ifdef ENABLE_UL_MEM_CHECK
atomic_t ul_mem_check_counter=ATOMIC_INIT(0);
#endif /* ENABLE_UL_MEM_CHECK */

int uld_null_fnc(ul_drv *udrv, int ret_code)
{
  return 0;
}

#ifdef UL_WITH_PCI
  #include "ul_16950pci.c"
#endif

#ifdef UL_WITH_UARTS
/*******************************************************************/
/* Chip driver for 82510 */

#if 1
 /* this enables use of shadow 82510 bank switch */
 /* be !!! carefull !!!, use U510_SBANK_FINI at begin of every */
 /* function containing u510_outb, u510_inb */
 /* such function must not call another such function */
 #define _U510_SBANK_SHADOW_VAR sbank_sv
 #define U510_SBANK_FINI int sbank_sv=-1;
#else
 #define U510_SBANK_FINI
#endif

#include "ul_82510.h"

/*** Test interrupt request state ***/
int u510_pool(ul_drv *udrv)
{
 return u510_pool_gir(udrv->port);
};

/*** Receive character into char_buff ***/
int u510_recch(ul_drv *udrv, int ret_code)
{
  U510_SBANK_FINI
  uchar uc;
  udrv->char_buff=0;
  if(u510_inb(udrv->port,U510_MCR)&U510_MCR_OEQ)
  { /* direction switch necessary */
    if(!(u510_inb(udrv->port,U510_GSR)&U510_GER_TM))
    { /* wait to end of previous transfer */
     u510_outb(udrv->port,U510_GER, U510_GER_TM);
     return UL_RC_WIRQ;
    }
    u510_outb(udrv->port,U510_MCR, U510_MCR_IE);
    u510_outb(udrv->port,U510_RCM, U510_RCM_FLUSH);
  };

  u510_outb(udrv->port,U510_GER,U510_GER_RI);
  uc=u510_inb(udrv->port,U510_RST);
  if(!(uc&0x01)) return UL_RC_WIRQ;

  uc=u510_inb(udrv->port,U510_RXF);
  udrv->char_buff=u510_inb(udrv->port,U510_RXD);
  if(uc&0x08)
  { /* store last ctrl for connect line busy/ready */
    udrv->char_buff|=0x100; /* controll char */
    udrv->last_ctrl=udrv->char_buff;
  } else udrv->last_ctrl&=0x7F;
  udrv->xor_sum^=udrv->char_buff;udrv->xor_sum++;

  UL_FRET;
  if(uc&0x02) 
  {
    LOG_CHIO(" ER:%03X",udrv->char_buff);
    return UL_RC_EFRAME; /* frame error */
  };
  LOG_CHIO(" R:%03X",udrv->char_buff);
  return UL_RC_PROC;
};

/*-- Helper functions for sndchar --*/

int u510_sndch(ul_drv *udrv, int ret_code);

int u510_sndch_w2(ul_drv *udrv, int ret_code)
{
  U510_SBANK_FINI
  u510_outb(udrv->port,U510_MCR, U510_MCR_OE);
  UL_FNEXT(u510_sndch);
  return UL_RC_PROC;
}

int u510_sndch_w1(ul_drv *udrv, int ret_code)
{
  U510_SBANK_FINI
  unsigned u;
  u=udrv->baud_div*11*16;
  u510_outb(udrv->port,U510_TMCR,0);
  u510_inb(udrv->port,U510_TMST);
  u510_outb(udrv->port,U510_GER,U510_GER_TIE);
  u510_outb(udrv->port,U510_LCR,U510_LCR_C|U510_LCR_DLAB);
  u510_outb(udrv->port,U510_BBL,(uchar)u);
  u510_outb(udrv->port,U510_BBH,(uchar)(u>>8));
  u510_outb(udrv->port,U510_LCR,U510_LCR_C);
  u510_outb(udrv->port,U510_TMIE,2);
  u510_outb(udrv->port,U510_TMCR,0x22);
  UL_FNEXT(u510_sndch_w2);
  return UL_RC_WIRQ;
}

/*** Send character from char_buff ***/
int u510_sndch(ul_drv *udrv, int ret_code)
{
  U510_SBANK_FINI
  if(!(u510_inb(udrv->port,U510_MCR)&U510_MCR_OEQ))
  { /* direction switch necessary */
    UL_FNEXT(u510_sndch_w1);
    if(!(u510_inb(udrv->port,U510_GSR)&U510_GER_TM))
    { /* wait to end of previous transfer */
      u510_outb(udrv->port,U510_GER, U510_GER_TM);
      return UL_RC_WIRQ;
    }
    return UL_RC_PROC;
  };

  u510_outb(udrv->port,U510_GER,U510_GER_TI);
  if(!(u510_inb(udrv->port,U510_LSR)&U510_LSR_TRE)) return UL_RC_WIRQ;

  if(udrv->char_buff&0x100)
  { /* store last ctrl for connect line busy/ready */
    udrv->last_ctrl=udrv->char_buff;
    u510_outb(udrv->port,U510_TXF, 0x80); /* controll char */
  } else u510_outb(udrv->port,U510_TXF, 0x00); /* regular char */;
  u510_outb(udrv->port,U510_TXD,(uchar)udrv->char_buff); /* char code */
  udrv->xor_sum^=udrv->char_buff;udrv->xor_sum++;

  LOG_CHIO(" T:%03X",udrv->char_buff);
  UL_FRET;
  return UL_RC_PROC;
};

/*-- Helper functions for wait --*/

int u510_wait(ul_drv *udrv, int ret_code);

int u510_wait_1(ul_drv *udrv, int ret_code)
{
  U510_SBANK_FINI
  if(u510_inb(udrv->port,U510_RST)&1)
  {
    UL_FNEXT(u510_recch);
    return UL_RC_PROC;
  };

  if(!udrv->wait_time--)
  {
    LOG_CHIO(" Timeout!");
    UL_FRET;
    return UL_RC_ETIMEOUT;
  };

  u510_outb(udrv->port,U510_TMCR,0);
  u510_inb(udrv->port,U510_TMST);

  u510_outb(udrv->port,U510_LCR,U510_LCR_C|U510_LCR_DLAB);
  u510_outb(udrv->port,U510_BBL,0xFF);
  u510_outb(udrv->port,U510_BBH,0xFF);
  u510_outb(udrv->port,U510_LCR,U510_LCR_C);
  u510_outb(udrv->port,U510_TMIE,2);
  u510_outb(udrv->port,U510_TMCR,0x22);

  return UL_RC_WIRQ;
};

/*** Wait for time or received character ***/
int u510_wait(ul_drv *udrv, int ret_code)
{
  U510_SBANK_FINI
  unsigned u;

  udrv->char_buff=0;
  if(!(u510_inb(udrv->port,U510_GSR)&U510_GER_TM))
  { /* wait to end of previous transfer */
    u510_outb(udrv->port,U510_GER, U510_GER_TM);
    return UL_RC_WIRQ;
  };
  if(u510_inb(udrv->port,U510_MCR)&U510_MCR_OEQ)
  { /* direction switch necessary */
    u510_outb(udrv->port,U510_MCR, U510_MCR_IE);
    u510_outb(udrv->port,U510_RCM, U510_RCM_FLUSH);
  };

  UL_FNEXT(u510_wait_1);

  u=udrv->wait_time*11*16*udrv->baud_div;
  udrv->wait_time=u>>16;
  u510_outb(udrv->port,U510_TMCR,0);
  u510_inb(udrv->port,U510_TMST);
  u510_outb(udrv->port,U510_GER,U510_GER_TIE+U510_GER_RI);
  if((u&0xFFFF)<=2) return UL_RC_PROC;

  u510_outb(udrv->port,U510_LCR,U510_LCR_C|U510_LCR_DLAB);
  u510_outb(udrv->port,U510_BBL,(uchar)u);
  u510_outb(udrv->port,U510_BBH,(uchar)(u>>8));
  u510_outb(udrv->port,U510_LCR,U510_LCR_C);
  u510_outb(udrv->port,U510_TMIE,2);
  u510_outb(udrv->port,U510_TMCR,0x22);

  return UL_RC_WIRQ;
};

/*-- Helper functions for connect --*/

int u510_connect_1(ul_drv *udrv, int ret_code);

int u510_connect_2(ul_drv *udrv, int ret_code)
{
  U510_SBANK_FINI
  u510_inb(udrv->port,U510_GSR);
  u510_outb(udrv->port,U510_LCR,U510_LCR_C);
  u510_outb(udrv->port,U510_MCR,U510_MCR_IE);
  u510_outb(udrv->port,U510_RCM, U510_RCM_FLUSH);
  u510_inb(udrv->port,U510_RST);
  if(udrv->chip_temp==1)
  {LOG_CHIO(" Connected");
    UL_FRET;
  }else{ 
    udrv->wait_time=(udrv->chip_temp&3)+1; 
    udrv->chip_temp>>=2;
    UL_FCALL2(*udrv->fnc_wait,u510_connect_1);
  };
  return UL_RC_PROC;
};

int u510_connect_1(ul_drv *udrv, int ret_code)
{
  U510_SBANK_FINI
  if(ret_code!=UL_RC_ETIMEOUT)
    { UL_FRET; return UL_RC_EARBIT;};
  if((u510_inb(udrv->port,U510_MSR)&U510_MSR_RxD)||
     (u510_inb(udrv->port,U510_RST)&1))
  { UL_FRET;
    LOG_CHIO(" EARBIT!");
    return UL_RC_EARBIT;
  };
  u510_outb(udrv->port,U510_LCR,U510_LCR_C|U510_LCR_B);
  u510_outb(udrv->port,U510_MCR,U510_MCR_OE);
  UL_FNEXT(u510_connect_2);
  /* wait to end of transfer */
  u510_inb(udrv->port,U510_GSR);
  u510_outb(udrv->port,U510_GER, U510_GER_TM);
  u510_outb(udrv->port,U510_TXF, 0x00);
  u510_outb(udrv->port,U510_TXD, 0x00); 
  return UL_RC_WIRQ;
};

/*** Connect to RS485 bus ***/
int u510_connect(ul_drv *udrv, int ret_code)
{
  unsigned u;
  u=udrv->last_ctrl;
  udrv->wait_time=((udrv->my_adr-u-1)&0xF)+4; 
  if(((u&0x180)!=0x180)||
     (u==0x1FF)) udrv->wait_time+=0x10;
  udrv->last_ctrl=0;
  udrv->char_buff=0;
  udrv->chip_temp=(udrv->my_adr&0x3F)|0x40;
  UL_FCALL2(*udrv->fnc_wait,u510_connect_1);
  return UL_RC_PROC;
};

/*
 UL_FNEXT(u510_recch);
 UL_FCALL(u510_recch);
 UL_FCALL2(u510_recch,u510_recch);
 UL_FRET;
*/

static char *u510_port_name="ulan_u510";

/*** 82510 initialize ports ***/
int u510_pinit(ul_drv *udrv)
{
  U510_SBANK_FINI
  unsigned u;
  int baud=udrv->baud_val;
  if (UL_CHK_IOS(udrv->port,8)) 
  { LOG_FATAL(KERN_CRIT "uLan u510_pinit : cannot reguest ports !\n");
    return UL_RC_EPORT;
  };
  UL_REQ_IOS(udrv->port,8,u510_port_name);
  for(u=0;u<0x80;u+=0x20) 
  {
    ul_outb(udrv->port+U510_GIR,u);
    if((ul_inb(udrv->port+U510_GIR)^u)&0x60)
    {
      ul_outb(udrv->port+U510_GIR,0);
      UL_REL_IOS(udrv->port,8);
      return UL_RC_EPORT;
    };
  };
  if (!baud) baud=19200;
  u=(0x8CA00+baud/2)/baud;
  if (u>0xFFFF) u=0xFFFF;
  udrv->baud_div=u;
  u510_outb(udrv->port,U510_ICM, 0x10);
  u510_outb(udrv->port,U510_GER, 0);
  u510_outb(udrv->port,U510_LCR, U510_LCR_C);
  u510_outb(udrv->port,U510_IMD, 0xB);
  u510_outb(udrv->port,U510_TMD, 0);	/* 0x18 -> RTS */
  u510_outb(udrv->port,U510_FMD, 0);
  u510_outb(udrv->port,U510_RMD, 0x40);	/* Addresss opens RxFIFO */
  u510_outb(udrv->port,U510_CLCF,0x50);	/* RxC a TxC are 16x from BRGA */
  u510_outb(udrv->port,U510_BACF,0x04);	/* Baud gen system clock */
  u510_outb(udrv->port,U510_BBCF,0x00);	/* Timer system clock */
  u510_outb(udrv->port,U510_PMD, 0x7C);	/* Configure pins */

  u510_outb(udrv->port,U510_LCR, U510_LCR_C|U510_LCR_DLAB);
  u510_outb(udrv->port,U510_BAL, (uchar)udrv->baud_div);
  u510_outb(udrv->port,U510_BAH, (uchar)(udrv->baud_div>>8));
  u510_outb(udrv->port,U510_LCR, U510_LCR_C);

  u510_outb(udrv->port,U510_LCR, U510_LCR_C|U510_LCR_C);
  u510_outb(udrv->port,U510_LCR, U510_LCR_C);
  u510_outb(udrv->port,U510_GER, U510_GER_MI);
  return UL_RC_PROC;
};

/*** 82510 deinitialize ports ***/
int u510_pdone(ul_drv *udrv)
{
  U510_SBANK_FINI
  u510_outb(udrv->port,U510_GER, 0);	/* disable interrupts */
  u510_outb(udrv->port,U510_MCR, U510_MCR_IE);  /* transmitter off */
  u510_outb(udrv->port,U510_GIR, U510_GIR_NO_INT); /* flush irq */
  UL_REL_IOS(udrv->port,8);
  return UL_RC_PROC;
};

/*** 82510 generate irq for irq_probe */
int u510_genirq(ul_drv *udrv,int param)
{
  U510_SBANK_FINI
  if(param) 
  {
    u510_outb(udrv->port,U510_MCR, U510_MCR_IE);  /* transmitter off */
    u510_outb(udrv->port,U510_GER, U510_GER_TI);  /* enable interrupts */
    u510_outb(udrv->port,U510_TXD, 0xFF);	  /* trig TxFIFO int */
  }else{
    u510_outb(udrv->port,U510_GER, 0);	/* disable interrupts */
    u510_outb(udrv->port,U510_MCR, U510_MCR_IE);  /* transmitter off */
    u510_outb(udrv->port,U510_GIR, U510_GIR_NO_INT); /* flush irq */
  };
  return UL_RC_PROC;
};

/* support for hardware tests */
int u510_hwtest(ul_drv *udrv,int param)
{
  unsigned u;
  U510_SBANK_FINI
  switch(param)
  {
   case 0x10:
   case 0x11:
	u510_outb(udrv->port,U510_GER, 0); /* disable interrupts */
	u510_outb(udrv->port,U510_MCR, U510_MCR_OE); /* transmitter on */
	u510_outb(udrv->port,U510_LCR, param&1?
	              U510_LCR_C:U510_LCR_B); /* set TD lines */
   case 0x12:
        u=u510_inb(udrv->port,U510_MSR);
	return u510_inb(udrv->port,U510_LSR)|(u<<8)|
	                 (u&U510_MSR_TxD?0:0x100000)|
			 (u&U510_MSR_RxD?0:0x10000);
   case 0x13:
	u510_outb(udrv->port,U510_MCR, U510_MCR_IE); /* transmitter off */
        u=u510_inb(udrv->port,U510_MSR);
	return u510_inb(udrv->port,U510_LSR)|(u<<8)|
	                 (u&U510_MSR_TxD?0:0x100000)|
			 (u&U510_MSR_RxD?0:0x10000);
  }
  return UL_RC_ENOFNC;
};

/*** Control functions of chip driver  ***/
int u510_cctrl(ul_drv *udrv, int ctrl_fnc, int param)
{
  U510_SBANK_FINI
  switch (ctrl_fnc)
  {
   case UL_CC_DIN:	/* switch off line transmitter */
	{ u510_outb(udrv->port,U510_MCR, U510_MCR_IE); 
	  return UL_RC_PROC;
	};
   case UL_CC_DOUT:	/* switch on line transmitter */
	{ u510_outb(udrv->port,U510_MCR, U510_MCR_OE);
	  return UL_RC_PROC;
	};
   case UL_CC_PINIT:	/* initialize ports (parameter baud)*/
  	{ return u510_pinit(udrv);
  	};
   case UL_CC_PDONE:	/* deinitialize ports */
  	{ return u510_pdone(udrv);
  	};
   case UL_CC_GENIRQ:	/* generate irq for irq_probe */
  	{return u510_genirq(udrv,param);
  	};
   case UL_CC_HWTEST:	/* support for hardware testing */
  	{return u510_hwtest(udrv,param);
  	};
  };
  return UL_RC_ENOFNC;
};

/*** 82510 chip driver initialize ***/
int u510_init(ul_drv *udrv, int port, int irq, int baud, int options)
{
  unsigned u;
  if (UL_CHK_IOS(port,8)) 
  { LOG_FATAL(KERN_CRIT "uLan u510_init : cannot reguest ports !\n");
    return UL_RC_EPORT;
  };
  UL_REQ_IOS(port,8,u510_port_name);
  for(u=0;u<0x80;u+=0x20) 
  {
    ul_outb(port+U510_GIR,u);
    if((ul_inb(port+U510_GIR)^u)&0x60)
    {
      ul_outb(port+U510_GIR,0);
      UL_REL_IOS(port,8);
      return UL_RC_EPORT;
    };
  };
  udrv->chip_options=options;
  if(!baud) baud=19200;
  udrv->baud_val=baud;
  u=(0x8CA00+baud/2)/baud;
  if (u>0xFFFF) u=0xFFFF;
  udrv->baud_div=u;
  udrv->port=port;
  udrv->irq=irq; 
  udrv->fnc_recch=&u510_recch;
  udrv->fnc_sndch=&u510_sndch;
  udrv->fnc_wait=&u510_wait;
  udrv->fnc_pool=&u510_pool;
  udrv->fnc_connect=&u510_connect;
  udrv->fnc_cctrl=&u510_cctrl;
  UL_REL_IOS(port,8);
  return 0;
};

/*******************************************************************/
/* Chip driver for 16450 */

#include "serial_reg.h"

#define u450_inb(base,port)	 ul_inb(base+port)
#define u450_outb(base,port,val) ul_outb(base+port,val)

#define U450_MCR_IE	0x0D	/* receive */
#define U450_MCR_OE	0x0E	/* transmit */
#define U450_MCR_OEQ	0x02	/* transmit ? */

#define U450_MSR_RxD	UART_MSR_CTS /* RxD line readback */
#define U450_MSR_TxD	UART_MSR_DSR /* TxD line readback */
#define U450_MSR_DTxD	UART_MSR_DDSR /* TxD line changed */

#define U450_LCR_UL	0x3B	/* receive/transmit */
#define U450_LCR_ULC	0x2B	/* transmit control character */
#define U450_LCR_ULB	0x7B	/* transmit line break */

#define U450_LSR_C	0x04	/* controll character received */
#define U450_LSR_ERR	0x1A	/* error in receiver */

/*** Test interrupt request state ***/
int u450_pool(ul_drv *udrv)
{
  unsigned u;
  u=u450_inb(udrv->port,UART_IIR);
  return (u&UART_IIR_NO_INT)?0:((u&UART_IIR_ID)|1);
};

/*** Wait end of transmit ***/
int u450_weot(ul_drv *udrv, int ret_code)
{
  int ret;
  unsigned u;
  uchar lsr_val;
  uchar uc;
  if(!udrv->chip_buff[0]||
     !(u450_inb(udrv->port,UART_MCR)&U450_MCR_OEQ)) 
  {
    UL_FRET;
    u450_outb(udrv->port,UART_LCR,U450_LCR_UL);
    return UL_RC_PROC;
  };
  u450_outb(udrv->port,UART_IER,UART_IER_RDI);
  lsr_val=u450_inb(udrv->port,UART_LSR);
  ret=UL_RC_WIRQ;
  if(!(~lsr_val&(UART_LSR_THRE|UART_LSR_TEMT)))
  {
    LOG_CHIO(" TMI");
    ret=UL_RC_PROC; /* was commented out */
  }
  if(lsr_val&U450_LSR_ERR) 
    ret=UL_RC_EFRAME;	/* frame error */
  if(lsr_val&UART_LSR_DR)
  {
    if((u=udrv->chip_buff[1])) udrv->chip_buff[1]=0;
     else if((u=udrv->chip_buff[0])) udrv->chip_buff[0]=0;
    uc=u450_inb(udrv->port,UART_RX);
    if((u^uc)&0xff||!u)
    {
      LOG_CHIO(" TE%c!",u?'0':'1');
      ret=UL_RC_EBADCHR;
    }else{
      LOG_CHIO(".");
      if(ret>=0&&!udrv->chip_buff[0])
	ret=UL_RC_PROC;
    }; 
  };
  if(ret!=UL_RC_WIRQ)
  {
    u450_outb(udrv->port,UART_LCR,U450_LCR_UL);
    UL_FRET;
  };
  LOG_CHIO(" weot ret %d ",ret);
  return ret;
};

/*** Switch from receive to transmit ***/
  int u450_sw_R2T_1(ul_drv *udrv, int ret_code);
  int u450_sw_R2T_2(ul_drv *udrv, int ret_code);
  int u450_sw_R2T_3(ul_drv *udrv, int ret_code);

int u450_sw_R2T(ul_drv *udrv, int ret_code)
{
  udrv->chip_buff[0]=udrv->chip_buff[1]=0;
 /* A_wait */
  UL_FCALL2(u450_weot,u450_sw_R2T_1);
  return UL_RC_PROC;
};

int u450_sw_R2T_1(ul_drv *udrv, int ret_code)
{
  if(ret_code<0) {UL_FRET;return ret_code;};
  u450_outb(udrv->port,UART_MCR,U450_MCR_IE);
  u450_outb(udrv->port,UART_IER,UART_IER_THRI);
  u450_outb(udrv->port,UART_TX,0);
  UL_FNEXT(u450_sw_R2T_2);
  if(u450_inb(udrv->port,UART_LSR)&UART_LSR_THRE) return UL_RC_PROC;
    else return UL_RC_WIRQ;
}; 

int u450_sw_R2T_2(ul_drv *udrv, int ret_code)
{
 /* A_wait1: */
  u450_inb(udrv->port,UART_MSR);
  u450_outb(udrv->port,UART_IER,UART_IER_MSI);
  UL_FNEXT(u450_sw_R2T_3);
  return UL_RC_WIRQ;
}; 

int u450_sw_R2T_3(ul_drv *udrv, int ret_code)
{
  uchar uc;
 #if 0
  static int reported_bad_chip=0;
 #endif
 /* A_wait2: */
  uc=u450_inb(udrv->port,UART_MSR);
  if((uc&(U450_MSR_DTxD|U450_MSR_TxD))!=U450_MSR_DTxD)
    return UL_RC_WIRQ;
 #if 0
  while(~u450_inb(udrv->port,UART_LSR)&
	(UART_LSR_TEMT|UART_LSR_THRE))
   if(!reported_bad_chip++)
     LOG_FATAL(KERN_CRIT "uLan u450_sw_R2T_3 : problematic IC => looping in ISR\n");
 #endif
  UL_FRET;
  return UL_RC_PROC;
}; 

/*** Receive character into char_buff ***/
  int u450_recch_1(ul_drv *udrv, int ret_code);
  int u450_recch_2(ul_drv *udrv, int ret_code);

int u450_recch(ul_drv *udrv, int ret_code)
{
  udrv->char_buff=0;
  if(u450_inb(udrv->port,UART_MCR)&U450_MCR_OEQ)
  {
    UL_FCALL2(u450_weot,u450_recch_1);
    return UL_RC_PROC;
  };
  UL_FNEXT(u450_recch_2);
  return UL_RC_PROC;
};
 
int u450_recch_1(ul_drv *udrv, int ret_code)
{
  u450_outb(udrv->port,UART_MCR,U450_MCR_IE);
  if(ret_code<0) {UL_FRET;return ret_code;};
  UL_FNEXT(u450_recch_2);
  return UL_RC_PROC;
};

INLINE int u450_recch_sub(ul_drv *udrv,int lsr_val,int rec_char)
{ /* helper function to process received character */
  if(lsr_val&U450_LSR_C) udrv->last_ctrl=rec_char|=0x100; 
   else udrv->last_ctrl&=0x7F; 
  udrv->char_buff=rec_char;
  udrv->xor_sum^=rec_char;udrv->xor_sum++;
  if(lsr_val&U450_LSR_ERR)
  {
    LOG_CHIO(" ER:%03X",udrv->char_buff);
    return UL_RC_EFRAME; /* frame error */
  };
  LOG_CHIO(" R:%03X",udrv->char_buff);
  return UL_RC_PROC;
};
 
int u450_recch_2(ul_drv *udrv, int ret_code)
{
  uchar lsr_val;
  u450_outb(udrv->port,UART_IER,UART_IER_RDI);
  lsr_val=u450_inb(udrv->port,UART_LSR);
  if(!(lsr_val&UART_LSR_DR)) return UL_RC_WIRQ;
  UL_FRET;
  return u450_recch_sub(udrv,lsr_val,u450_inb(udrv->port,UART_RX));
};

/*** Send character from char_buff ***/
  int u450_sndch_1(ul_drv *udrv, int ret_code);
  int u450_sndch_2(ul_drv *udrv, int ret_code);

int u450_sndch(ul_drv *udrv, int ret_code)
{
  uchar uc;
  uc=(udrv->char_buff&0x100)?U450_LCR_ULC:U450_LCR_UL;
  if(u450_inb(udrv->port,UART_LCR)!=uc)
  { /* change between control and data character format */
   if((u450_inb(udrv->port,UART_MCR)&U450_MCR_OEQ))
   { /* wait to end of transmit */
     UL_FCALL2(u450_weot,u450_sndch_1);
     return UL_RC_PROC;
   }else{ /* direction switch necessary */
     udrv->chip_buff[0]=udrv->chip_buff[1]=0;
     UL_FCALL2(u450_sw_R2T_1,u450_sndch_1);
     UL_FCALL2(u450_weot,u450_sw_R2T_1);
     return UL_RC_PROC;
   };
  };
  if(!(u450_inb(udrv->port,UART_MCR)&U450_MCR_OEQ))
  { /* direction switch necessary */
    udrv->chip_buff[0]=udrv->chip_buff[1]=0;
    UL_FCALL2(u450_sw_R2T_1,u450_sndch_1);
    UL_FCALL2(u450_weot,u450_sw_R2T_1);
  } else UL_FNEXT(u450_sndch_1);
  return UL_RC_PROC;
};

int u450_sndch_1(ul_drv *udrv, int ret_code)
{
  uchar uc;
  if(ret_code<0) {UL_FRET;return ret_code;};
  u450_outb(udrv->port,UART_MCR,U450_MCR_OE);
  if(udrv->char_buff&0x100)
  { /* store last ctrl for connect line busy/ready */
    udrv->last_ctrl=udrv->char_buff;
    uc=U450_LCR_ULC;
  }else{ 
    uc=U450_LCR_UL;
  };
  udrv->chip_buff[1]=udrv->chip_buff[0];
  udrv->chip_buff[0]=udrv->char_buff|0x1000; 
  udrv->xor_sum^=udrv->char_buff;udrv->xor_sum++;
  u450_outb(udrv->port,UART_IER,UART_IER_THRI);
  u450_outb(udrv->port,UART_LCR,uc);  /* control/data character */
  u450_outb(udrv->port,UART_TX,(uchar)udrv->char_buff);
  LOG_CHIO(" T:%03X",udrv->char_buff);
  UL_FNEXT(u450_sndch_2);
  return UL_RC_WIRQ;
}; 

int u450_sndch_2(ul_drv *udrv, int ret_code)
{
  uchar uc;
  unsigned u;
  uc=u450_inb(udrv->port,UART_LSR);
  UL_FRET;
  if(uc&U450_LSR_ERR)
      return UL_RC_EFRAME;	/* frame error */
  if(!(uc&UART_LSR_DR))
      return UL_RC_PROC;		/* no char */ 
  if((u=udrv->chip_buff[1])) udrv->chip_buff[1]=0;
    else if((u=udrv->chip_buff[0])) udrv->chip_buff[0]=0;
      else return UL_RC_PROC;
  uc=u450_inb(udrv->port,UART_RX);
  if((u^uc)&0xff||!u)
  {
    LOG_CHIO(" TE%c!",u?'2':'3');
    return UL_RC_EBADCHR;
  };
  LOG_CHIO(".");
  return UL_RC_PROC;
};
 
/*** Wait for time or received character ***/
  int u450_wait_1(ul_drv *udrv, int ret_code);
  int u450_wait_2(ul_drv *udrv, int ret_code);

int u450_wait(ul_drv *udrv, int ret_code)
{
  udrv->char_buff=0;
  if(u450_inb(udrv->port,UART_MCR)&U450_MCR_OEQ)
  {
    UL_FCALL2(u450_weot,u450_wait_1);
  } else UL_FNEXT(u450_wait_1);
  return UL_RC_PROC;
};

int u450_wait_1(ul_drv *udrv, int ret_code)
{
 uchar lsr_val;
 if(ret_code<0) {UL_FRET;return ret_code;};
 u450_outb(udrv->port,UART_MCR,U450_MCR_IE);
 u450_outb(udrv->port,UART_IER,UART_IER_THRI|UART_IER_RDI);
 if((lsr_val=u450_inb(udrv->port,UART_LSR))&UART_LSR_DR)
 {
   UL_FRET;
   return u450_recch_sub(udrv,lsr_val,u450_inb(udrv->port,UART_RX));
 };
 if(!(lsr_val&UART_LSR_THRE)) return UL_RC_WIRQ;
 if(!udrv->wait_time)
 {
   UL_FRET;
   LOG_CHIO(" Timeout!");
   return UL_RC_ETIMEOUT;
 };
 UL_FNEXT(u450_wait_2);
 u450_outb(udrv->port,UART_TX,0);
 return UL_RC_WIRQ;
};

int u450_wait_2(ul_drv *udrv, int ret_code)
{
  uchar lsr_val;
  if((lsr_val=u450_inb(udrv->port,UART_LSR))&UART_LSR_DR)
  {
    UL_FRET;
    return u450_recch_sub(udrv,lsr_val,u450_inb(udrv->port,UART_RX));
  };
  if(!udrv->wait_time--)
  {
    UL_FRET;
    LOG_CHIO(" Timeout!");
    return UL_RC_ETIMEOUT;
  };
  u450_outb(udrv->port,UART_TX,0);
  return UL_RC_WIRQ;
};

/*** Connect to RS485 bus ***/
  int u450_connect_1(ul_drv *udrv, int ret_code);
  int u450_connect_2(ul_drv *udrv, int ret_code);

int u450_connect(ul_drv *udrv, int ret_code)
{
  unsigned u;
  LOG_CHIO(" C_0 ");
  u=udrv->last_ctrl;
  udrv->wait_time=((udrv->my_adr-u-1)&0xF)+4; 
  if(((u&0x180)!=0x180)||
     (u==0x1FF)) udrv->wait_time+=0x10;
  udrv->last_ctrl=0;
  udrv->char_buff=0;
  udrv->chip_temp=(udrv->my_adr&0x3F)|0x40;
  u450_inb(udrv->port,UART_MSR);
  UL_FCALL2(*udrv->fnc_wait,u450_connect_1);
  return UL_RC_PROC;
};

int u450_connect_1(ul_drv *udrv, int ret_code)
{
  LOG_CHIO(" C_1 ");
  if(ret_code!=UL_RC_ETIMEOUT)
  { UL_FRET;
    LOG_CHIO(" EARBIT1!");
    return UL_RC_EARBIT;
  };
  /* check for RxD low */
  if(u450_inb(udrv->port,UART_MSR)&U450_MSR_RxD)
  { UL_FRET;
    LOG_CHIO(" EARBIT2!");
    return UL_RC_EARBIT;
  };
  u450_outb(udrv->port,UART_LCR,U450_LCR_ULB);
  u450_inb(udrv->port,UART_LSR);
  u450_outb(udrv->port,UART_IER, UART_IER_RLSI);
  u450_outb(udrv->port,UART_MCR,U450_MCR_OE);
  UL_FNEXT(u450_connect_2);
  /* wait for break condition now */
  return UL_RC_WIRQ;
};

int u450_connect_2(ul_drv *udrv, int ret_code)
{
  LOG_CHIO(" C_2 ");
  if(!u450_inb(udrv->port,UART_LSR)&UART_LSR_BI)
    return UL_RC_WIRQ;
  u450_outb(udrv->port,UART_LCR,U450_LCR_ULC);
  u450_outb(udrv->port,UART_MCR,U450_MCR_IE);
  u450_inb(udrv->port,UART_LSR);
  u450_inb(udrv->port,UART_RX);
  if(udrv->chip_temp==1)
  {LOG_CHIO(" Connected");
    UL_FRET;
  }else{ 
    udrv->wait_time=(udrv->chip_temp&3)+1; 
    udrv->chip_temp>>=2;
    UL_FCALL2(*udrv->fnc_wait,u450_connect_1);
  };
  return UL_RC_PROC;
};

/*
 UL_FNEXT(u450_recch);
 UL_FCALL(u450_recch);
 UL_FCALL2(u450_recch,u450_recch);
 UL_FRET;
*/

static char *u450_port_name="ulan_u450";

/*** 16450 initialize ports ***/
int u450_pinit(ul_drv *udrv)
{
  unsigned u;
  int baud=udrv->baud_val;
  if (UL_CHK_IOS(udrv->port,8)) 
  { LOG_FATAL(KERN_CRIT "uLan u450_pinit : cannot reguest ports !\n");
    return UL_RC_EPORT;
  };
  UL_REQ_IOS(udrv->port,8,u450_port_name);
  if (!baud) baud=19200;
  u=(0x1C200+baud/2)/baud;
  if (u>0xFFFF) u=0xFFFF;
  udrv->baud_div=u;

  udrv->chip_buff[0]=udrv->chip_buff[1]=0;
  u450_outb(udrv->port,UART_MCR,UART_MCR_OUT1|UART_MCR_DTR);
  u450_outb(udrv->port,UART_LCR,UART_LCR_DLAB);
  u450_outb(udrv->port,(uchar)UART_DLL,(uchar)udrv->baud_div);
  u450_outb(udrv->port,(uchar)UART_DLM,(uchar)(udrv->baud_div>>8));
  u450_outb(udrv->port,UART_LCR,U450_LCR_UL);
  u450_outb(udrv->port,UART_IER,0);
  u450_outb(udrv->port,UART_FCR,0);	/* disable FIFOs */
  u450_inb(udrv->port,UART_LSR);		/* reset errors */
  u450_inb(udrv->port,UART_RX);		/* and other irq */
  u450_outb(udrv->port,UART_MCR,U450_MCR_IE); /* connect irq */
  return UL_RC_PROC;
};

/*** 16450 deinitialize ports ***/
int u450_pdone(ul_drv *udrv)
{
  u450_outb(udrv->port,UART_IER, 0);	/* disable interrupts */
  u450_outb(udrv->port,UART_MCR, U450_MCR_IE&3);  /* transmitter off */
  u450_inb(udrv->port,UART_IIR);		/* flush irq */
  u450_inb(udrv->port,UART_LSR);		/* flush irq */
  UL_REL_IOS(udrv->port,8);
  return UL_RC_PROC;
};

/*** 16450 generate irq for irq_probe */
int u450_genirq(ul_drv *udrv,int param)
{
  if(param) 
  {
    u450_outb(udrv->port,UART_MCR, U450_MCR_IE);  /* transmitter off */
    u450_outb(udrv->port,UART_IER, UART_IER_MSI);	/* enable interrupts */
    u450_outb(udrv->port,UART_LCR, U450_LCR_ULB);	/* trig break */
  }else{
    u450_outb(udrv->port,UART_IER, 0);		/* disable interrupts */
    u450_outb(udrv->port,UART_MCR, U450_MCR_IE);	/* transmitter off */
    u450_inb(udrv->port,UART_MSR);		/* flush irq */
  };
  return UL_RC_PROC;
};

/* support for hardware tests */
int u450_hwtest(ul_drv *udrv,int param)
{
  unsigned u;
  switch(param)
  {
   case 0x10:
   case 0x11:
	u450_outb(udrv->port,UART_IER, 0); /* disable interrupts */
	u450_outb(udrv->port,UART_MCR, U450_MCR_OE); /* transmitter on */
	u450_outb(udrv->port,UART_LCR, param&1?
	             U450_LCR_UL:U450_LCR_ULB); /* set TD lines */
   case 0x12:
        u=u450_inb(udrv->port,UART_MSR);
	return u450_inb(udrv->port,UART_LSR)|(u<<8)|
	                (u&(U450_MSR_TxD)?0:0x100000)|
			(u&(U450_MSR_RxD)?0:0x10000);
   case 0x13:
	u450_outb(udrv->port,UART_MCR, U450_MCR_IE); /* transmitter off */
        u=u450_inb(udrv->port,UART_MSR);
	return u450_inb(udrv->port,UART_LSR)|(u<<8)|
	                (u&(U450_MSR_TxD)?0:0x100000)|
			(u&(U450_MSR_RxD)?0:0x10000);
  }
  return UL_RC_ENOFNC;
};

/*** Control functions of chip driver  ***/
int u450_cctrl(ul_drv *udrv, int ctrl_fnc, int param)
{
  switch (ctrl_fnc)
  {
   case UL_CC_DIN:	/* switch off line transmitter */
	{ u450_outb(udrv->port,UART_MCR, U450_MCR_IE); 
	  u450_outb(udrv->port,UART_LCR, U450_LCR_UL); 
	  return UL_RC_PROC;
	};
   case UL_CC_DOUT:	/* switch on line transmitter */
	{ u450_outb(udrv->port,UART_MCR, U450_MCR_OE);
	  u450_outb(udrv->port,UART_LCR, U450_LCR_UL); 
	  return UL_RC_PROC;
	};
   case UL_CC_PINIT:	/* initialize ports (parameter baud)*/
  	{ return u450_pinit(udrv);
  	};
   case UL_CC_PDONE:	/* deinitialize ports */
  	{ return u450_pdone(udrv);
  	};
   case UL_CC_GENIRQ:	/* generate irq for irq_probe */
  	{return u450_genirq(udrv,param);
  	};
   case UL_CC_HWTEST:	/* support for hardware testing */
  	{return u450_hwtest(udrv,param);
  	};
  };
  return UL_RC_ENOFNC;
};

/*** 16450 chip driver initialize ***/
int u450_init(ul_drv *udrv, int port, int irq, int baud, int options)
{
  unsigned u;
  unsigned old_mcr;
  unsigned old_lcr;
  if (UL_CHK_IOS(port,8)) 
  { LOG_FATAL(KERN_CRIT "uLan u450_init : cannot reguest ports !\n");
    return UL_RC_EPORT;
  };
  UL_REQ_IOS(port,8,u450_port_name);
  LOG_PORTS("uLan u450_init : probe MCR\n");
  old_mcr=u450_inb(port,UART_MCR);
  LOG_PORTS("uLan u450_init : MCR = 0x%02X\n",old_mcr);

  u450_outb(port,UART_MCR,UART_MCR_OUT1|UART_MCR_DTR);
  #ifdef FOR_LINUX_KERNEL
   set_current_state(TASK_INTERRUPTIBLE);
   schedule_timeout(ULD_HZ/10);
  #elif defined(_WIN32)
   /* !!!!!!!!!!!!!! */
   /* KeStallExecutionProcessor(20000); */
   { LARGE_INTEGER delay;
     delay=RtlConvertLongToLargeInteger(-30*10000);
     KeDelayExecutionThread(KernelMode,FALSE,&delay);
   }
  #else
   sleep(1);
  #endif
  old_lcr=u450_inb(port,UART_LCR);
  u450_outb(port,UART_LCR,UART_LCR_SBC);
  #ifdef FOR_LINUX_KERNEL
   set_current_state(TASK_INTERRUPTIBLE);
   schedule_timeout(ULD_HZ/10);
  #elif defined(_WIN32)
   /* !!!!!!!!!!!!!! */
   /* KeStallExecutionProcessor(2000); */
   { LARGE_INTEGER delay;
     delay=RtlConvertLongToLargeInteger(-10*10000);
     KeDelayExecutionThread(KernelMode,FALSE,&delay);
   }
  #else
   sleep(1);
  #endif
  if(u450_inb(port,UART_MSR)&U450_MSR_TxD)
  { u450_outb(port,UART_LCR,U450_LCR_UL);
    LOG_PORTS("uLan u450_init : MCR test 1 OK\n");
    #ifdef FOR_LINUX_KERNEL
     set_current_state(TASK_INTERRUPTIBLE);
     schedule_timeout(ULD_HZ/10);
    #elif defined(_WIN32)
     /* !!!!!!!!!!!!!! */
     /* KeStallExecutionProcessor(2000); */
     { LARGE_INTEGER delay;
       delay=RtlConvertLongToLargeInteger(-10*10000);
       KeDelayExecutionThread(KernelMode,FALSE,&delay);
     }
    #else
     sleep(1);
    #endif
    if(!(u450_inb(port,UART_MSR)&U450_MSR_TxD))
    { LOG_PORTS("uLan u450_init : MCR test 2 OK\n");
      udrv->chip_options=options;
      if(!baud) baud=19200;
      udrv->baud_val=baud;
      u=(0x1C200+baud/2)/baud;
      if (u>0xFFFF) u=0xFFFF;
      udrv->baud_div=u;
      udrv->port=port;
      udrv->irq=irq; 
      udrv->fnc_recch=&u450_recch;
      udrv->fnc_sndch=&u450_sndch;
      udrv->fnc_wait=&u450_wait;
      udrv->fnc_pool=&u450_pool;
      udrv->fnc_connect=&u450_connect;
      udrv->fnc_cctrl=&u450_cctrl;
      UL_REL_IOS(port,8);
      return 0;
    };
    LOG_PORTS("uLan u450_init : MCR test 2 Error\n");
  };
  u450_outb(port,UART_LCR,(uchar)old_lcr);
  u450_outb(port,UART_MCR,(uchar)old_mcr);
  LOG_PORTS("uLan u450_init : MCR test Error\n");
  UL_REL_IOS(port,8);
  return UL_RC_EPORT;
};

#endif /*UL_WITH_UARTS*/

#if defined(UL_WITH_UARTS) || defined(UL_WITH_PCI)
/*******************************************************************/
/* Frame begin/end */

/*-- Helper functions for sndbeg --*/

int uld_sndbeg_2(ul_drv *udrv, int ret_code)
{
  if(ret_code<0) {UL_FRET;return ret_code;};
  udrv->char_buff=udrv->con_cmd;
  UL_FNEXT(*udrv->fnc_sndch);
  return UL_RC_PROC;
};

int uld_sndbeg_1(ul_drv *udrv, int ret_code)
{
  if(ret_code<0) {UL_FRET;return ret_code;};
  udrv->char_buff=udrv->char_buff=udrv->con_sadr;
  UL_FCALL2(*udrv->fnc_sndch,uld_sndbeg_2);
  return UL_RC_PROC;
};

/*** Send frame begin ***/
int uld_sndbeg(ul_drv *udrv, int ret_code)
{
  udrv->xor_sum=0;
  if(udrv->con_flg&4) udrv->char_buff=UL_BEG;
  else udrv->char_buff=udrv->con_dadr|0x100;
  UL_FCALL2(*udrv->fnc_sndch,uld_sndbeg_1);
  return UL_RC_PROC;
};

/*-- Helper functions for sndend --*/

int uld_sndend_3(ul_drv *udrv, int ret_code)
{
  if(ret_code<0) {UL_FRET;return ret_code;};
  UL_FRET;
  if(ret_code<0) return ret_code;
  if(udrv->char_buff==UL_ACK) return UL_RC_PROC;
  if(udrv->char_buff==UL_NAK) return UL_RC_ENAK;
  if(udrv->char_buff==UL_WAK) return UL_RC_EWAK;
  return UL_RC_EREPLY;
};

int uld_sndend_2(ul_drv *udrv, int ret_code)
{
  if(ret_code<0) {UL_FRET;return ret_code;};
  if(!(udrv->con_flg&1))
  {
    UL_FRET;
    return UL_RC_PROC;
  };
  udrv->wait_time=4;
  UL_FCALL2(*udrv->fnc_wait,uld_sndend_3);
  return UL_RC_PROC;
};

int uld_sndend_1(ul_drv *udrv, int ret_code)
{
  if(ret_code<0) {UL_FRET;return ret_code;};
  udrv->char_buff=udrv->xor_sum&0xFF;
  UL_FCALL2(*udrv->fnc_sndch,uld_sndend_2);
  return UL_RC_PROC;
};

/*** Send frame end ***/
int uld_sndend(ul_drv *udrv, int ret_code)
{
  static int sndend_arr[]={UL_END,UL_ARQ,UL_PRQ,UL_AAP};
  udrv->char_buff=sndend_arr[udrv->con_flg];
  UL_FCALL2(*udrv->fnc_sndch,uld_sndend_1);
  return UL_RC_PROC;
};

/*-- Helper functions for recbeg --*/

int uld_recbeg_3(ul_drv *udrv, int ret_code)
{
  if(ret_code<0) {UL_FRET;return ret_code;};
  if(udrv->char_buff&0xFF00) {UL_FRET; return UL_RC_ERECBEG;};
  udrv->con_cmd=(uchar)udrv->char_buff;
  UL_FRET;
  return UL_RC_PROC;
};

int uld_recbeg_2(ul_drv *udrv, int ret_code)
{
  if(ret_code<0) {UL_FRET;return ret_code;};
  if(udrv->char_buff&0xFF00) {UL_FRET; return UL_RC_ERECBEG;};
  udrv->con_sadr=(uchar)udrv->char_buff;
  udrv->wait_time=4;
  UL_FCALL2(*udrv->fnc_wait,uld_recbeg_3);
  return UL_RC_PROC;
};

int uld_recbeg_1(ul_drv *udrv, int ret_code)
{
  if((udrv->char_buff!=(udrv->my_adr|0x100))&&
     (udrv->char_buff!=0x100)&&
     ((udrv->char_buff!=UL_BEG)||(!(udrv->con_flg&8))))
  { UL_FRET;
    if((udrv->char_buff&0x180)==0x100)
     return UL_RC_ERECADR;
    else
     return UL_RC_ERECBEG;
  };
  udrv->con_dadr=udrv->char_buff&0xFF;
  udrv->con_flg|=8;
  udrv->wait_time=4;
  UL_FCALL2(*udrv->fnc_wait,uld_recbeg_2);
  return UL_RC_PROC;
};

/*** Receive frame begin ***/
int uld_recbeg(ul_drv *udrv, int ret_code)
{
  udrv->xor_sum=0;
  udrv->wait_time=4;
  UL_FCALL2(*udrv->fnc_wait,uld_recbeg_1);
  return UL_RC_PROC;
};

/*-- Helper functions for recend --*/

int uld_recend_3(ul_drv *udrv, int ret_code)
{
  UL_FRET;
  if(udrv->con_flg&0xC0) return UL_RC_ERECEND;
  return UL_RC_PROC;
};

int uld_recend_2(ul_drv *udrv, int ret_code)
{
  if(ret_code<0) {UL_FRET;return ret_code;};
  if(udrv->char_buff&0xFF00) {UL_FRET; return UL_RC_ERECPROT;};
  if((udrv->xor_sum-1)&0xFF) udrv->con_flg|=0x80;
  if(!(udrv->con_flg&1))
  {
    UL_FRET;
    if(udrv->con_flg&0xC0) return UL_RC_ERECEND;
    return UL_RC_PROC;
  };
  if(udrv->con_flg&0x80) udrv->char_buff=UL_NAK;
  else if(udrv->con_flg&0x40) udrv->char_buff=UL_WAK;
  else udrv->char_buff=UL_ACK;
  UL_FCALL2(*udrv->fnc_sndch,uld_recend_3);
  return UL_RC_PROC;
};

int uld_recend_1(ul_drv *udrv, int ret_code)
{
  if(ret_code<0) {UL_FRET;return ret_code;};
  if(!(udrv->char_buff&0x100))
  {
    udrv->wait_time=4;
    UL_FCALL(*udrv->fnc_wait);
    udrv->con_flg|=0x80;
    return UL_RC_PROC;
  };
  udrv->con_flg&=~3;
  switch(udrv->char_buff)
  {
    case UL_END: break;
    case UL_ARQ: udrv->con_flg|=1; break;
    case UL_PRQ: udrv->con_flg|=2; break;
    case UL_AAP: udrv->con_flg|=3; break;
    default: UL_FRET; return UL_RC_ERECPROT;
  };
  udrv->wait_time=4;
  UL_FCALL2(*udrv->fnc_wait,uld_recend_2);
  return UL_RC_PROC;
};

/*** Receive frame end ***/
int uld_recend(ul_drv *udrv, int ret_code)
{
  if((udrv->char_buff&0xFF00)==0x100) 
  {
    UL_FNEXT(uld_recend_1);
  }else{
    udrv->wait_time=4;
    UL_FCALL2(*udrv->fnc_wait,uld_recend_1);
  };
  return UL_RC_PROC;
};

#endif /*defined(UL_WITH_UARTS) || defined(UL_WITH_PCI)*/

/*******************************************************************/
/* Data iterator */

long ul_di_seek(ul_data_it *di,long pos, int whence)
{
  switch(whence)
  {
    case 1: pos+=di->pos; break;
    case 2: pos+=UL_BLK_HEAD(di->head_blk).len; break;
  };
  if(pos<di->pos)
  {
    di->blk=di->head_blk;
    di->pos=0;
    di->ptr=UL_BLK_HEAD(di->head_blk).fdata;
    if(pos<0) return 0;
  };
  di->ptr+=pos-di->pos;
  di->pos=pos;
  return pos;
};

uchar ul_di_read1(ul_data_it *di)
{
  uchar ch;
  if(ul_di_adjust(di))
    ch=*ul_di_byte(di);
  else ch=0;
  ul_di_inc(di);
  return ch;
};

uchar ul_di_write1(ul_data_it *di, uchar ch)
{
  ul_mem_blk *blk;
  while(!ul_di_adjust(di))
  {
    if(!(blk=ul_alloc_blk(UL_BLK_HEAD(di->head_blk).bll->udrv))) return 0;
    di->blk->next=blk;
  };
  *ul_di_byte(di)=ch;
  ul_di_inc(di);
  return 1;
};

int ul_di_read(ul_data_it *di, char *buf, int count)
{
  int cn;
  int len;

  cn=count;
  while(cn>0)
  {
    if(!ul_di_adjust(di))
    {
      memset(buf,0,cn);
      cn=0;
      break;
    };
    len=ul_di_atonce(di);
    if(len>cn) len=cn;
    memcpy(buf,ul_di_byte(di),len);
    ul_di_add(di,len);
    buf+=len;
    cn-=len;
  };
  return count-cn;
}

int ul_di_write(ul_data_it *di, const char *buf, int count)
{
  int cn;
  int len;
  ul_mem_blk *blk;

  cn=count;
  while(cn>0)
  {
    while(!ul_di_adjust(di))
    {
      if(!(blk=ul_alloc_blk(UL_BLK_HEAD(di->head_blk).bll->udrv))) 
	{count-=cn; break;};
      memset(UL_BLK_NDATA(blk),0,UL_BLK_SIZE);
      di->blk->next=blk;
    };
    len=ul_di_atonce(di);
    if(len>cn) len=cn;
    memcpy(ul_di_byte(di),buf,len);
    ul_di_add(di,len);
    buf+=len;
    cn-=len;
  };
  return count;
}

#if defined(UL_WITH_UARTS) || defined(UL_WITH_PCI)

/*-- Helper functions for snddata --*/
int uld_snddata_1(ul_drv *udrv, int ret_code)
{
  if(ret_code<0) {UL_FRET;return ret_code;};
  if(udrv->con_di.trans_len--)
  {
    udrv->char_buff=ul_di_read1(&udrv->con_di);
    UL_FCALL(*udrv->fnc_sndch);
  } else UL_FRET;
  return UL_RC_PROC;
};

/*** Send data ***/
int uld_snddata(ul_drv *udrv, int ret_code)
{
  ul_di_init(&udrv->con_di,udrv->con_frame);
  udrv->con_di.trans_len=UL_BLK_HEAD(udrv->con_frame).len;
  if(!udrv->con_di.trans_len) 
    {UL_FRET; return UL_RC_PROC;};
  UL_FNEXT(uld_snddata_1);
  return UL_RC_PROC;
};

/*-- Helper functions for recdata --*/
int uld_recdata_1(ul_drv *udrv, int ret_code)
{
  if(ret_code<0) {UL_FRET;return ret_code;};
  if(!(udrv->char_buff&0xFF00))
  {
    ret_code=ul_di_write1(&udrv->con_di,(uchar)udrv->char_buff);
    if(!ret_code) return UL_RC_ENOMEM;
    if(--udrv->con_di.trans_len)
    {
      UL_FCALL(*udrv->fnc_recch);
      return UL_RC_PROC;
    };
  };
  if(!(UL_BLK_HEAD(udrv->con_frame).flg&UL_BFL_LNMM))
    UL_BLK_HEAD(udrv->con_frame).len=udrv->con_di.pos;
  UL_FRET;
  return UL_RC_PROC;
};

/*** Receive data ***/
int uld_recdata(ul_drv *udrv, int ret_code)
{
  ul_di_init(&udrv->con_di,udrv->con_frame);
  if(UL_BLK_HEAD(udrv->con_frame).flg&UL_BFL_LNMM)
    udrv->con_di.trans_len=UL_BLK_HEAD(udrv->con_frame).len;
  else
    udrv->con_di.trans_len=-1;
  if(!udrv->con_di.trans_len) {UL_FRET; return UL_RC_PROC;};
  UL_FCALL2(*udrv->fnc_recch,uld_recdata_1);
  return UL_RC_PROC;
};

/*******************************************************************/
/* Message and frame proccessor */

int uld_prmess_frame(ul_drv *udrv, int ret_code);
int uld_prmess(ul_drv *udrv, int ret_code);

/*-- Helper functions for prmess --*/
int uld_prmess_error(ul_drv *udrv, int ret_code)
{
  UL_FRET;
  return ret_code;
};

int uld_prmess_next(ul_drv *udrv, int ret_code)
{
  if(ret_code<0) {UL_FNEXT(uld_prmess_error);return UL_RC_EPRMESS;};
  if(UL_BLK_HEAD(udrv->con_frame).flg&UL_BFL_TAIL)
  {
    if(UL_BLK_HEAD(udrv->con_frame).next)
    {
      udrv->con_frame=UL_BLK_HEAD(udrv->con_frame).next;
      UL_FNEXT(uld_prmess_frame);
      return UL_RC_PROC;
    };
    UL_FRET;
    return UL_RC_ENOTAIL;
  };
  UL_FRET;
  return UL_RC_PROC;
};

int uld_prmess_rec2(ul_drv *udrv, int ret_code)
{
  if(ret_code<0) {UL_FNEXT(uld_prmess_error);return UL_RC_EPRMESS;};
  UL_FCALL2(uld_recend,uld_prmess_next);
  return UL_RC_PROC;
};

int uld_prmess_rec1(ul_drv *udrv, int ret_code)
{
  if(ret_code<0) {UL_FNEXT(uld_prmess_error);return UL_RC_EPRMESS;};
  UL_BLK_HEAD(udrv->con_frame).dadr=udrv->con_dadr;
  UL_BLK_HEAD(udrv->con_frame).sadr=udrv->con_sadr;
  UL_BLK_HEAD(udrv->con_frame).cmd=udrv->con_cmd;
  UL_FCALL2(uld_recdata,uld_prmess_rec2);
  return UL_RC_PROC;
};

int uld_prmess_snd2(ul_drv *udrv, int ret_code)
{
  if(ret_code<0) {UL_FNEXT(uld_prmess_error);return UL_RC_EPRMESS;};
  UL_FCALL2(uld_sndend,uld_prmess_next);
  return UL_RC_PROC;
};

int uld_prmess_snd1(ul_drv *udrv, int ret_code)
{
  if(ret_code<0) {UL_FNEXT(uld_prmess_error);return UL_RC_EPRMESS;};
  UL_FCALL2(uld_snddata,uld_prmess_snd2);
  return UL_RC_PROC;
};

int uld_prmess_frame(ul_drv *udrv, int ret_code)
{
  if(ret_code<0) {UL_FNEXT(uld_prmess_error);return UL_RC_EPRMESS;};
  if(UL_BLK_HEAD(udrv->con_frame).flg&UL_BFL_SND)
  {
    udrv->con_dadr=UL_BLK_HEAD(udrv->con_frame).dadr;
    udrv->con_sadr=UL_BLK_HEAD(udrv->con_frame).sadr;
    udrv->con_cmd=UL_BLK_HEAD(udrv->con_frame).cmd;
    udrv->con_flg=UL_BLK_HEAD(udrv->con_frame).flg&3;
    UL_FCALL2(uld_sndbeg,uld_prmess_snd1);
  }
  else if(UL_BLK_HEAD(udrv->con_frame).flg&UL_BFL_REC)
  {
    udrv->con_flg=8;
    UL_FCALL2(uld_recbeg,uld_prmess_rec1);
  }
  else UL_FNEXT(uld_prmess_next);
  return UL_RC_PROC;
};

/*** Proccess one outgoing message ***/
int uld_prmess(ul_drv *udrv, int ret_code)
{
  if(!udrv->con_message)
  {
    UL_FRET;
    return UL_RC_PROC;
  };
  udrv->con_frame=udrv->con_message;
  UL_FNEXT(uld_prmess_frame);
  return UL_RC_PROC;
};

#endif /*defined(UL_WITH_UARTS) || defined(UL_WITH_PCI)*/

/*******************************************************************/
/* uLan memory management and bidirection linked lists */

int ul_mem_init(ul_drv *udrv, int req_size)
{
  void **pptr;
  ul_mem_blk *bptr;
  pptr=(void **)(udrv->mem_ptr=MALLOC(req_size+2*sizeof(void*)));
  if(!pptr) return -1;
  *(pptr++)=0;
  *(pptr++)=(void *)req_size;
  udrv->free_blk=bptr=(ul_mem_blk *)pptr;
  udrv->free_blk_cnt=1;

  while((req_size-=sizeof(ul_mem_blk))>=sizeof(ul_mem_blk))
  {
    bptr->next=bptr+1;
    bptr++;
    udrv->free_blk_cnt++;
  }; 
  bptr->next=NULL;
  udrv->prep_bll.first=udrv->prep_bll.last=NULL;
  udrv->prep_bll.cnt=0;
  udrv->prep_bll.udrv=udrv; 
  udrv->work_bll.first=udrv->work_bll.last=NULL;
  udrv->work_bll.cnt=0;
  udrv->work_bll.udrv=udrv; 
  udrv->proc_bll.first=udrv->proc_bll.last=NULL;
  udrv->proc_bll.cnt=0;
  udrv->proc_bll.udrv=udrv;
  udrv->opan_bll.first=udrv->opan_bll.last=NULL;
  udrv->opan_bll.cnt=0;
  udrv->opan_bll.udrv=udrv;
  udrv->con_message=NULL;

  return 0;
};

int ul_mem_done(ul_drv *udrv)
{
  void *ptr;
  if(udrv->mem_ptr)
  { ptr=*(void **)udrv->mem_ptr;
    FREE(udrv->mem_ptr);
    udrv->mem_ptr=0;
  };
  return 0;
};

/*** Insert frame blk on list bll ***/
void ul_bll_ins(ul_blk_bll *bll, ul_mem_blk *blk)
{
  UL_DRV_LOCK_FINI
  /* ul_drv *udrv=bll->udrv; */
  UL_BLK_HEAD(blk).bll=bll;
  /* UL_BLK_HEAD(blk).next=NULL; */
  /* UL_BLK_HEAD(blk).ref_cnt=1; */
  /* UL_BLK_HEAD(blk).flg|=UL_BFL_LOCK; */
  UL_DRV_LOCK;
  if((UL_BLK_HEAD(blk).prev=bll->last)!=NULL)
    UL_BLK_HEAD(bll->last).next=blk;
  else bll->first=blk;
  bll->last=blk;
  bll->cnt++;
  UL_DRV_UNLOCK;
};

/*** Remove frame blk from its bll ***/
void ul_bll_del(ul_mem_blk *blk)
{
  UL_DRV_LOCK_FINI
  ul_blk_bll *bll;
  ul_mem_blk *p;
  bll=UL_BLK_HEAD(blk).bll;
  UL_DRV_LOCK;
  bll->cnt--;

  if((p=UL_BLK_HEAD(blk).next)!=NULL) 
  { UL_BLK_HEAD(p).prev=UL_BLK_HEAD(blk).prev;
    UL_BLK_HEAD(blk).next=NULL;
  }else bll->last=UL_BLK_HEAD(blk).prev;

  if((p=UL_BLK_HEAD(blk).prev)!=NULL) 
  { UL_BLK_HEAD(p).next=UL_BLK_HEAD(blk).next;
    UL_BLK_HEAD(blk).prev=NULL;
  }else bll->first=UL_BLK_HEAD(blk).next;

  UL_BLK_HEAD(blk).bll=NULL;
  UL_DRV_UNLOCK;
};

/*** Move one message starting at beg_blk to list new_bll ***/
void ul_bll_move_mes(ul_blk_bll *new_bll, ul_mem_blk *beg_blk)
{
  UL_DRV_LOCK_FINI
  /* ul_drv *udrv=bll->udrv; */
  ul_mem_blk *end_blk;
  int         cnt_blk=1;
  ul_blk_bll *old_bll;
  ul_mem_blk *p;

  if(beg_blk==NULL) return;
  end_blk=beg_blk;
  while((UL_BLK_HEAD(end_blk).flg&UL_BFL_TAIL)&&UL_BLK_HEAD(end_blk).next)
    { end_blk=UL_BLK_HEAD(end_blk).next; cnt_blk++; };
  UL_BLK_HEAD(end_blk).flg&=~UL_BFL_TAIL;

  UL_DRV_LOCK;
  if((old_bll=UL_BLK_HEAD(beg_blk).bll)!=NULL)
  {
    old_bll->cnt-=cnt_blk;
    if((p=UL_BLK_HEAD(beg_blk).prev)!=NULL)
    { UL_BLK_HEAD(p).next=UL_BLK_HEAD(end_blk).next;
      UL_BLK_HEAD(beg_blk).prev=NULL;
    } else old_bll->first=UL_BLK_HEAD(end_blk).next;
    if((p=UL_BLK_HEAD(end_blk).next)!=NULL)
    { UL_BLK_HEAD(p).prev=UL_BLK_HEAD(beg_blk).prev;
      UL_BLK_HEAD(end_blk).next=NULL;
    } else old_bll->last=UL_BLK_HEAD(beg_blk).prev;
  };
  UL_DRV_UNLOCK;

  p=beg_blk;
  do UL_BLK_HEAD(p).bll=new_bll; 
  while ((p=UL_BLK_HEAD(p).next)!=NULL);

  if(new_bll)
  {
    UL_DRV_LOCK;
    if((UL_BLK_HEAD(beg_blk).prev=new_bll->last)!=NULL)
      UL_BLK_HEAD(new_bll->last).next=beg_blk;
    else new_bll->first=beg_blk;
    new_bll->last=end_blk;
    new_bll->cnt+=cnt_blk;
    UL_DRV_UNLOCK;
  };
};

void ul_free_mes(ul_drv *udrv,ul_mem_blk *beg_blk)
{
  ul_mem_blk *blk, *next_blk;
  if(beg_blk==NULL) return;
  ul_bll_move_mes(NULL,beg_blk);
  do {
    next_blk=beg_blk;
    beg_blk=UL_BLK_HEAD(beg_blk).next;
    do {
      blk=next_blk;
      next_blk=blk->next;
      ul_free_blk(udrv,blk);
    } while(next_blk!=NULL);
  } while(beg_blk!=NULL);
};

/*** Free one message starting at beg_blk ***/
INLINE void ul_bll_free_mes(ul_mem_blk *beg_blk)
{
  if(beg_blk==NULL) return;
  ul_free_mes(UL_BLK_HEAD(beg_blk).bll->udrv,beg_blk);
};

ul_mem_blk *ul_new_frame_head(ul_drv *udrv, uchar dadr,
                         uchar sadr, uchar cmd,  unsigned flg)
{ ul_mem_blk *blk;
  blk=ul_alloc_blk(udrv);
  if(blk==NULL) return NULL;
  UL_BLK_HEAD(blk).prev=NULL;
  UL_BLK_HEAD(blk).next=NULL;
  UL_BLK_HEAD(blk).bll=NULL;
  UL_BLK_HEAD(blk).retry_cnt=0;
  UL_BLK_HEAD(blk).ref_cnt=0;
  UL_BLK_HEAD(blk).flg=flg;
  UL_BLK_HEAD(blk).len=0;
  UL_BLK_HEAD(blk).cmd=cmd;
  UL_BLK_HEAD(blk).dadr=dadr;
  UL_BLK_HEAD(blk).sadr=sadr;
  UL_BLK_HEAD(blk).stamp=0;
  return blk;
};

void ul_tail_frame_head(ul_mem_blk *beg_blk, ul_mem_blk *blk)
{
  while(UL_BLK_HEAD(beg_blk).next!=NULL) 
    beg_blk=UL_BLK_HEAD(beg_blk).next;
  UL_BLK_HEAD(beg_blk).next=blk;
  UL_BLK_HEAD(blk).prev=beg_blk;
  UL_BLK_HEAD(beg_blk).flg|=UL_BFL_TAIL;
};

unsigned ul_gen_stamp(void)
{
  UL_DRV_LOCK_FINI
  static unsigned ul_stamp_cnt=0; 
  unsigned stamp;
  UL_DRV_LOCK;
  ul_stamp_cnt++;
  ul_stamp_cnt&=0x7fffffff;
  if(!ul_stamp_cnt)ul_stamp_cnt++;
  stamp=ul_stamp_cnt;
  UL_DRV_UNLOCK;
  return stamp; 
};

int ul_inc_ref_cnt(ul_mem_blk *mes)
{
  int cnt;
  UL_DRV_LOCK_FINI
  UL_DRV_LOCK;
  UL_BLK_HEAD(mes).ref_cnt++;
  cnt=UL_BLK_HEAD(mes).ref_cnt;
  UL_DRV_UNLOCK;
  return cnt;
};

int ul_dec_ref_cnt(ul_mem_blk *mes)
{
  int cnt;
  UL_DRV_LOCK_FINI
  UL_DRV_LOCK;
  if(UL_BLK_HEAD(mes).ref_cnt)UL_BLK_HEAD(mes).ref_cnt--;
  cnt=UL_BLK_HEAD(mes).ref_cnt;
  UL_DRV_UNLOCK;
  if(!cnt)ul_bll_free_mes(mes);
  return cnt;
};

#if defined(UL_WITH_UARTS) || defined(UL_WITH_PCI)
/*******************************************************************/
/* uLan main driver automata loop */

ul_call_fnc uld_drvloop_proc_1, uld_drvloop_proc_2;
ul_call_fnc uld_drvloop_rec_1, uld_drvloop_rec_2, uld_drvloop_rec_3;

int uld_drvloop(ul_drv *udrv, int ret_code)
{
  if(((udrv->char_buff==(udrv->my_adr|0x100))||
      (udrv->char_buff==0x100))&&
     (udrv->char_buff==udrv->last_ctrl))
  { /* it seems, that somebody wants us  */
    udrv->con_flg=0;
    udrv->xor_sum=udrv->char_buff+1;
    UL_FCALL2(uld_recbeg_1,uld_drvloop_rec_1);
    return UL_RC_PROC;
  };

  if(udrv->prep_bll.first!=NULL) /* we have something to proccess  */
  {
    LOG_CHIO(" call connect ");
    UL_FCALL2(*udrv->fnc_connect,uld_drvloop_proc_1);
    return UL_RC_PROC;
  };

  udrv->char_buff=0;
  UL_FCALL(*udrv->fnc_recch);
  return UL_RC_PROC;
};

int uld_drvloop_proc_error(ul_drv *udrv)
{ ul_mem_blk *mes;
  udrv->fnc_timeout=NULL;
  LOG_MESSAGES("uLan : send message ERROR\n");
  if((mes=udrv->con_message)!=NULL)
  {udrv->con_message=NULL;
   if((UL_BLK_HEAD(mes).flg&UL_BFL_NORE)||
      (++UL_BLK_HEAD(mes).retry_cnt>=UL_RETRY_CNT))
   { UL_BLK_HEAD(mes).flg|=UL_BFL_FAIL;
     ul_bll_move_mes(&udrv->proc_bll,mes);
     LOG_FAILS(KERN_ERR "uLan : send message (d:%d,s:%d,c:%X,f:%X,l:%d) FAILED after %d attempts\n",
                   UL_BLK_HEAD(mes).dadr,UL_BLK_HEAD(mes).sadr,
                   UL_BLK_HEAD(mes).cmd,UL_BLK_HEAD(mes).flg,
                   UL_BLK_HEAD(mes).len,UL_BLK_HEAD(mes).retry_cnt);
     SCHEDULE_BH(udrv);
   }else if(UL_BLK_HEAD(mes).flg&UL_BFL_REWA)
     ul_bll_move_mes(&udrv->prep_bll,mes);
  };
  udrv->char_buff=0x1FF; 
  return 0;
};

int uld_drvloop_proc_1(ul_drv *udrv, int ret_code)
{
  if(ret_code<0) {UL_FNEXT(uld_drvloop); return UL_RC_PROC;};
  LOG_MESSAGES("uLan : begin of send message\n");
  udrv->fnc_timeout=&uld_drvloop_proc_error;
  udrv->con_message=udrv->prep_bll.first;
  UL_FCALL2(uld_prmess,uld_drvloop_proc_2);
  return UL_RC_PROC;
};

int uld_drvloop_proc_2(ul_drv *udrv, int ret_code)
{
  ul_mem_blk *mes;
  udrv->fnc_timeout=NULL;
  udrv->char_buff=udrv->my_adr|0x180; 
  UL_FCALL2(*udrv->fnc_sndch,uld_drvloop);
  if(udrv->con_message==NULL) return UL_RC_PROC;
  if(ret_code<0)
  {uld_drvloop_proc_error(udrv);
  }else{
    LOG_MESSAGES("uLan : send message finished OK\n");
    mes=udrv->con_message;
    udrv->con_message=NULL;
    if(UL_BLK_HEAD(mes).flg&UL_BFL_M2IN || 
       UL_BLK_HEAD(mes).dadr==udrv->my_adr)
    {
      ul_bll_move_mes(&udrv->proc_bll,mes);
      SCHEDULE_BH(udrv);
    } else
     ul_bll_free_mes(mes);
  };
  return UL_RC_PROC;
};

int uld_drvloop_rec_error(ul_drv *udrv)
{ ul_mem_blk *mes;
  udrv->fnc_timeout=NULL;
  LOG_MESSAGES("uLan : receive message ERROR\n");
  if((mes=udrv->con_message)!=NULL)
  { udrv->con_message=NULL;
    ul_bll_free_mes(mes);
  };
  udrv->char_buff=0x1FF;
  return 0;
};

int uld_drvloop_rec_1(ul_drv *udrv, int ret_code)
{ ul_mem_blk *mes;
  if(ret_code<0) {UL_FNEXT(uld_drvloop); return UL_RC_PROC;};
  LOG_MESSAGES("uLan : begin of receive message\n");
  mes=ul_new_frame_head(udrv,udrv->con_dadr,udrv->con_sadr, 
                	udrv->con_cmd,0);
  if(mes==NULL) 
  { LOG_FAILS(KERN_ERR "uLan : NO memory for receive\n");
    UL_FNEXT(uld_drvloop); return UL_RC_PROC;
  };
  ul_bll_ins(&udrv->work_bll,mes);
  udrv->con_message=mes;
  udrv->fnc_timeout=&uld_drvloop_rec_error;
  udrv->con_frame=mes;
  UL_FCALL2(uld_recdata,uld_drvloop_rec_2);
  return UL_RC_PROC;
};

int uld_drvloop_rec_2(ul_drv *udrv, int ret_code)
{
  if(ret_code<0) 
  { uld_drvloop_rec_error(udrv);
    UL_FNEXT(uld_drvloop); return UL_RC_PROC;
  };
  UL_FCALL2(uld_recend,uld_drvloop_rec_3);
  return UL_RC_PROC;
};

int uld_drvloop_rec_3(ul_drv *udrv, int ret_code)
{ ul_mem_blk *mes;
  if(ret_code<0) 
  { uld_drvloop_rec_error(udrv);
    UL_FNEXT(uld_drvloop); return UL_RC_PROC;
  };
  mes=udrv->con_message;
  udrv->fnc_timeout=NULL;
  udrv->con_message=NULL;
  UL_BLK_HEAD(mes).stamp=ul_gen_stamp();
  ul_bll_move_mes(&udrv->proc_bll,mes);
  SCHEDULE_BH(udrv);
  UL_FNEXT(uld_drvloop);
  LOG_MESSAGES("uLan : end of receive message\n");
  return UL_RC_PROC;
};

int ul_max_cycle_cnt=0;
int ul_max_cycle_time=0;
int ul_irq_seq_num=0;
#ifdef ENABLE_UL_IRQ_STALE_WDG	
int ul_irq_stale_wdg_cnt=0;
#endif /* ENABLE_UL_IRQ_STALE_WDG */

/*** Interrupt entry point ***/
#ifdef _WIN32
BOOLEAN uld_irq_handler(IN PKINTERRUPT intno, IN PVOID dev_id)
#elif defined(__DJGPP__)||defined(USER_SPACE_EMU_IRQ)
void uld_irq_handler(int intno, void *dev_id, void *regs)
#else /* _WIN32 */
irqreturn_t uld_irq_handler(int intno, void *dev_id, struct pt_regs *regs)
#endif /* _WIN32 */
{
  #ifdef UL_GLOBAL_IRQ_LOCK_FINI
   UL_GLOBAL_IRQ_LOCK_FINI
  #endif
  int irq_loop_cnt=0;
  int cycle_cnt=0;
  int cycle_time;
  int ret_code;
  int seq_num=++ul_irq_seq_num;
  ul_drv* udrv=(ul_drv*)dev_id;
  if(udrv->magic!=UL_DRV_MAGIC)
  {
    #ifdef FOR_LINUX_KERNEL
     panic("uld_irq_handler : BAD udrv magic !!!");
    #elif defined(_WIN32)
     UL_PRINTF("uld_irq_handler : BAD udrv magic !!!\n");
     return FALSE;
    #elif defined(__DJGPP__)
     UL_PRINTF("uld_irq_handler : BAD udrv magic !!!\n");
     return;
    #else
     error("uld_irq_handler : BAD udrv magic !!!");
    #endif
  }
  #ifdef UL_GLOBAL_IRQ_LOCK
   UL_GLOBAL_IRQ_LOCK;
  #endif
  if(uld_test_dfl(udrv,CHIPOK))
  do
  { 
    #ifdef ENABLE_UL_IRQ_STALE_WDG	
    /* last rescue code for stale IRQ !!!!!! */
    if(ul_irq_stale_wdg_cnt++>10000)
    {
      SCHEDULE_BH(udrv); /* Test activation of BH */
      if(ul_irq_stale_wdg_cnt>20000)
      {
	LOG_FATAL(KERN_CRIT "Stale IRQ uLan IRQ signal !!!\n");
	udrv->fnc_cctrl(udrv,UL_CC_GENIRQ,0);
	if(ul_irq_stale_wdg_cnt<25000) break;
	#ifdef FOR_LINUX_KERNEL
	 LOG_FATAL(KERN_CRIT "Disabling IRQ %d forever !!!\n",intno);
	 disable_irq_nosync(intno);
	#endif
	break;
      }
    } 
    #endif /* ENABLE_UL_IRQ_STALE_WDG */
    do /* IRQ fully SMP and re-entrant processing */
    {
      uld_set_dfl(udrv,ASK_ISR);
      if(uld_test_and_set_dfl(udrv,IN_ISR))
      { LOG_IRQ("uLan IRQ %d arrived, but else is in ISR\n",intno);
        break;
      };

      /* only one CPU in no nested levels at time */
      uld_clear_dfl(udrv,ASK_ISR);

      while((ret_code=udrv->fnc_pool(udrv))!=0)
      {
	LOG_SEQNUM("\n<<< seq %d ",seq_num);
	LOG_IRQ("uLan IRQ %d arrived, code = %d\n",intno,ret_code);
	cycle_time=RDTSC;
	while((ret_code=udrv->fnc_act(udrv,ret_code))!=0) 
	  if(cycle_cnt++>100) 
	    {LOG_FATAL(KERN_CRIT "Infinite loop in uLan IRQ handler !!!\n");break;};
	cycle_time=RDTSC-cycle_time;
	if(cycle_cnt>ul_max_cycle_cnt) ul_max_cycle_cnt=cycle_cnt;
	if(cycle_time>ul_max_cycle_time) ul_max_cycle_time=cycle_time;
	LOG_SEQNUM(" seq %d >>>",seq_num);
	if(irq_loop_cnt++>10)
	{ LOG_FATAL(KERN_CRIT "Stale IRQ - uLan state handlers !!!\n");
	  udrv->fnc_cctrl(udrv,UL_CC_GENIRQ,0);
	  break;
	};
	uld_atomic_clear_dfl(udrv,NACTIV);
      };
      uld_atomic_clear_dfl(udrv,IN_ISR);
    }
    while(uld_atomic_test_dfl(udrv,ASK_ISR));

  }while(0);
  #ifdef UL_GLOBAL_IRQ_UNLOCK
   UL_GLOBAL_IRQ_UNLOCK;
  #endif
  #ifdef _WIN32
   return irq_loop_cnt?TRUE:FALSE;
  #elif defined(KERNEL)
   return IRQ_RETVAL(irq_loop_cnt);
  #endif
};

/*** Timeout entry point ***/
void uld_timeout(ul_drv *udrv)
{
  #ifdef UL_GLOBAL_IRQ_LOCK_FINI
   UL_GLOBAL_IRQ_LOCK_FINI
  #endif
  #if defined(_WIN32)&&defined(UL_WITH_PCI)
   KIRQL old_irql;
  #endif
  int cycle_cnt=0;
  int cycle_time;
  int ret_code=0;
  #ifdef UL_GLOBAL_IRQ_LOCK
   UL_GLOBAL_IRQ_LOCK;
  #endif
  LOG_IRQ("uLan timeout activated\n");
  if(uld_test_dfl(udrv,CHIPOK))
  {
    #ifdef FOR_LINUX_KERNEL
    if(udrv->irq) disable_irq_nosync(udrv->irq);
    #elif defined(_WIN32)&&defined(UL_WITH_PCI)
    KeRaiseIrql(uL_SpinLock_Irql,&old_irql);
    #endif
    do
    {
      if(uld_test_and_set_dfl(udrv,IN_ISR))
      { LOG_IRQ("uLan timeout, but else is in ISR\n");
        break;
      };

      /* only one CPU with no nested levels goes there */
      uld_clear_dfl(udrv,ASK_ISR);
      if((ret_code=udrv->fnc_pool(udrv))==0)
      {/* if first pass check for timeout conditions */
	if(!cycle_cnt&&(uld_atomic_test_dfl(udrv,NACTIV)||
            (udrv->last_ctrl&0x80)))
	{ cycle_cnt++;
	  if(udrv->fnc_timeout) ret_code=udrv->fnc_timeout(udrv);
	  if(ret_code==0)
	  {udrv->fnc_sp=&udrv->fnc_stack[0];
	   udrv->fnc_act=&uld_drvloop;
	   ret_code=UL_RC_ACTIVATE;
	  };
	};
      };
      cycle_time=RDTSC;
      while(ret_code)
      { ret_code=udrv->fnc_act(udrv,ret_code);
        if(cycle_cnt++>100) 
	  {LOG_FATAL(KERN_CRIT "Infinite loop in uLan Timeout handler !!!\n");break;};
      };
      cycle_time=RDTSC-cycle_time;
      if(cycle_cnt>ul_max_cycle_cnt) ul_max_cycle_cnt=cycle_cnt;
      if(cycle_time>ul_max_cycle_time) ul_max_cycle_time=cycle_time;
      uld_clear_dfl(udrv,NACTIV);
      uld_atomic_clear_dfl(udrv,IN_ISR);
    }
    while(uld_atomic_test_dfl(udrv,ASK_ISR));
    #ifdef FOR_LINUX_KERNEL
     if(udrv->irq) enable_irq(udrv->irq);
    #elif defined(_WIN32)&&defined(UL_WITH_PCI)
     KeLowerIrql(old_irql);
    #endif
  };
  #ifdef UL_GLOBAL_IRQ_UNLOCK
   UL_GLOBAL_IRQ_UNLOCK;
  #endif
};

/* link watchdog timer function */
#ifdef _WIN32
VOID  ulan_wd_dpc(IN PKDPC Dpc,IN PVOID data,
		  IN PVOID arg1,IN PVOID arg2)
#else /* _WIN32 */
 void ulan_do_wd_timer(unsigned long data)
#endif /* _WIN32 */
{
  ul_drv* udrv=(ul_drv*)data;
  if(udrv->magic!=UL_DRV_MAGIC)
  {
    #ifdef FOR_LINUX_KERNEL
     panic("ulan_do_wd_timer : BAD udrv magic !!!");
    #elif defined(_WIN32)
     UL_PRINTF("ulan_do_wd_timer : BAD udrv magic !!!\n");
     return;
    #else
     error("ulan_do_wd_timer : BAD udrv magic !!!");
    #endif
  }
  if(!uld_test_dfl(udrv,CHIPOK)) return;

  if(udrv->prep_bll.first!=NULL)
  {
    if(uld_test_and_set_dfl(udrv,NACTIV)||
      (udrv->last_ctrl&0x80)) uld_timeout(udrv);
    #ifndef _WIN32  
     SCHEDULE_UDRV_WDTIM(udrv,jiffies+ULD_HZ/25+1);
    #else /* _WIN32 */
     { LARGE_INTEGER delay;
       delay=RtlConvertLongToLargeInteger(-40*10000);
       KeSetTimer(&udrv->wd_timer,delay,&udrv->wd_timer_dpc);
     }
    #endif /* _WIN32 */
  };
};

#endif /*defined(UL_WITH_UARTS) || defined(UL_WITH_PCI)*/

void ulan_stroke(ul_drv *udrv)
{
  UL_DRV_LOCK_FINI
  if(!uld_test_dfl(udrv,CHIPOK)) return;
  if(udrv->prep_bll.first==NULL) return;
  if(udrv->fnc_stroke)
  { udrv->fnc_stroke(udrv);
    return;
  }
  UL_DRV_LOCK;
  #ifndef _WIN32
  { int delay;
    delay=udrv->last_ctrl&0x80?0:ULD_HZ/100+2;
    SCHEDULE_UDRV_WDTIM(udrv,jiffies+delay);
  }
  #else /* _WIN32 */
  { LARGE_INTEGER delay;
    delay=udrv->last_ctrl&0x80?RtlConvertLongToLargeInteger(-1):
		 	      RtlConvertLongToLargeInteger(-40*10000);
    KeSetTimer(&udrv->wd_timer,delay,&udrv->wd_timer_dpc);
  } 
  #endif /* _WIN32 */
  UL_DRV_UNLOCK;
};

/*******************************************************************/
/* initialize and free of one driver */ 

ul_chip_type_ent ul_chip_types[12]={
    #ifdef UL_WITH_UARTS
	{"82510",u510_init,UL_CHIPT_RQPORT,1,0},
	{"16450",u450_init,UL_CHIPT_RQPORT,1,0},
    #endif /*UL_WITH_UARTS*/
    #ifdef UL_WITH_PCI
	{"16950-adv",u950pci_init,UL_CHIPT_PCI|UL_CHIPT_NOPROBE,4,
		0x16954000|U950PCI_CHOPT_TXDTRNEG|U950PCI_CHOPT_HSPDOSC},
	{"16950-pci",u950pci_init,UL_CHIPT_PCI|UL_CHIPT_NOPROBE,4,
		0x16954000|U950PCI_CHOPT_RXDONRI},
    #endif  /* UL_WITH_PCI */
	{NULL,NULL},
};

/* set initial state */
void ul_drv_new_init_state(ul_drv* udrv, int my_adr)
{
  /* init sequencer */
  udrv->fnc_sp=&udrv->fnc_stack[0];
 #if defined(UL_WITH_UARTS) || defined(UL_WITH_PCI)
  udrv->fnc_act=&uld_drvloop;
 #else
  udrv->fnc_act=&uld_null_fnc;
 #endif
  udrv->my_adr=my_adr;
 #ifdef FOR_LINUX_KERNEL
  init_waitqueue_head(&udrv->wq_kwt);
 #endif
  udrv->fnc_timeout=NULL;
  /* udrv->flags=0; */
  udrv->magic=UL_DRV_MAGIC;
  udrv->operators=NULL;
}

/* setups buffers, ports and irq for sucesfully detected device */
int ul_drv_new_start(ul_drv* udrv, unsigned buffer_size)
{ int ret;
  int test_cnt;
  unsigned long probe_irqs;
  /* init ports */
  ret=udrv->fnc_cctrl(udrv,UL_CC_PINIT,0);
  UL_PRINTF(KERN_INFO "uLan_pinit ret %d\n",ret);
  if(ret>=0)
  { /* take buffer memory */
    ret=ul_mem_init(udrv,buffer_size);
    UL_PRINTF(KERN_INFO "ul_mem_init ret %d\n",ret);
    if(ret>=0)
    { /* request irq */
      test_cnt=3;
      ret=udrv->fnc_cctrl(udrv,UL_CC_RQIRQ,0);
     #if defined(UL_WITH_PCI) || defined(UL_WITH_UARTS)
      if(ret==UL_RC_ENOFNC){ /* default IRQ handling */
	ret=0;
	#ifdef FOR_LINUX_KERNEL
	while(test_cnt--&&udrv->irq<=0)
	{ /* probe for IRQ */
	  probe_irq_off(probe_irq_on());
	  probe_irqs=probe_irq_on();
	  ret=udrv->fnc_cctrl(udrv,UL_CC_GENIRQ,1);
	  udelay(100);
	  udrv->irq=probe_irq_off(probe_irqs);
	  ret=udrv->fnc_cctrl(udrv,UL_CC_GENIRQ,0);
	  UL_PRINTF(KERN_INFO "uLan ul_drv_new : probed irq %d\n",udrv->irq);
	};
	#endif /* FOR_LINUX_KERNEL */
	if(udrv->irq<=0) ret=-1;
	  #ifndef _WIN32
	  else ret=request_irq(udrv->irq, uld_irq_handler, 
                              UL_SA_IRQ_FLAGS , "ulan", udrv);
	  #endif /* _WIN32 */
      }
     #endif /*defined(UL_WITH_PCI) || defined(UL_WITH_UARTS)*/
      if (ret>=0)
      { /* sucessful initialization */
	INIT_UDRV_WDTIM(udrv);
	INIT_UDRV_BH(udrv);
	uld_set_dfl(udrv,CHIPOK);
	return 0;
      } else UL_PRINTF(KERN_ERR "uLan ul_drv_new : cannot request irq %d\n",udrv->irq);
      ul_mem_done(udrv);
    } else UL_PRINTF(KERN_ERR "uLan ul_drv_new : no memory for buffers\n");
    ret=udrv->fnc_cctrl(udrv,UL_CC_PDONE,0);
  } else UL_PRINTF(KERN_ERR "uLan ul_drv_new : port 0x%lx init error\n",
                   udrv->port);
  return -1;
}

#ifndef _WIN32

/* initialize new driver from given parameters */
ul_drv *ul_drv_new(int port, int irq, int baud, int my_adr, char *chip_name)
{
  ul_drv* udrv;
  int ret;
  int chipidx=0;
  int chipprobe=1;
  /* covert chip name to index */
  if(chip_name&&strcmp(chip_name,"auto")) {
    for(chipidx=0;1;chipidx++){
      if(!ul_chip_types[chipidx].chip_name){
	UL_PRINTF(KERN_ERR "uLan ul_drv_new : unknown chip type %s\n",chip_name);
	return NULL;
      }
      if(!strcmp(ul_chip_types[chipidx].chip_name,chip_name)){
	chipprobe=0; break;
      }
    }
  }
  UL_PRINTF(KERN_INFO "uLan ul_drv_new : chip %s index %d probe %d\n",
 		 chip_name?chip_name:"auto",chipidx,chipprobe);
  /* mem for ul_drv */
  if(!(udrv=MALLOC(sizeof(ul_drv)))) return NULL;
  /* clear memory */
  memset(udrv,0,sizeof(ul_drv));
  /* set initial state */
  ul_drv_new_init_state(udrv, my_adr);
  /* init chip driver */
  ret=-1;
  do{
    ul_chip_init_fnc *chip_init;
    int chip_flags=ul_chip_types[chipidx].chip_flags;
    int chip_options=ul_chip_types[chipidx].chip_options;
    chip_init=ul_chip_types[chipidx].chip_init;
    if(chipprobe&&(ul_chip_types[chipidx].chip_flags&UL_CHIPT_NOPROBE))
      continue;
    /* if(ret=u510_init(udrv, port, irq, baud))>=0) */
    if((chip_flags&UL_CHIPT_RQPORT)&&!port) continue;
    if((chip_flags&UL_CHIPT_RQIRQ)&&!irq) continue;
    if((ret=(*chip_init)(udrv, port, irq, baud, chip_options))>=0){
      UL_PRINTF(KERN_INFO "uLan ul_drv_new : %s found at 0x%lx\n",
   		 ul_chip_types[chipidx].chip_name,udrv->port);
      break;
    }
  }while(chipprobe&&ul_chip_types[++chipidx].chip_name);
  if(ret>=0)
  { /* setups buffers, ports and irq for sucesfully detected device */
    if(ul_drv_new_start(udrv,ulbuffer)>=0)
      return udrv;
  }
  UL_PRINTF(KERN_ERR "uLan ul_drv_new : failed to init\n");
  FREE(udrv);
  return NULL;
};

/* destroy driver */
void ul_drv_free(ul_drv *udrv)
{
  if(!udrv) return;
  if(uld_test_dfl(udrv,CHIPOK))
  { if(udrv->fnc_cctrl(udrv,UL_CC_FREEIRQ,0)==UL_RC_ENOFNC)
      if(udrv->irq)
        free_irq(udrv->irq,udrv); 
  }
  if(udrv->irq>0) ul_synchronize_irq(udrv->irq);
  KILL_UDRV_BH(udrv);
  DONE_UDRV_WDTIM(udrv);
  schedule();
  if(uld_test_dfl(udrv,CHIPOK))
  { uld_clear_dfl(udrv,CHIPOK);
    udrv->fnc_cctrl(udrv,UL_CC_PDONE,0);
  };
  if(udrv->irq>0) ul_synchronize_irq(udrv->irq);
  KILL_UDRV_BH(udrv);
  { /* first not complete sanity check of client list */
    ul_opdata *opptr;
    UL_DRV_LOCK_FINI
    UL_DRV_LOCK;
    while((opptr=udrv->operators)!=NULL){
      udrv->operators=opptr->opnext; 
      if(udrv->operators!=NULL) udrv->operators->opprew=NULL;
      opptr->udrv=NULL;
     #ifdef FOR_LINUX_KERNEL
      wake_up_interruptible(&opptr->wqrec);
     #endif /*FOR_LINUX_KERNEL*/
    }
    UL_DRV_UNLOCK;
  };
  ul_mem_done(udrv);
  FREE(udrv);
};

#endif /* _WIN32 */

/*******************************************************************/
/* Some internal test routines */

#undef UL_INT_TST

#ifdef UL_INT_TST

/* standard command codes 
   00H .. 3FH    store to buffer 
   40H .. 7FH    store to buffer without ACK
   80H .. 9FH    proccess at onece
   A0H .. BFH    process with additional receive
   C0H .. FFH    process with additional send
*/

#define UL_CMD_RES	0x80	/* Reinitialize RS485 */
#define UL_CMD_SFT	0x81	/* Test free space in input buffer */
#define UL_CMD_SID	0xF0	/* Send identification */
#define UL_CMD_SFI	0xF1	/* Send amount of free space in IB */
#define UL_CMD_TF0	0x98	/* End of stepping */
#define UL_CMD_TF1	0x99	/* Begin of stepping */
#define UL_CMD_STP	0x9A	/* Do step */
#define UL_CMD_DEB	0x9B	/* Additional debug commands */
#define UL_CMD_SPC	0xDA	/* Send state - for 8051 PCL PCH PSW ACC */

#define UL_CMD_RDM	0xF8	/* Read memory   T T B B L L */
#define UL_CMD_WRM	0xB8	/* Write mamory  T T B B L L */

void genrdmemblk(ul_drv *udrv, uchar dadr, int mtyp, int mbeg, int mlen)
{
  ul_mem_blk *blk, *beg_blk;
  ul_data_it di;
  blk=ul_new_frame_head(udrv, 3, 2, UL_CMD_RDM,
       UL_BFL_M2IN|UL_BFL_TAIL|UL_BFL_SND|
       UL_BFL_PRQ|UL_BFL_ARQ);
  if(blk==NULL) return;
  beg_blk=blk;
  ul_di_init(&di,blk);
  /* ul_di_write1(&di,0); */
  ul_di_write1(&di,(uchar)mtyp);
  ul_di_write1(&di,(uchar)(mtyp>>8));
  ul_di_write1(&di,(uchar)mbeg);
  ul_di_write1(&di,(uchar)(mbeg>>8));
  ul_di_write1(&di,(uchar)mlen);
  ul_di_write1(&di,(uchar)(mlen>>8));
  UL_BLK_HEAD(blk).len=di.pos;

  blk=ul_new_frame_head(udrv, 0, 0, 0,
       UL_BFL_M2IN|UL_BFL_REC|
       UL_BFL_PRQ|UL_BFL_ARQ /*|UL_BFL_LNMM*/);
  if(blk==NULL) { ul_free_blk(udrv,beg_blk); return;};
  /* UL_BLK_HEAD(blk).len=mlen; */
  ul_tail_frame_head(beg_blk,blk);
  ul_bll_move_mes(&udrv->prep_bll,beg_blk);
};

void geninfoblk(ul_drv *udrv)
{
  static int seq_num=0;
  ul_mem_blk *blk, *beg_blk;
  ul_data_it di;
  blk=ul_new_frame_head(udrv, 3, 2, UL_CMD_SID,
       UL_BFL_M2IN|UL_BFL_TAIL|UL_BFL_SND|
       UL_BFL_PRQ|UL_BFL_ARQ);
  if(blk==NULL) return;
  beg_blk=blk;
  ul_di_init(&di,blk);
  /* ul_di_write1(&di,0); */
  ul_di_write1(&di,1);
  ul_di_write1(&di,2);
  ul_di_write1(&di,3);
  ul_di_write1(&di,(uchar)(seq_num++));
  UL_BLK_HEAD(blk).len=di.pos;

  blk=ul_new_frame_head(udrv, 0, 0, 0,
       UL_BFL_M2IN|UL_BFL_REC|
       UL_BFL_PRQ|UL_BFL_ARQ);
  if(blk==NULL) { ul_free_blk(udrv,beg_blk); return;};
  ul_tail_frame_head(beg_blk,blk);
  ul_bll_move_mes(&udrv->prep_bll,beg_blk);
};

void gentestblk(ul_drv *udrv)
{
  static int seq_num=0;
  ul_mem_blk *blk;
  ul_data_it di;
  blk=ul_new_frame_head(udrv, 3, 2, 0x23,
       UL_BFL_M2IN|UL_BFL_SND|UL_BFL_ARQ);
  if(blk==NULL) return;
  ul_di_init(&di,blk);
  /* ul_di_write1(&di,0); */
  ul_di_write1(&di,0x11);
  ul_di_write1(&di,0x12);
  ul_di_write1(&di,0x13);
  ul_di_write1(&di,(uchar)(seq_num++));
  UL_BLK_HEAD(blk).len=di.pos;

  ul_bll_move_mes(&udrv->prep_bll,blk);
};

#endif /* UL_INT_TST */
	
void printblk(ul_mem_blk *blk)
{
  ul_mem_blk *prev_blk=NULL;
  while(blk)
  {
    if(prev_blk) if(UL_BLK_HEAD(blk).prev!=prev_blk)
       UL_PRINTF(KERN_CRIT "Inconsistency in BLL list !!!\n");

    UL_PRINTF("BLK : dadr=%02X sadr=%02X cmd=%02X flg=%04X retry=%02X len=%04X\n",
     UL_BLK_HEAD(blk).dadr,
     UL_BLK_HEAD(blk).sadr,
     UL_BLK_HEAD(blk).cmd,
     UL_BLK_HEAD(blk).flg,
     UL_BLK_HEAD(blk).retry_cnt,
     UL_BLK_HEAD(blk).len
    );
    {
      ul_data_it di;
      uchar ch;
      int cnt=128;
      UL_PRINTF("      |");
      ul_di_init(&di,blk);
      di.trans_len=UL_BLK_HEAD(blk).len;
      while((di.pos<di.trans_len)&& --cnt)
      {
	ch=ul_di_read1(&di);
	if(ch>=0x20&&ch<0x7F) UL_PRINTF("%c",ch);
	else UL_PRINTF("<%X>",ch);
      };
      if(!cnt) UL_PRINTF("<...>...");
      UL_PRINTF("|\n");
    };
    prev_blk=blk;
    blk=UL_BLK_HEAD(blk).next;
  };
};

void printudrvbll(ul_drv *udrv)
{
  UL_PRINTF("\nprep_bll : frames = %d\n",udrv->prep_bll.cnt);
  printblk(udrv->prep_bll.first);

  UL_PRINTF("\nwork_bll : frames = %d\n",udrv->work_bll.cnt);
  printblk(udrv->work_bll.first);

  UL_PRINTF("\nproc_bll : frames = %d\n",udrv->proc_bll.cnt);
  printblk(udrv->proc_bll.first);
  UL_PRINTF("\nopan_bll : frames = %d\n",udrv->opan_bll.cnt);
  printblk(udrv->opan_bll.first);
  UL_PRINTF("\nfree_blk = %d",udrv->free_blk_cnt);
  UL_PRINTF("\n");
 #ifdef ENABLE_UL_MEM_CHECK
  UL_PRINTF("MEM_CHECK: malloc-free=%d\n",
		 (int)atomic_read(&ul_mem_check_counter));
 #endif /* ENABLE_UL_MEM_CHECK */
};

void printudrvoperators(ul_drv *udrv)
{
  UL_DRV_LOCK_FINI
  ul_opdata *opdata;
  ul_opchain *opmember,**opchain;
  int reccnt,filtcnt,opcnt;
  UL_PRINTF("\nul_drv operators (open handles)\n");
  UL_DRV_LOCK;
  opcnt=0;
  for(opdata=udrv->operators;opdata;opdata=opdata->opnext)
  { opchain=&opdata->recchain;reccnt=0;
    if((opmember=*opchain)!=NULL)do{
      reccnt++; if(reccnt>1000) break;
    }while((opmember=opmember->next)!=*opchain);
    opchain=&opdata->filtchain;filtcnt=0;
    if((opmember=*opchain)!=NULL)do{
      filtcnt++; if(filtcnt>1000) break;
    }while((opmember=opmember->next)!=*opchain);
    UL_PRINTF("  operator %d, in recchain %d, in filtchain %d\n",
		    opcnt,reccnt,filtcnt);
    opcnt++;if(opcnt>100) break;
  }
  UL_DRV_UNLOCK;
}

void printudrvfncstack(ul_drv *udrv)
{
  ul_call_fnc **p;
  UL_PRINTF("ul_drv: fnc_act:%08lx fnc_stack:",(long)udrv->fnc_act);
  p=udrv->fnc_sp;
  while(p!=&udrv->fnc_stack[0]){
    p--;
    UL_PRINTF("%08lx ",(long)*p);
  };
  UL_PRINTF("\n");
}

/*******************************************************************/
/* Linux kernel and module specific part */ 

#ifdef FOR_LINUX_KERNEL

MODULE_SUPPORTED_DEVICE("ul_drv");
MODULE_AUTHOR("Pavel Pisa <pisa@cmp.felk.cvut.cz>");
MODULE_DESCRIPTION("\
uLan communication module ( 9 bit RS-485 protocol )\n\
for communication with embedded controllers.\n\
It uses i82510 addon card or RS-232 dongle");
MODULE_LICENSE("GPL");
EXPORT_NO_SYMBOLS;

static int ulan_major_dev=ULAN_MAJOR;	/* uLan major device number */

#define UL_MINORS 9
ul_drv *ul_drv_arr[UL_MINORS]={[0 ... UL_MINORS-1]=NULL};
char *chip[UL_MINORS]={NULL,};
#if defined(UL_WITH_PCI) || defined(UL_WITH_USB)
char *slot[UL_MINORS]={NULL,};
#endif
int my_adr[UL_MINORS]={[0 ... UL_MINORS-1]=2};
int baud[UL_MINORS]={[0 ... UL_MINORS-1]=19200};
int irq[UL_MINORS]={0,};
int port[UL_MINORS]={0x3e8 
	#if (UL_MINORS>1)
		,[1 ... UL_MINORS-1]=0
	#endif
		};
int debug=-1;		/* debug flags */

MODULE_PARM_DESC(chip,"chip type: auto, 82510, 16450, 16950-pci");
MODULE_PARM(chip,"1-" __MODULE_STRING(UL_MINORS) "s");
MODULE_PARM_DESC(my_adr,"uLan address of defined interface");
MODULE_PARM(my_adr,"1-" __MODULE_STRING(UL_MINORS) "i");
MODULE_PARM_DESC(baud,"define baudrates");
MODULE_PARM(baud,"1-" __MODULE_STRING(UL_MINORS) "i");
MODULE_PARM_DESC(irq,"define irqs");
MODULE_PARM(irq,"1-" __MODULE_STRING(UL_MINORS) "i");
MODULE_PARM_DESC(port,"define ports");
MODULE_PARM(port,"1-" __MODULE_STRING(UL_MINORS) "i");
#if defined(UL_WITH_PCI) || defined(UL_WITH_USB)
MODULE_PARM_DESC(slot,"slot type (pci) and specification");
MODULE_PARM(slot,"1-" __MODULE_STRING(UL_MINORS) "s");
#endif
MODULE_PARM_DESC(debug,"debug flags (FATAL,CHIO,IRQ,MSG,FAILS,"\
		       "SEQ,PORTS,FILE,FILT)");
MODULE_PARM(debug,"i");
MODULE_PARM_DESC(ulbuffer,"size of per interface buffer");
MODULE_PARM(ulbuffer,"i");

static int ulan_open(struct inode *inode, struct file *file);
static CLOSERET ulan_close(struct inode *inode, struct file *file);
static int ulan_ioctl(struct inode *inode, struct file *file, 
			unsigned int cmd, unsigned long arg);
static RWRET ulan_write(WRITE_PARAMETERS);
static RWRET ulan_read(READ_PARAMETERS);
static loff_t ulan_lseek(LSEEK_PARAMETERS);
static unsigned int ulan_poll(struct file *, poll_table *);

/* static struct file_operations ulan_fops */
KC_CHRDEV_FOPS_BEG(ulan_fops)
	KC_FOPS_LSEEK(ulan_lseek)/* lseek              */
	read:ulan_read,		/* read                */
	write:ulan_write,	/* write               */
	readdir:NULL,		/* readdir             */
	poll:ulan_poll,		/* poll                */
	ioctl:ulan_ioctl,	/* ioctl               */
	mmap:NULL,		/* mmap                */
	open:ulan_open,		/* open                */
	KC_FOPS_FLUSH(NULL)	/* flush               */
	KC_FOPS_RELEASE(ulan_close)/* close/release    */
	fsync:NULL,		/* fsync               */
	fasync:NULL,		/* fasync              */
	NULL,			/* check_media_change  */
	NULL,			/* revalidate          */
KC_CHRDEV_FOPS_END;

#ifdef ENABLE_UL_MEM_CHECK
void * ul_mem_check_malloc(size_t size)
{void *ptr;
 ptr=kmalloc(size,GFP_KERNEL);
 if(ptr) atomic_inc(&ul_mem_check_counter);
 return ptr;
}

void ul_mem_check_free(void *ptr)
{
  if(ptr) atomic_dec(&ul_mem_check_counter);
  kfree(ptr);
}
#endif /* ENABLE_UL_MEM_CHECK */
#endif /* FOR_LINUX_KERNEL */

static void add_to_opchain(ul_opchain **chain,ul_opchain *member)
{
  UL_DRV_LOCK_FINI
  UL_DRV_LOCK;
  if(!*chain) *chain=member->next=member->prev=member;
  else {
    (volatile ul_opchain*)member->next=*chain;
    (volatile ul_opchain*)member->prev=(*chain)->prev;
    UL_MB();
    (volatile ul_opchain*)(*chain)->prev->next=member;
    (volatile ul_opchain*)(*chain)->prev=member;
  };
  UL_DRV_UNLOCK;
};

static void add_to_opchain_front(ul_opchain **chain,ul_opchain *member)
{
  UL_DRV_LOCK_FINI
  UL_DRV_LOCK;
  if(!*chain) *chain=member->next=member->prev=member;
  else {
    (volatile ul_opchain*)member->next=*chain;
    (volatile ul_opchain*)member->prev=(*chain)->prev;
    UL_MB();
    (volatile ul_opchain*)(*chain)->prev->next=member;
    (volatile ul_opchain*)(*chain)->prev=member;
    *(volatile ul_opchain**)chain=member;
  };
  UL_DRV_UNLOCK;
};

static void del_from_opchain(ul_opchain **chain,ul_opchain *member)
{
  UL_DRV_LOCK_FINI
  UL_DRV_LOCK;
  member->next->prev=member->prev;
  member->prev->next=member->next;
  if(*chain==member) 
  { if(member->next==member) *chain=NULL;
    else *chain=member->next;
  };
  UL_DRV_UNLOCK;
};

/* free message form ulan operator */
static int ulan_freemsg(ul_opdata* opdata)
{
  int stamp;
  ul_mem_blk *mes;
  ul_opchain *opmember;

  if(!opdata->udrv) return -ENODEV;
  if(!(mes=opdata->message)) return -ENOMSG;

  if(!UL_BLK_HEAD(mes).stamp)
  {
    opdata->message=NULL;
    stamp=ul_gen_stamp();
    UL_BLK_HEAD(mes).stamp=stamp;
    if(UL_BLK_HEAD(mes).flg&UL_BFL_M2IN)
    {
      opmember=(ul_opchain*)MALLOC(sizeof(ul_opchain));
      if(opmember)
      {
	memset(opmember,0,sizeof(ul_opchain));
	opmember->stamp=stamp;
	opmember->state=UL_OPST_ONCE;
	add_to_opchain_front(&opdata->filtchain,opmember);
      };
    };
    ul_bll_move_mes(&opdata->udrv->prep_bll,mes);
    ulan_stroke(opdata->udrv);
    return stamp;
  }else{
    ul_dec_ref_cnt(mes);
    opdata->message=NULL;
    return 0;
  };
};

/* abort preparation of message */
static int ulan_abortmsg(ul_opdata* opdata)
{
  ul_mem_blk *mes;

  if(!opdata->udrv) return -ENODEV;
  if(!(mes=opdata->message)) return -ENOMSG;
  opdata->message=NULL;

  if(!UL_BLK_HEAD(mes).stamp)
  {
    ul_free_mes(opdata->udrv,mes);
    return 1;
  }else{
    ul_dec_ref_cnt(mes);
    return 0;
  };
};

/* rewind of message */
static int ulan_rewmsg(ul_opdata* opdata)
{
  ul_mem_blk *mes;
  if(!opdata->udrv) return -ENODEV;
  if(!(mes=opdata->message)) return -ENOMSG;
  opdata->message=NULL;
  ul_di_init(&opdata->data,opdata->message);
  return 0;
};

/* create new outgoing message */
static int ulan_newmsg(ul_opdata* opdata,const ul_msginfo *msginfo)
{
  int flg=msginfo->flg;
  flg &= ~(UL_BFL_LOCK | UL_BFL_FAIL | UL_BFL_TAIL | UL_BFL_REC);
  flg |= UL_BFL_SND;
  if(!opdata->udrv) return -ENODEV;
  if (opdata->message) ulan_freemsg(opdata);
  opdata->message=ul_new_frame_head(opdata->udrv, msginfo->dadr,
                     opdata->udrv->my_adr, msginfo->cmd,flg);
  if(!opdata->message) return -ENOMEM;
  memset(UL_BLK_FDATA(opdata->message),0,UL_BLK_FSIZE);
  ul_di_init(&opdata->data,opdata->message);
  return 0;
}

static int ulan_tailmsg(ul_opdata* opdata,const ul_msginfo *msginfo)
{
  ul_mem_blk *mes;
  ul_mem_blk *tail;
  int flg=msginfo->flg;
  if(!opdata->udrv) return -ENODEV;
  if(!(mes=opdata->message)) return -ENOMSG;
  if(UL_BLK_HEAD(mes).stamp) return -EPERM;

  flg &= ~(UL_BFL_LOCK | UL_BFL_FAIL | UL_BFL_TAIL);
  if(!flg&UL_BFL_REC) flg|=UL_BFL_SND;
  tail=ul_new_frame_head(opdata->udrv, msginfo->dadr,
	opdata->udrv->my_adr, msginfo->cmd,flg);
  if(!tail) return -ENOMEM;
  memset(UL_BLK_FDATA(tail),0,UL_BLK_FSIZE);
  UL_BLK_HEAD(tail).len=msginfo->len;
  ul_tail_frame_head(mes,tail);
  ul_di_init(&opdata->data,tail);
  return 0;
}

static int ulan_acceptmsg(ul_opdata* opdata,ul_msginfo *msginfo)
{
  ul_mem_blk *mes;
  ul_opchain *opmember;

  if(!opdata->udrv) return -ENODEV;
  if (opdata->message) ulan_freemsg(opdata);
  memset(msginfo,0,sizeof(ul_msginfo));
  if(!(opmember=opdata->recchain)) return -ENOMSG;
  del_from_opchain(&opdata->recchain,opmember);
  mes=opmember->message;
  if(opmember->state==UL_OPST_MES) FREE(opmember);
  opdata->message=mes;
  ul_di_init(&opdata->data,mes);
  msginfo->flg=UL_BLK_HEAD(mes).flg;
  msginfo->dadr=UL_BLK_HEAD(mes).dadr;
  msginfo->sadr=UL_BLK_HEAD(mes).sadr;
  msginfo->cmd=UL_BLK_HEAD(mes).cmd;
  msginfo->len=UL_BLK_HEAD(mes).len;
  msginfo->stamp=UL_BLK_HEAD(mes).stamp;
  return 0;
}

static int ulan_actailmsg(ul_opdata* opdata,ul_msginfo *msginfo)
{
  ul_mem_blk *mes;
  ul_mem_blk *tail;

  if(!opdata->udrv) return -ENODEV;
  memset(msginfo,0,sizeof(ul_msginfo));
  if(!(mes=opdata->message)) return -ENOMSG;
  if(!UL_BLK_HEAD(mes).stamp) return -EPERM;
  tail=UL_BLK_HEAD(opdata->data.head_blk).next;
  if(!tail) return -ENOMSG;
  ul_di_init(&opdata->data,tail);
  msginfo->flg=UL_BLK_HEAD(tail).flg;
  msginfo->dadr=UL_BLK_HEAD(tail).dadr;
  msginfo->sadr=UL_BLK_HEAD(tail).sadr;
  msginfo->cmd=UL_BLK_HEAD(tail).cmd;
  msginfo->len=UL_BLK_HEAD(tail).len;
  msginfo->stamp=UL_BLK_HEAD(tail).stamp;
  return 0;
}

/* create new outgoing message */
static int ulan_addfilt(ul_opdata* opdata,const ul_msginfo *msginfo)
{
  ul_drv *udrv;
  ul_opchain *opmember;
  int filtonce;
  if(!opdata->udrv) return -ENODEV;
  opmember=(ul_opchain*)MALLOC(sizeof(ul_opchain));
  udrv=opdata->udrv;
  if(!opmember) return -ENOMEM;
  memset(opmember,0,sizeof(ul_opchain));
  opmember->dadr=msginfo->dadr;
  opmember->sadr=msginfo->sadr;
  opmember->cmd=msginfo->cmd;
  filtonce=msginfo->flg&UL_BFL_NORE;
  opmember->state=filtonce?UL_OPST_FILTNEW:UL_OPST_FILT;
  add_to_opchain(&opdata->filtchain,opmember);
  if(filtonce){
    uld_set_dfl(udrv,CHECK_FILT);
    SCHEDULE_BH(udrv);
  }
  LOG_FILTS("ulan_addfilt : filter %d %d %d %s\n",
	 opmember->dadr,opmember->sadr,opmember->cmd,filtonce?"once":"");
  return 0;
}

static int ulan_inepoll(ul_opdata* opdata)
{
  return opdata->recchain!=NULL?1:0;
}

static int ulan_hwtest(ul_opdata* opdata,unsigned long fnc)
{
  int ret;
  ul_drv *udrv;
  UL_DRV_LOCK_FINI
  if(!opdata->udrv) return -ENODEV;
  udrv=opdata->udrv;
  UL_DRV_LOCK;
  ret=udrv->fnc_cctrl(udrv,UL_CC_HWTEST,fnc);
  if(ret<0) ret=-EINVAL;
  UL_DRV_UNLOCK;
  return ret;
}

#ifdef FOR_LINUX_KERNEL

/* open ulan device */
static int ulan_open(struct inode *inode, struct file *file)
{ 
  ul_drv *udrv;
  ul_opdata *opdata;
  unsigned int minor = kc_dev2minor(inode->i_rdev);
  #ifdef UL_WITH_DEVFS
  if(minor==-1){
    udrv=file->private_data;
  }else
  #endif /* UL_WITH_DEVFS */
  {
    if ((minor >= UL_MINORS)||(minor<0)) return -ENODEV;
    udrv=ul_drv_arr[minor];
  }
  if (!udrv) return -ENODEV;
  if(udrv->magic!=UL_DRV_MAGIC)
  {
    printk(KERN_CRIT "ulan_open: Wrong uLan MAGIC number!!!\n");
    return -ENODEV;
  }

  kc_MOD_INC_USE_COUNT;
  if(!(file->private_data=MALLOC(sizeof(ul_opdata))))
  { kc_MOD_DEC_USE_COUNT;
    return(-ENOMEM);
  };
  opdata=(ul_opdata*)file->private_data;
  opdata->file=file;
  opdata->magic=ULOP_MAGIC;
  opdata->message=NULL;
  opdata->udrv=udrv;
  init_waitqueue_head(&opdata->wqrec);
  opdata->opprew=NULL;
  opdata->opnext=NULL;
  opdata->recchain=NULL;
  opdata->filtchain=NULL;
  { /* add us onto list of clients of udrv */
    ul_opdata *opptr;
    UL_DRV_LOCK_FINI
    UL_DRV_LOCK;
    opptr=udrv->operators;
    if(opptr) {opptr->opprew=opdata;opdata->opnext=opptr;};
    UL_MB();
    udrv->operators=opdata;
    UL_DRV_UNLOCK;
  };

  return 0;
};

static CLOSERET ulan_close(struct inode *inode, struct file *file)
{
  ul_opchain *opmember;
  ul_opdata *opdata=(ul_opdata*)file->private_data;
  if(opdata->magic!=ULOP_MAGIC) panic("ulan_close : BAD opdata magic !!!");

  if (opdata->message) ulan_freemsg(opdata);
  { /* delete us from list of clients of udrv */
    ul_drv *udrv;
    ul_opdata *opptr;
    UL_DRV_LOCK_FINI
    UL_DRV_LOCK;
    if((udrv=opdata->udrv)!=NULL){
      if((opptr=opdata->opnext)) opptr->opprew=opdata->opprew;
      if((opptr=opdata->opprew)) opptr->opnext=opdata->opnext;
      else udrv->operators=opdata->opnext;
    }
    UL_DRV_UNLOCK;
    if(udrv)
      if(udrv->irq>0) ul_synchronize_irq(udrv->irq);
    while((udrv=opdata->udrv)!=NULL){
      if(!uld_atomic_test_dfl(udrv,IN_BOTTOM))
        break;
      kc_yield();
    }
  }
  schedule();
  while((opmember=opdata->recchain))
  {
    del_from_opchain(&opdata->recchain,opmember);
    if(opmember->message) ul_dec_ref_cnt(opmember->message);
    FREE(opmember);
  };
  while((opmember=opdata->filtchain))
  {
    del_from_opchain(&opdata->filtchain,opmember);
    FREE(opmember);
  };
  FREE(file->private_data);
  kc_MOD_DEC_USE_COUNT;
  return (CLOSERET)0;
};

static int ulan_ioctl(struct inode *inode, struct file *file, 
			unsigned int cmd, unsigned long arg)
{
  ul_msginfo msginfo;
  int ret;
  ul_opdata *opdata=(ul_opdata*)file->private_data;
  if(opdata->magic!=ULOP_MAGIC) panic("ulan_ioctl : BAD opdata magic !!!");

  switch (cmd) {
   case UL_DRV_VER :
	return UL_DRV_VERCODE;
   case UL_NEWMSG :
	if(kc_copy_from_user(&msginfo,(void *)arg,sizeof(msginfo)))
	   return -EFAULT;
  	return ulan_newmsg(opdata,&msginfo);
   case UL_TAILMSG :
	if(kc_copy_from_user(&msginfo,(void *)arg,sizeof(msginfo)))
	   return -EFAULT;
  	return ulan_tailmsg(opdata,&msginfo);
   case UL_FREEMSG :
  	return ulan_freemsg(opdata);
   case UL_ACCEPTMSG :  	
	if(kc_copy_to_user((void *)arg,&msginfo,sizeof(msginfo)))
	   return -EFAULT;
	ret=ulan_acceptmsg(opdata,&msginfo);
	kc_copy_to_user((void *)arg,&msginfo,sizeof(msginfo));
  	return  ret;
   case UL_ACTAILMSG :  	
	if(kc_copy_to_user((void *)arg,&msginfo,sizeof(msginfo)))
	   return -EFAULT;
	ret=ulan_actailmsg(opdata,&msginfo);
	kc_copy_to_user((void *)arg,&msginfo,sizeof(msginfo));
  	return  ret;
   case UL_ADDFILT :
	if(kc_copy_from_user(&msginfo,(void *)arg,sizeof(msginfo)))
	   return -EFAULT;
  	return ulan_addfilt(opdata,&msginfo);
   case UL_ABORTMSG :
  	return ulan_abortmsg(opdata);
   case UL_REWMSG :
  	return ulan_rewmsg(opdata);
   case UL_KLOGBLL :
        if(!opdata->udrv) return -ENODEV;
	printudrvbll(opdata->udrv);
    #ifdef ENABLE_UL_MEM_CHECK
	printudrvoperators(opdata->udrv);
    #endif /* ENABLE_UL_MEM_CHECK */
	return 0;
   case UL_STROKE :
        if(!opdata->udrv) return -ENODEV;
	ulan_stroke(opdata->udrv);
	return 0;
   case UL_DEBFLG :
	uld_debug_flg=arg;
	return 0;
   case UL_INEPOLL :
        return ulan_inepoll(opdata);
   case UL_HWTEST :
        if(current->euid!=0) return -EPERM;
	return ulan_hwtest(opdata,arg);
  };


  return -EINVAL;
};

/* common parameters (struct file *file, const char *buf, 
                      size_t count [, loff_t *ppos]) */
static RWRET ulan_write(WRITE_PARAMETERS)
{
  int cn;
  int len;
  ul_mem_blk *blk;
  ul_opdata *opdata=(ul_opdata*)file->private_data;
  if(opdata->magic!=ULOP_MAGIC) panic("ulan_write : BAD opdata magic !!!");

  if(!opdata->udrv) return -ENODEV;
  if(!opdata->message) return -ENOMSG;

  cn=count;
  while(cn>0)
  {
    while(!ul_di_adjust(&opdata->data))
    {
      if(!(blk=ul_alloc_blk(opdata->udrv))) 
	{count-=cn;cn=0;break;};
      memset(UL_BLK_NDATA(blk),0,UL_BLK_SIZE);
      opdata->data.blk->next=blk;
    };
    len=ul_di_atonce(&opdata->data);
    if(len>cn) len=cn;
    kc_copy_from_user(ul_di_byte(&opdata->data),buf,len);
    ul_di_add(&opdata->data,len);
    buf+=len;
    cn-=len;
  };

  if(opdata->data.pos>UL_BLK_HEAD(opdata->data.head_blk).len)
    UL_BLK_HEAD(opdata->data.head_blk).len=opdata->data.pos;

  return count-cn;
};

/* common parameters (struct file *file, char *buf, 
                      size_t count [, loff_t *ppos]) */
static RWRET ulan_read(READ_PARAMETERS)
{
  int cn;
  int len;
  ul_opdata *opdata=(ul_opdata*)file->private_data;
  if(opdata->magic!=ULOP_MAGIC) panic("ulan_read : BAD opdata magic !!!");

  if(!opdata->udrv) return -ENODEV;
  if(!opdata->message) return -ENOMSG;

  if(opdata->data.pos+count>UL_BLK_HEAD(opdata->data.head_blk).len)
   count=UL_BLK_HEAD(opdata->data.head_blk).len-opdata->data.pos;

  cn=count;
  while(cn>0)
  {
    if(!ul_di_adjust(&opdata->data))
    {
      while(cn--) kc_put_user(0,buf++);
      break;
    };
    len=ul_di_atonce(&opdata->data);
    if(len>cn) len=cn;
    kc_copy_to_user(buf,ul_di_byte(&opdata->data),len);
    ul_di_add(&opdata->data,len);
    buf+=len;
    cn-=len;
  };

  return count-cn;
};

/* common parameters (struct file *file, loff_t pos, int whence) */
static loff_t ulan_lseek(LSEEK_PARAMETERS)
{
  ul_opdata *opdata=(ul_opdata*)file->private_data;
  if(opdata->magic!=ULOP_MAGIC) panic("ulan_lseek : BAD opdata magic !!!");

  if(!opdata->message) return -ENOMSG;

  return ul_di_seek(&opdata->data,pos,whence);
};

static unsigned int ulan_poll(struct file *file, poll_table *wait)
{
  /* POLLIN | POLLOUT | POLLERR | POLLHUP | POLLRDNORM | POLLWRNORM */
  unsigned int mask = 0;
  ul_opdata *opdata=(ul_opdata*)file->private_data;
  if(opdata->magic!=ULOP_MAGIC) panic("ulan_poll : BAD opdata magic !!!");

  kc_poll_wait(file, &opdata->wqrec, wait);

  if (opdata->recchain)
    mask |= (POLLIN | POLLRDNORM);

  if (!opdata->udrv)
    mask |= (POLLIN | POLLOUT | POLLERR);

  return mask;
};

#endif /* FOR_LINUX_KERNEL */


/* this function is responsible to inform all ulan client operators
   about message of waited type -- !!!!! it is not SMP clean still */
static int ulan_proc_arrived(ul_drv *udrv, ul_mem_blk *message)
{
  ul_opdata *opdata;
  ul_opchain **opchain;
  ul_opchain *opmember;
  ul_opchain *opmember2;
  UL_DRV_LOCK_FINI

  ul_inc_ref_cnt(message);
  UL_DRV_LOCK;
  opdata=udrv->operators;
  UL_DRV_UNLOCK;
  while(opdata)
  {
    opchain=&opdata->filtchain;
    opmember=*opchain;
    if(opmember) do
    {
      if((!opmember->stamp||opmember->stamp==UL_BLK_HEAD(message).stamp)&&
	 (!opmember->cmd||opmember->cmd==UL_BLK_HEAD(message).cmd)&&
	 (!opmember->dadr||opmember->dadr==UL_BLK_HEAD(message).dadr)&&
	 (!opmember->sadr||opmember->sadr==UL_BLK_HEAD(message).sadr))
      {
	if(opmember->state==UL_OPST_ONCE)
	{
	  del_from_opchain(opchain,opmember);
	  opmember->state=UL_OPST_MES;
	  ul_inc_ref_cnt(message);
	  opmember->message=message;
	  add_to_opchain(&opdata->recchain,opmember);
	}else if(opmember->state==UL_OPST_FILT){
	 #ifdef FOR_LINUX_KERNEL
	  #ifdef ENABLE_UL_MEM_CHECK
	  atomic_inc(&ul_mem_check_counter);
	  #endif
	  opmember2=(ul_opchain*)kmalloc(sizeof(ul_opchain),GFP_ATOMIC);
	 #else /* FOR_LINUX_KERNEL */
	  opmember2=(ul_opchain*)MALLOC(sizeof(ul_opchain));
	 #endif /* FOR_LINUX_KERNEL */
	  if(opmember2)
	  { 
	    memset(opmember2,0,sizeof(ul_opchain));
	    opmember2->state=UL_OPST_MES;
	    ul_inc_ref_cnt(message);
	    opmember2->message=message;
	    add_to_opchain(&opdata->recchain,opmember2);
	  } else  UL_PRINTF(KERN_CRIT "ulan_proc_arrived : no kernel memory : FIXME handling\n");
	};
       #ifdef FOR_LINUX_KERNEL
	wake_up_interruptible(&opdata->wqrec);
       #elif defined(_WIN32)
	    wake_and_xch_pending_irp(&opdata->wait_irp,NULL,STATUS_SUCCESS);
       #endif /* FOR_LINUX_KERNEL */
	break;
      };
    } while((opmember=opmember->next)!=*opchain);
    opdata=opdata->opnext;
  };

  /* if no one interested in message, it is discarded */
  return  ul_dec_ref_cnt(message);
};

/* check for UL_OPST_FILTNEW and change them to single instance
   of UL_OPST_FILT */
void check_for_filtnew(ul_drv *udrv)
{
  ul_opdata *opdata;
  ul_opchain **opchain;
  ul_opchain *opmember;
  ul_opchain *opmember2;
  ul_opchain *opnext;
  UL_DRV_LOCK_FINI

  LOG_FILTS("check_for_filtnew : called\n");
  UL_DRV_LOCK;
  opdata=udrv->operators;
  UL_DRV_UNLOCK;
  for(;opdata;opdata=opdata->opnext)
  {
    opchain=&opdata->filtchain;
    opmember=*opchain;
    if(opmember) do
    { opnext=opmember->next;
      if(opnext==*opchain) opnext=NULL;
      if(opmember->state==UL_OPST_FILTNEW){
	LOG_FILTS("check_for_filtnew : processing UL_OPST_FILTNEW\n");
	opmember2=*opchain;
	while(opmember2){
	  if((opmember->cmd==opmember2->cmd)&&
             (opmember->sadr==opmember2->sadr)&&
	     (opmember->dadr==opmember2->dadr)&&
	     (opmember->stamp==opmember2->stamp)&&
	     (opmember2->state==UL_OPST_FILT)) break;
	  opmember2=opmember2->next;
	  if(opmember2==*opchain){
	    opmember2=NULL;
	  }
        };
	if(opmember2){
	  LOG_FILTS("check_for_filtnew : deletting UL_OPST_FILTNEW\n");
	  del_from_opchain(opchain,opmember);
	  FREE(opmember);
	}else{
	  LOG_FILTS("check_for_filtnew : changing to UL_OPST_FILT\n");
	  opmember->state=UL_OPST_FILT;
	}
      }
    } while((opmember=opnext)!=NULL);
  };
};

#ifndef _WIN32

/* notice about new message in udrv->proc_bll */
void ulan_do_bh(ulan_do_bh_data_type data)
{
  ul_mem_blk *message;
  ul_drv *udrv=(ul_drv *)data;

  #ifdef ENABLE_UL_IRQ_STALE_WDG
  ul_irq_stale_wdg_cnt=0;
  #endif /* ENABLE_UL_IRQ_STALE_WDG */

  if(udrv->magic!=UL_DRV_MAGIC)
  {
    #ifdef FOR_LINUX_KERNEL
     panic("ulan_do_bh : BAD udrv magic !!!");
    #else /* FOR_LINUX_KERNEL */
     return;
    #endif /* FOR_LINUX_KERNEL */
  }

  LOG_MESSAGES("ulan_do_bh : announcing messages\n");
  while(udrv->proc_bll.first||uld_test_dfl(udrv,ASK_BOTTOM))
  {
    if(uld_test_and_set_dfl(udrv,IN_BOTTOM))
    { LOG_MESSAGES("ulan_do_bh : else is already in");
	  return;
    }
    uld_clear_dfl(udrv,ASK_BOTTOM);
    if(uld_test_dfl(udrv,CHECK_FILT)){
      uld_clear_dfl(udrv,CHECK_FILT);
      check_for_filtnew(udrv);
    }
    while((message=udrv->proc_bll.first))
    {
      int i;
      ul_bll_move_mes(&udrv->opan_bll,message);
      i=ulan_proc_arrived(udrv,message);
      LOG_MESSAGES("ulan_do_bh : message sent to %d recchains\n",i);
    };
    uld_atomic_clear_dfl(udrv,IN_BOTTOM);
  };
};

#endif /* _WIN32 */

#ifdef FOR_LINUX_KERNEL

/* config space access (byte|word|dword) return to text pcibios_strerror */
/* pci_read_config_byte(struct pci_dev *dev, int where, u8 *val) */
/* pci_write_config_byte(struct pci_dev *dev, int where, u8 *val) */
/* request_region(), request_mem_region() */
/* void *pci_get_drvdata (struct pci_dev *pdev) */
/* void pci_set_drvdata (struct pci_dev *pdev, void *data) */


#if defined(UL_WITH_PCI) || defined(UL_WITH_USB)

static int /*__devinit*/ 
 ulan_init_find_minor(char *abus, char *aslot, char *asubdev,
                      int *pminor, int *ppar)
{
  int i;
  int abus_len;
  int aslot_len;
  int sub_offs;
  int minor;
  int par;
  int match=0;
  char *islot;
  abus_len=strlen(abus);
  aslot_len=strlen(aslot);
  sub_offs=abus_len+aslot_len;
  *pminor=*ppar=minor=par=-1;
  for(i=0;i<UL_MINORS;i++){
    islot=slot[i];
    if(islot)
      if(!*islot) islot=NULL;
    switch (match){
      case 0: 
        if(!ul_drv_arr[i]&&!islot){
	  minor=i;match=1;
	  if(par==-1) par=i;
	}
      case 1:
        if(!islot) continue;
        if(!strcmp(islot,abus)){
	  par=i;
	  if(!ul_drv_arr[i]){ minor=i; match=2;}
	}
      case 2:
        if(!islot||!aslot) continue;
        if(!strncmp(islot,abus,abus_len)&&
	   !strcmp(islot+abus_len,aslot)){
	  par=i;
	  if(!ul_drv_arr[i]){ minor=i; match=3;}
        }
      case 3:
        if(!islot||!aslot||!asubdev) continue;
        if(!strncmp(islot,abus,abus_len)&&
	   !strncmp(islot+abus_len,aslot,aslot_len)&&
	   (islot[sub_offs]=='.')&&!strcmp(islot+sub_offs+1,asubdev)){
	  par=i;
	  if(!ul_drv_arr[i]){ minor=i; match=4;}
        }
    }
  }
  *pminor=minor;*ppar=par;
  return match;
}

#endif /*defined(UL_WITH_PCI)||defined(UL_WITH_USB)*/

/*******************************************************************/
/* PCI Devices Support */

#ifdef UL_WITH_PCI

/* default invocation of chip_init for PCI devices */
static int /*__devinit*/ 
  ulan_init_chan(struct pci_dev *dev,char *subdev, 
  		 char *chip_name, ul_drv **pudrv,
                 ul_chip_init_fnc *chip_init,
                 int port, int irq, int options)
{
  ul_drv *udrv;
  int amy_adr=0;
  int abaud=0;
  int match;
  int minor;
  int ret;
  int i;
#ifdef UL_WITH_DEVFS
  kc_devfs_handle_t devfs_handle;
  char dev_name[32];
#endif /* UL_WITH_DEVFS */
  
  *pudrv=NULL;
  /* try to find best minor and parameters */
  match=ulan_init_find_minor("pci",dev->slot_name,subdev,&minor,&i);
  if(i>=0){
    abaud=baud[i];amy_adr=my_adr[i];
  }

  /* mem for ul_drv */
  if(!(udrv=MALLOC(sizeof(ul_drv)))) return -ENOMEM;
  /* clear memory */
  memset(udrv,0,sizeof(ul_drv));
  /* set initial state */
  ul_drv_new_init_state(udrv, amy_adr);
  udrv->dev=dev;
  /* init chip driver */
  if((ret=(*chip_init)(udrv, port, irq, abaud, options))<0){
    printk(KERN_CRIT "ulan_init_chan: ERROR - chip_init returned %d\n",ret);
    FREE(udrv);
    return -EIO;
  }
  /* setups buffers, ports and irq for sucesfully detected device */
  if((ret=ul_drv_new_start(udrv,ulbuffer))<0){
    printk(KERN_CRIT "ulan_init_chan: ERROR - ul_drv_new_start returned %d\n",ret);
    FREE(udrv);
    return -EIO;
  }
  #ifdef UL_WITH_DEVFS
  sprintf (dev_name, "ulan%d", minor);
  devfs_handle=kc_devfs_new_cdev(NULL, MKDEV(ulan_major_dev, minor), 
			S_IFCHR | S_IRUSR | S_IWUSR | S_IRGRP | S_IWGRP, 
			&ulan_fops, udrv, dev_name);
  udrv->devfs_handle=devfs_handle;
  #endif /* UL_WITH_DEVFS */

  printk(KERN_INFO "ulan_init_chan: chip=%s minor=%d baud=%d my_adr=%d ready\n",
	 chip_name,/*dev_name*/minor,udrv->baud_val,udrv->my_adr);
  
  if(minor>=0) ul_drv_arr[minor]=udrv;
  *pudrv=udrv;
  udrv->next_chan=pci_get_drvdata(dev);
  pci_set_drvdata(dev, udrv);
  return 0;
}


static int /*__devinit*/ ulan_init_one_950pci (struct pci_dev *dev,
				   const struct pci_device_id *ent)
{
  int chip_options=ent->driver_data;
  int ret, retall=-ENODEV;
  char subdev[2];
  int subno, port, irq;
  ul_drv *udrv;

  if (pci_enable_device (dev)){
    printk(KERN_CRIT "ulan_init_one_950pci: Cannot enable device\n");
    return -EIO;    
  }
  port=pci_resource_start(dev,0);
  irq=dev->irq;
  if(!irq||!port||!(pci_resource_flags(dev,0)&IORESOURCE_IO)){
    LOG_FATAL(KERN_CRIT "uLan u950pci_init : bad PCI port or irq !\n");
    return -ENODEV;
  }
  
  for(subno=0;subno<2;subno++,port+=8){
    subdev[0]='0'+subno;subdev[1]=0;
    /* (dev,subdev,chip_name,pudrv,chip_init,port,irq,options) */
    ret=ulan_init_chan(dev,subdev,"16950-pci",&udrv,&u950pci_init,port,irq,chip_options);
    if(ret>=0){
      retall=0;
    } else {
      if(retall<0) retall=ret;
    }
  }
  return retall;
}

static int /*__devinit*/ ulan_init_one (struct pci_dev *dev,
				   const struct pci_device_id *ent)
{
  unsigned long driver_data=ent->driver_data;
  printk(KERN_INFO "ulan_init_one: PCI device found at slot : %s\n",
         dev->slot_name?dev->slot_name:"unknown");
  pci_set_drvdata(dev, NULL);
  switch(driver_data&~0xff){
    case 0x16954000 :
      return ulan_init_one_950pci(dev,ent);
  }
  printk(KERN_CRIT "ulan_init_one: No device of specified driver_data type\n");
  return -ENODEV;
}

static void /*__devexit*/ ulan_remove_one (struct pci_dev *dev)
{
  ul_drv *udrv, *next_udrv;
  int i;
  udrv=(ul_drv *)pci_get_drvdata(dev);
  if(!udrv){
    printk(KERN_CRIT "ulan_remove_one: no uLan drvdata\n");
    return;
  }
  for(;udrv;udrv=next_udrv){
    if(udrv->magic!=UL_DRV_MAGIC){
      printk(KERN_CRIT "ulan_remove_one: Wrong uLan MAGIC number!!!\n");
      return;
    }
    next_udrv=udrv->next_chan;
    if(dev!=udrv->dev){
      printk(KERN_CRIT "ulan_remove_one: BAD - cross PCI device remove\n");
    }
    for(i=0;i<UL_MINORS;i++){
      if (udrv==ul_drv_arr[i]) ul_drv_arr[i]=NULL;
    }
    #ifdef UL_WITH_DEVFS
    if(udrv->devfs_handle) kc_devfs_delete(udrv->devfs_handle);
    #endif /* UL_WITH_DEVFS */
    ul_drv_free(udrv);
  }
  pci_set_drvdata(dev, NULL);
  printk(KERN_INFO "ulan_remove_one: PCI device removed\n");
}

static struct pci_device_id ulan_pci_tbl[] /*__devinitdata*/ = {
	/* Advantech PCI-1602A, negated DTR */
	{ /* PCI_VENDOR_ID */ 0x13fe, /* PCI_DEVICE_ID */ 0x1600,
	  /* PCI_VENDOR_ID */ PCI_ANY_ID, /* PCI_SUBDEVICE_ID */ PCI_ANY_ID,
	  0, 0, 0x16954000|U950PCI_CHOPT_TXDTRNEG|U950PCI_CHOPT_HSPDOSC},
	/* Tedia PCI-1482 */
	{ /* PCI_VENDOR_ID */ 0x1415, /* PCI_DEVICE_ID */ 0x950A,
	  /* PCI_VENDOR_ID */ PCI_ANY_ID, /* PCI_SUBDEVICE_ID */ PCI_ANY_ID,
	  0, 0, 0x16954000|U950PCI_CHOPT_RXDONRI},
	/* Tedia PCI-1482 with PiKRON signature */
	{ /* PCI_VENDOR_ID */ 0x1760, /* PCI_DEVICE_ID */ 0x8004,
	  /* PCI_VENDOR_ID */ PCI_ANY_ID, /* PCI_SUBDEVICE_ID */ PCI_ANY_ID,
	  0, 0, 0x16954000|U950PCI_CHOPT_RXDONRI},
	{ 0, } /* terminate list */
};

MODULE_DEVICE_TABLE(pci, ulan_pci_tbl);

static struct pci_driver ulan_pci_driver = {
	name:		"ul_drv",
	id_table:	ulan_pci_tbl,
	probe:		ulan_init_one,
	remove:		ulan_remove_one,
};

#endif  /* UL_WITH_PCI */


/*** initialize kernel module ***/
int init_module(void)
{
 #ifdef UL_WITH_DEVFS
  kc_devfs_handle_t devfs_handle;
  char devname[32];
 #endif /* UL_WITH_DEVFS */
  ul_drv* udrv;
  int ret;
  int minor;

  printk(KERN_INFO "uLan v" UL_DRV_VERSION " init_module\n");
  printk(KERN_INFO "Compiled at %s %s\n",__DATE__, __TIME__);

  if(debug!=-1){
    uld_debug_flg=debug;
  }

  ret=0;
 #ifdef UL_WITH_UARTS
  for(minor=0;minor<UL_MINORS;minor++)
  {
    if(port[minor]||chip[minor])
    { /* get new ul_drv (port, irq, baud, my_adr); */
      udrv=ul_drv_new(port[minor],irq[minor],
   		      baud[minor],my_adr[minor],chip[minor]);
      ul_drv_arr[minor]=udrv;
      if(udrv){
	#ifdef UL_WITH_DEVFS
	  sprintf (devname, "ulan%d", minor);
          devfs_handle=kc_devfs_new_cdev(NULL, MKDEV(ulan_major_dev, minor), 
			S_IFCHR | S_IRUSR | S_IWUSR | S_IRGRP | S_IWGRP, 
			&ulan_fops, udrv, devname);
	  udrv->devfs_handle=devfs_handle;
	#endif /* UL_WITH_DEVFS */
	ret++;
      };
    };
  };
 #endif /*UL_WITH_UARTS*/

 #ifdef UL_WITH_PCI
  { int rc;
    rc=pci_register_driver (&ulan_pci_driver);
    ret+=rc;
  }
 #endif  /* UL_WITH_PCI */

  if(!ret)
  { printk(KERN_ERR "uLan module_init : no hardware initialized, aborting !!!\n");
    #ifdef UL_WITH_PCI
    pci_unregister_driver (&ulan_pci_driver);
    #endif  /* UL_WITH_PCI */
    return -ENODEV;
  };

 #ifdef UL_WITH_DEVFS
  ret = devfs_register_chrdev(ulan_major_dev, "ulan", &ulan_fops);
 #else /* UL_WITH_DEVFS */
  ret = register_chrdev(ulan_major_dev, "ulan", &ulan_fops);
 #endif /* UL_WITH_DEVFS */
  printk(KERN_INFO "uLan init : called register dev %d and got %d\n",
			 ulan_major_dev,ret);
  if (ret < 0) {
    printk(KERN_ERR "uLan init : unable to grab device #\n");
    ulan_major_dev=0;
    #ifdef UL_WITH_PCI
    pci_unregister_driver (&ulan_pci_driver);
    #endif  /* UL_WITH_PCI */
    for(minor=0;minor<UL_MINORS;minor++)
    { udrv=ul_drv_arr[minor];
      if(udrv)
      { ul_drv_arr[minor]=NULL;
	#ifdef UL_WITH_DEVFS
	if(udrv->devfs_handle) kc_devfs_delete(udrv->devfs_handle);
	#endif /* UL_WITH_DEVFS */
	ul_drv_free(udrv);
      };
    };
    return ret;
  }
  if(!ulan_major_dev) ulan_major_dev=ret;
  printk(KERN_INFO "uLan init : driver succesfully registered as major %d\n",
	          ulan_major_dev);

  #ifdef UL_INT_TST
  udrv=ul_drv_arr[0];
  if(udrv)
  { printk("uLan : geninfoblk\n");
    geninfoblk(udrv);
    printk("uLan : gentestblk\n");
    gentestblk(udrv);
    printk("uLan : uld_timeout\n");
    uld_timeout(udrv);
  };
  #endif /* UL_INT_TST */

  return 0;
}

/*** cleanup kernel module ***/
void cleanup_module(void)
{
  ul_drv* udrv;
  int i;
  int ret;

  if (ulan_major_dev != 0) {
   #ifdef UL_WITH_DEVFS
    ret = devfs_unregister_chrdev(ulan_major_dev, "ulan");
   #else /* UL_WITH_DEVFS */
    ret = unregister_chrdev(ulan_major_dev, "ulan");
   #endif /* UL_WITH_DEVFS */
    if(ret>=0)
      printk(KERN_INFO "uLan cleanup: Unregister  O.K.\n");
    else
      printk(KERN_CRIT "uLan cleanup: Unregister failed\n");
  };

  udrv=ul_drv_arr[0];
  if(udrv)
  { printk("uLan : printudrvbll\n");
    printudrvbll(udrv);
    printk("uLan : max cycle cnt  %d\n",ul_max_cycle_cnt);
    printk("uLan : max cycle time %d\n",ul_max_cycle_time);
  }

 #ifdef UL_WITH_PCI
  pci_unregister_driver (&ulan_pci_driver);
 #endif  /* UL_WITH_PCI */
  for(i=0;i<UL_MINORS;i++)
  { udrv=ul_drv_arr[i];
    ul_drv_arr[i]=NULL;
    if(udrv){
     #ifdef UL_WITH_DEVFS
     if(udrv->devfs_handle) kc_devfs_delete(udrv->devfs_handle);
     #endif /* UL_WITH_DEVFS */
     ul_drv_free(udrv);
    }
  }
 #ifdef ENABLE_UL_MEM_CHECK
  UL_PRINTF("MEM_CHECK: malloc-free=%d\n",
		 (int)atomic_read(&ul_mem_check_counter));
 #endif /* ENABLE_UL_MEM_CHECK */
}

#endif /* FOR_LINUX_KERNEL */

/*******************************************************************/
/* Interface for driver in library */

#ifdef UL_DRV_IN_LIB

#define UL_MINORS 4
ul_drv *ul_drv_arr[UL_MINORS]={[0 ... UL_MINORS-1]=NULL};

void uld_check_timeout(void);
ul_drv *uld_try_open_new(const char *dev_name,
                         const char *options, int minor);

ul_opdata *ul_open(const char *dev_name, const char *options)
{
  ul_drv *udrv;
  ul_opdata *opdata;
  unsigned int minor;
  const char *p;
  if(dev_name==NULL) { errno=ENODEV; return NULL;}
  if(!strncasecmp(dev_name,"COM",3)){  /* strnicmp */
    p=dev_name+3;
    if(*p<'1') { errno=ENODEV; return NULL;}
    minor=*p-'1';
  }else{ if(!strncmp(dev_name,"/dev/ulan",9)){
      p=dev_name+9;
    }else {
      p=dev_name;
    }
    if(!*p||(*p==' ')||(*p==':')) minor=0;
    else{
      if(*p<'0') { errno=ENODEV; return NULL;}
      minor=*p-'0';
    }
  }
  if (minor >= UL_MINORS) { errno=ENODEV; return NULL;}
  if (!ul_drv_arr[minor]){
    ul_drv_arr[minor]=uld_try_open_new(dev_name,options,minor);
  }
  if (!ul_drv_arr[minor]){
    errno=ENODEV; return NULL;
  }
  udrv=ul_drv_arr[minor];

  if(!(opdata=MALLOC(sizeof(ul_opdata))))
  { errno=ENOMEM;
    return NULL;
  };
  memset(opdata,0,sizeof(ul_opdata));
  /* opdata->file=file; */
  opdata->magic=ULOP_MAGIC;
  opdata->message=NULL;
  opdata->udrv=udrv;
  /* init_waitqueue_head(&opdata->wqrec); */
  opdata->opprew=NULL;
  opdata->opnext=NULL;
  opdata->recchain=NULL;
  opdata->filtchain=NULL;
  { /* add us onto list of clients of udrv */
    ul_opdata *opptr;
    UL_DRV_LOCK_FINI
    UL_DRV_LOCK;
    opptr=udrv->operators;
    if(opptr) {opptr->opprew=opdata;opdata->opnext=opptr;};
    UL_MB();
    udrv->operators=opdata;
    UL_DRV_UNLOCK;
  };
  uld_printk_flush();
  return opdata;
}

int ul_close(ul_opdata* opdata)
{
  ul_opchain *opmember;

  if (opdata->message) ulan_freemsg(opdata);
  { /* delete us from list of clients of udrv */
    ul_drv *udrv=opdata->udrv;
    ul_opdata *opptr;
    UL_DRV_LOCK_FINI
    UL_DRV_LOCK;
    if((opptr=opdata->opnext)) opptr->opprew=opdata->opprew;
    if((opptr=opdata->opprew)) opptr->opnext=opdata->opnext;
    else udrv->operators=opdata->opnext;
    UL_DRV_UNLOCK;
    if(udrv->irq>0) ul_synchronize_irq(udrv->irq);
  };
  schedule();
  while((opmember=opdata->recchain))
  {
    del_from_opchain(&opdata->recchain,opmember);
    if(opmember->message) ul_dec_ref_cnt(opmember->message);
    FREE(opmember);
  };
  while((opmember=opdata->filtchain))
  {
    del_from_opchain(&opdata->filtchain,opmember);
    FREE(opmember);
  };
  FREE(opdata);

  /* UL_PRINTF("uLan : printudrvbll\n"); */
  /* printudrvbll(opdata->udrv); */
  
  uld_printk_flush();
  return 0;
}

ssize_t	ul_read(ul_opdata* opdata, void *buf, size_t count)
{
  int cn;
  int len;

  if(!opdata->message)
  {
    errno=ENOMSG;
    return -1;
  }

  if(opdata->data.pos+count>UL_BLK_HEAD(opdata->data.head_blk).len)
   count=UL_BLK_HEAD(opdata->data.head_blk).len-opdata->data.pos;

  cn=count;
  while(cn>0)
  {
    if(!ul_di_adjust(&opdata->data))
    {
      memset(buf,0,cn);
      cn=0;
      break;
    };
    len=ul_di_atonce(&opdata->data);
    if(len>cn) len=cn;
    memcpy(buf,ul_di_byte(&opdata->data),len);
    ul_di_add(&opdata->data,len);
    buf+=len;
    cn-=len;
  };

  return count-cn;
};

ssize_t	ul_write(ul_opdata* opdata, const void *buf, size_t count)
{
  int cn;
  int len;
  ul_mem_blk *blk;

  if(!opdata->message)
  {
    errno=ENOMSG;
    return -1;
  }

  cn=count;
  while(cn>0)
  {
    while(!ul_di_adjust(&opdata->data))
    {
      if(!(blk=ul_alloc_blk(opdata->udrv))) 
        {count-=cn;cn=0;break;};
      memset(UL_BLK_NDATA(blk),0,UL_BLK_SIZE);
      opdata->data.blk->next=blk;
    };
    len=ul_di_atonce(&opdata->data);
    if(len>cn) len=cn;
    memcpy(ul_di_byte(&opdata->data),buf,len);
    ul_di_add(&opdata->data,len);
    buf+=len;
    cn-=len;
  };

  if(opdata->data.pos>UL_BLK_HEAD(opdata->data.head_blk).len)
    UL_BLK_HEAD(opdata->data.head_blk).len=opdata->data.pos;

  return count-cn;
};

int ul_newmsg(ul_opdata* opdata,const ul_msginfo *msginfo)
{ int ret;
  ret=ulan_newmsg(opdata,msginfo);
  if(ret<0) {errno=-ret;ret=-1;}
  return ret;
}

int ul_tailmsg(ul_opdata* opdata,const ul_msginfo *msginfo)
{ int ret;
  ret=ulan_tailmsg(opdata,msginfo);
  if(ret<0) {errno=-ret;ret=-1;}
  return ret;
}

int ul_freemsg(ul_opdata* opdata)
{ int ret;
  ret=ulan_freemsg(opdata);
  if(ret<0) {errno=-ret;ret=-1;}
  return ret;
}

int ul_acceptmsg(ul_opdata* opdata,ul_msginfo *msginfo)
{ int ret;
  ret=ulan_acceptmsg(opdata,msginfo);
  if(ret<0) {errno=-ret;ret=-1;}
  return ret;
}

int ul_actailmsg(ul_opdata* opdata,ul_msginfo *msginfo)
{ int ret;
  ret=ulan_actailmsg(opdata,msginfo);
  if(ret<0) {errno=-ret;ret=-1;}
  return ret;
}

int ul_addfilt(ul_opdata* opdata,const ul_msginfo *msginfo)
{ int ret;
  ret=ulan_addfilt(opdata,msginfo);
  if(ret<0) {errno=-ret;ret=-1;}
  return ret;
}

int ul_abortmsg(ul_opdata* opdata)
{ int ret;
  ret=ulan_abortmsg(opdata);
  if(ret<0) {errno=-ret;ret=-1;}
  return ret;
}

int ul_rewmsg(ul_opdata* opdata)
{ int ret;
  ret=ulan_rewmsg(opdata);
  if(ret<0) {errno=-ret;ret=-1;}
  return ret;
}

int ul_inepoll(ul_opdata* opdata)
{ int ret;
 #if 1 /* Processing of arrived or processed messages */
  uld_check_timeout();
  if(uld_test_dfl(opdata->udrv,ASK_BOTTOM))
  {
    ulan_do_bh(opdata->udrv);
  }
 #endif
  ret=ulan_inepoll(opdata);
  uld_printk_flush();
  if(ret<0) {errno=-ret;ret=-1;}
  return ret;
}

int ul_hwtest(ul_opdata* opdata, unsigned int fnc)
{ int ret;
  ret=ulan_hwtest(opdata,fnc);
  if(ret<0) {errno=EINVAL;ret=-1;}
  return ret;
}

int ul_drv_debflg(ul_opdata* opdata,int debug_msk)
  {uld_debug_flg=debug_msk;return uld_debug_flg;};

#include <stdarg.h>

#define ULD_PRINTK_BUF_LEN 4096
char uld_printk_buf[ULD_PRINTK_BUF_LEN];
char *uld_printk_pin=uld_printk_buf;
char *uld_printk_pout=uld_printk_buf;

int uld_printk(const char *format, ...)
{
  int len,rest,ret;
  char buf[255];
  va_list va=0;
  UL_IRQ_LOCK_FINI
  if(uld_debug_flg&LOG_MSK_DISABLE) return 0;
  va_start(va,format);
  ret=vsprintf(buf,format,va);
  va_end(va);
  len=strlen(buf);
  UL_IRQ_LOCK;
  rest=uld_printk_buf+ULD_PRINTK_BUF_LEN-uld_printk_pin;
  if(len>=rest){
    memcpy(uld_printk_pin,buf,rest); len-=rest;
    memcpy(uld_printk_buf,buf,len);
    uld_printk_pin=uld_printk_buf+len;
  }else{
    memcpy(uld_printk_pin,buf,len); uld_printk_pin+=len;
  }
  UL_IRQ_UNLOCK;
  return ret;
}

void uld_printk_flush(void)
{
  char *pout,*pin;
  UL_IRQ_LOCK_FINI
  UL_IRQ_LOCK;
  pin=uld_printk_pin;
  pout=uld_printk_pout;
  UL_IRQ_UNLOCK;
  if(pin==pout) return;
  if(pin>pout){
    fwrite(pout,pin-pout,1,stderr); uld_printk_pout=pin;
  }else{
    fwrite(pout,uld_printk_buf+ULD_PRINTK_BUF_LEN-pout,1,stderr);
    fwrite(uld_printk_buf,pin-uld_printk_buf,1,stderr);
    uld_printk_pout=pin;
  }
  fflush(stderr);
}

void uld_check_timeout(void)
{
  int i;
  ul_drv *udrv;
 #ifdef JIFFIES_FROM_GETTIMEOFDAY
  struct timeval actual;
  long int new_expires;
  static long int old_expires;
 #endif /* JIFFIES_FROM_GETTIMEOFDAY */
  UL_DRV_LOCK_FINI
 #ifdef JIFFIES_FROM_GETTIMEOFDAY
  gettimeofday(&actual,NULL);
  uld_jiffies=actual.tv_usec+actual.tv_sec*ULD_HZ;
 #endif /* JIFFIES_FROM_GETTIMEOFDAY */
  for(i=0;i<UL_MINORS;i++)
  {
    if((udrv=ul_drv_arr[i])!=NULL)
    {
      if(uld_test_dfl(udrv,WDSCHED))
      { 
       #ifdef JIFFIES_FROM_GETTIMEOFDAY
        if(old_expires!=udrv->wd_timer_expires)
        { UL_DRV_LOCK;
          new_expires=uld_jiffies+ULD_HZ/18;
          if((new_expires-udrv->wd_timer_expires)<0)
            new_expires=udrv->wd_timer_expires;
          old_expires=udrv->wd_timer_expires=new_expires;
          UL_DRV_UNLOCK;
          LOG_IRQ("sched timeout %ld at time %ld\n",new_expires,uld_jiffies);
        }
       #elif 1
        { static long int old_expires;
          if(old_expires!=udrv->wd_timer_expires)
            LOG_IRQ("sched timeout %ld at time %ld\n",
                    old_expires=udrv->wd_timer_expires,jiffies);
        }
       #endif /* JIFFIES_FROM_GETTIMEOFDAY */
        if((udrv->wd_timer_expires-jiffies)<0)
        { uld_clear_dfl(udrv,WDSCHED);
          ulan_do_wd_timer(udrv);
          /* uld_timeout(udrv); */
          /* printudrvfncstack(udrv) */
        }
      }
      /* Call bottom halves as needed */
      if(uld_test_dfl(udrv,ASK_BOTTOM))
      {
        uld_clear_dfl(udrv,ASK_BOTTOM);
        ulan_do_bh(udrv);
      }
    }
  }
 #ifdef __DJGPP__
  #if 1
  __save_flags(ul_irq_lock_flags);
  if(!(ul_irq_lock_flags&0x0200))
  { LOG_FATAL(KERN_CRIT "Somebody disables IRQ !!!!!\n");
    __sti();
  }
  #endif
  __dpmi_yield();
 #endif
}

#if defined(__DJGPP__)||1
typedef struct uld_dev_params{
  int port;
  int irq;
  int baud;
  int my_adr;
}uld_dev_params_t;

static const uld_dev_params_t
  uld_dev_params_defaults[4]={
    {0x3f8,0x4,19200,2},
    {0x2f8,0x3,19200,2},
    {0x3e8,0x5,19200,2},
    {0x2e8,0x5,19200,2}
  };

ul_drv *uld_try_open_new(const char *dev_name,
                         const char *options, int minor)
{
  uld_dev_params_t params;
  const char *p,*r;
  char *s;
  char chip_name[16];
  if(minor<4) params=uld_dev_params_defaults[minor];
  else  params=uld_dev_params_defaults[0];
  chip_name[0]=0;
  p=dev_name;
  while(*p){
    while(*p&&(*p!=' ')&&(*p!=':')) p++;
    while(*p&&((*p==' ')||(*p==':'))) p++;
    r=p;
    while(*r&&(*r!=' ')&&(*r!=':')&&((*r<'0')||(*r>'9'))) r++;
    switch (*p) {
      case 'p': params.port=strtol(r,NULL,0); break;
      case 'i': params.irq=strtol(r,NULL,0); break;
      case 'b': params.baud=strtol(r,NULL,0); break;
      case 'a': params.my_adr=strtol(r,NULL,0); break;
      case 'c': for(s=chip_name,p++;*p&&(*p!=':');p++,s++)
                   *s=*p;
		*s=0;
                break;
    }
  }
  /* printf("port %d irq %d baud %d my_adr %d\n",
    params.port,params.irq,params.baud,params.my_adr);
  return NULL; */
  return ul_drv_new(params.port,params.irq,params.baud,params.my_adr,
  		    chip_name[0]?chip_name:NULL);
}

#endif /* __DJGPP__ */

#endif /* UL_DRV_IN_LIB */

#ifdef __DJGPP__

#include <go32.h>

#ifdef ENABLE_UL_MEM_CHECK
void * ul_mem_check_malloc(size_t size)
{ void *ptr;
  ptr=malloc(size);
  if(ptr){
    ul_mem_check_counter++;
    _go32_dpmi_lock_data(ptr,size);
  }
  return ptr;
}

void ul_mem_check_free(void *ptr)
{
  if(!ptr) {
    LOG_FATAL(KERN_CRIT "ul_mem_check_free : triing to free NULL ptr\n");
  }else{
    ul_mem_check_counter--;
    free(ptr);
  }
}
#endif /* ENABLE_UL_MEM_CHECK */

static _go32_dpmi_seginfo wrapper;
static __dpmi_paddr old_handler, new_handler;
uchar pic1_old, pic2_old, pic_mask;
int pm_irq;

void *uld_djgpp_dev_id=NULL;

int uld_djgpp_irq_mask_set(int irqline, int enable)
{
  unsigned long irq_flags;
  unsigned mask;
  if(irqline>=16) return -1;
  __save_flags(irq_flags);__cli();
  if(irqline<8){
    mask=inportb(0x21);
    mask=enable?mask&~(1<<irqline):mask|(1<<irqline);
    outportb(0x21,mask);                /* Change IRQ mask */
  }else{
    outportb(0x21,inportb(0x21)&0xFB);  /* Enable IRQ2 */
    mask=inportb(0xA1);
    mask=enable?mask&~(1<<(irqline-8)):mask|(1<<(irqline-8));
    outportb(0xA1,mask);                /* Change IRQ mask */
  }
  __restore_flags(irq_flags);
  return 0;
}

void uld_djgpp_irq_eoi(int irqline)
{
  if(irqline<8){
    outportb(0x20,0x20|0x40|irqline);
  }else{
    outportb(0xA0,0x20|0x40|(irqline&7));
    outportb(0x20,0x20|0x40|0x2);
  }
}


/* !!!!!!!!!!!! */
void uld_djgpp_call_irq(void)
{
  return;
  if(ul_drv_arr[0])
    uld_irq_handler(0, ul_drv_arr[0], NULL);
}

void uld_djgpp_proc_irq(void)
{
  if(uld_djgpp_dev_id){
    uld_irq_handler(0, uld_djgpp_dev_id, NULL);
    uld_djgpp_irq_eoi(((ul_drv*)uld_djgpp_dev_id)->irq);
  }else{
    outportb(0xA0,0x20);
    outportb(0x20,0x20);
  }
}

int request_irq(int irq,void *handler,int flags,char *name,void *dev_id)
{
 #if 1
  /* !!!!!!!!!!!! */
  handler=uld_djgpp_proc_irq;
  uld_djgpp_dev_id=dev_id;

  pic1_old=inportb(0x21);
  pic2_old=inportb(0xA1);

  if(irq<8) {
    pm_irq=irq+0x08;
  } else {
    pm_irq=irq+0x68;
  }
  uld_djgpp_irq_mask_set(irq,0);

  wrapper.pm_offset=(int)handler;
  wrapper.pm_selector=_my_cs();
  _go32_dpmi_allocate_iret_wrapper(&wrapper);
  new_handler.offset32=wrapper.pm_offset;
  new_handler.selector=wrapper.pm_selector;
  __dpmi_get_protected_mode_interrupt_vector(pm_irq,&old_handler);
  if(__dpmi_set_protected_mode_interrupt_vector(pm_irq,&new_handler)<0)
  { LOG_FATAL("uLan request_irq: DPMI set interrupt failed\n");
    return -1;
  };
  uld_djgpp_irq_mask_set(irq,1);

 #endif
  return irq;
}

int free_irq(int irq,void *ctx)
{
  uld_djgpp_irq_mask_set(irq,0);
  __dpmi_set_protected_mode_interrupt_vector(irq,&old_handler);

  outportb(0x21,pic1_old);
  outportb(0xA1,pic2_old);
  _go32_dpmi_free_iret_wrapper(&wrapper);
  return 0;
}


#include <crt0.h>
int _crt0_startup_flags = _CRT0_FLAG_LOCK_MEMORY;
int uld_djgpp_data_end;
void uld_djgpp_code_end(void) {;};

void uld_djgpp_do_atexit(void)
{ int i;
  for(i=0;i<UL_MINORS;i++)
  {
    if(ul_drv_arr[i])
    {
      ul_drv_free(ul_drv_arr[i]);
      ul_drv_arr[i]=NULL;
    }
  }
  UL_PRINTF("ulan_driver : atexit finished driver\n");
  #ifdef ENABLE_UL_MEM_CHECK
    UL_PRINTF("              mem_check_counter = %ld\n",ul_mem_check_counter);
  #endif /* ENABLE_UL_MEM_CHECK */
  uld_printk_flush();
}

void uld_djgpp_do_atinit(void) __attribute__((constructor));

void uld_djgpp_do_atinit(void)
{
  if(!uld_debug_flg) uld_debug_flg=ULD_DEBUG_DEFAULT;
  if((_go32_dpmi_lock_code(&uld_djgpp_code_beg,(char*)&uld_djgpp_code_end-(char*)&uld_djgpp_code_beg))||
     (_go32_dpmi_lock_data(&uld_djgpp_data_beg,(char*)&uld_djgpp_data_end-(char*)&uld_djgpp_data_beg)))
  {
    UL_PRINTF("ulan_driver : Cannot lock code and data in memory !\n");
  }
  atexit(uld_djgpp_do_atexit);
  uld_check_timeout();

  UL_PRINTF("ulan_driver : atinit finished\n");
  uld_printk_flush();
}

#endif /* __DJGPP__ */

/*******************************************************************/
/* NT Kernel Mode Driver specific code */ 

#ifdef _WIN32

#define uLan_DbgPrint DbgPrint

//
// Defaults
//

#define DEF_PORT_ADDRESS    0x2F8
#define DEF_PORT_RANGE      0x08
#define DEF_IRQ_LINE        0x03
#define DEF_BAUD_RATE       19200
#define DEF_BUFFER_SIZE     2048
#define DEF_MY_ADDR         2


//-------------------------------------------------------------------
//
// Device extension 
// 

typedef struct ul_drv ULAN_DEVICE_EXTENSION,
                      *PULAN_DEVICE_EXTENSION;

//-------------------------------------------------------------------
//
// Define the driver names
// 
#define NT_DEVICE_NAME	    L"\\Device\\ul_drv"
#define DOS_DEVICE_NAME     L"\\DosDevices\\ul_drv"

//-------------------------------------------------------------------
//
// Declare forward function references
//

NTSTATUS DispatchRoutine (IN PDEVICE_OBJECT DeviceObject, IN PIRP Irp);

VOID UnloadDriver (IN PDRIVER_OBJECT DriverObject);

BOOLEAN ReportUsage(IN PDRIVER_OBJECT DriverObject,
	            IN PDEVICE_OBJECT DeviceObject,
                    IN INTERFACE_TYPE InterfaceType,
                    IN ULONG BusNumber,
                    IN PHYSICAL_ADDRESS PortAddress,
		    IN LONG PortRange,
		    IN KIRQL IRQLine,
                    IN BOOLEAN *ConflictDetected);

VOID ulan_bottom_dpc(IN PKDPC Dpc,IN PVOID contex,
		     IN PVOID arg1,IN PVOID arg2);

//-------------------------------------------------------------------
// Memory check

#ifdef ENABLE_UL_MEM_CHECK
void * ul_mem_check_malloc(size_t size)
{ void *ptr;
  ptr=ExAllocatePool(NonPagedPool,size);
  if(ptr) InterlockedIncrement(&ul_mem_check_counter);
  return ptr;
}

void ul_mem_check_free(void *ptr)
{
  if(ptr) InterlockedDecrement(&ul_mem_check_counter);
  ExFreePool(ptr);
}
#endif /* ENABLE_UL_MEM_CHECK */

//-------------------------------------------------------------------
//
//  Initialize and free driver
//
/* initialize new driver from given parameters */

NTSTATUS ul_drv_init_ext(ul_drv *udrv, int port,
	 int irq, int baud, int chip_options, int buffer_size, int my_adr)
{
  int ret;
  int test_cnt;
  unsigned long probe_irqs;
  udrv->flag_IN_ISR=0;
  udrv->flag_ASK_ISR=0;
  udrv->flag_NACTIV=0;
  udrv->flag_CHIPOK=0;
  udrv->flag_IN_BOTTOM=0;
  udrv->flag_ASK_BOTTOM=0;
  udrv->flag_CHECK_FILT=0;
  if(buffer_size<0x2000) buffer_size=0x2000;
  /* init sequencer */
  ul_drv_new_init_state(udrv,my_adr);
  /* init chip driver */
  #ifdef UL_WITH_PCI
  if((chip_options&~0xff)==0x16954000){
    if((ret=u950pci_init(udrv, port, irq, baud, chip_options))>=0){
      UL_PRINTF("uLan ul_drv_new : 16C954 at 0x%x\n",port);
    }
  }else
  #endif /*UL_WITH_PCI*/
  {
    if((ret=u510_init(udrv, port, irq, baud, chip_options))>=0){
      UL_PRINTF("uLan ul_drv_new : 82510 found at 0x%x\n",port);
    }else if((ret=u450_init(udrv, port, irq, baud, chip_options))>=0){
      UL_PRINTF("uLan ul_drv_new : 16450 found at 0x%x\n",port);
    };
  }
  if(ret>=0)
  { /* setups buffers, ports and first part of irq
      for sucesfully detected device */
    if(ul_drv_new_start(udrv,buffer_size)>=0)
      return STATUS_SUCCESS;
  }
  UL_PRINTF("uLan ul_drv_new : failed to init\n");
  return STATUS_INSUFFICIENT_RESOURCES;
};

/* destroy driver */
void ul_drv_done_ext(ul_drv *udrv)
{
  int old_CHIPOK;
  if(!udrv) return;
  old_CHIPOK=udrv->flag_CHIPOK;
  udrv->flag_CHIPOK=0;
  if(old_CHIPOK&&udrv->irq){
    /* free_irq(udrv->irq,udrv)*/;
    /* synchronize_irq(); */
  }
  DONE_UDRV_WDTIM(udrv);
  /* schedule(); */
  ul_mem_done(udrv);
  if(old_CHIPOK) { 
    udrv->fnc_cctrl(udrv,UL_CC_PDONE,0);
  };
};


//-------------------------------------------------------------------
//
//  File operations
//

NTSTATUS ul_drv_open(ul_drv *udrv,ul_opdata *opdata)
{ 
  opdata->magic=ULOP_MAGIC;
  opdata->message=NULL;
  opdata->udrv=udrv;
  /* init_waitqueue_head(&opdata->wqrec); */
  opdata->opprew=NULL;
  opdata->opnext=NULL;
  opdata->recchain=NULL;
  opdata->filtchain=NULL;
  opdata->wait_irp=NULL;
  { /* add us onto list of clients of udrv */
    ul_opdata *opptr;
    UL_DRV_LOCK_FINI
    UL_DRV_LOCK;
    opptr=udrv->operators;
    if(opptr) {opptr->opprew=opdata;opdata->opnext=opptr;};
    UL_MB();
    udrv->operators=opdata;
    UL_DRV_UNLOCK;
  };

  return 0;
}

NTSTATUS ul_drv_close(ul_drv *udrv,ul_opdata *opdata)
{
  ul_opchain *opmember;
  KIRQL OldIrql;

  if (opdata->message) ulan_freemsg(opdata);

  KeRaiseIrql(DISPATCH_LEVEL,&OldIrql);
  {while(InterlockedExchange(&udrv->flag_IN_BOTTOM,1))/* spinlock loop*/;}
  { 
    /* delete us from list of clients of udrv */
    ul_drv *udrv=opdata->udrv;
    ul_opdata *opptr;
    UL_DRV_LOCK_FINI
    UL_DRV_LOCK;
    if((opptr=opdata->opnext)) opptr->opprew=opdata->opprew;
    if((opptr=opdata->opprew)) opptr->opnext=opdata->opnext;
    else udrv->operators=opdata->opnext;
    UL_DRV_UNLOCK;
  };
  InterlockedExchange(&udrv->flag_IN_BOTTOM,0);
  ulan_bottom_dpc(NULL,udrv,NULL,NULL);
  KeLowerIrql(OldIrql);

  while((opmember=opdata->recchain))
  {
    del_from_opchain(&opdata->recchain,opmember);
    if(opmember->message) ul_dec_ref_cnt(opmember->message);
    FREE(opmember);
  };
  while((opmember=opdata->filtchain))
  {
    del_from_opchain(&opdata->filtchain,opmember);
    FREE(opmember);
  };
  return 0;
};


NTSTATUS ulan_read(ul_drv *udrv,ul_opdata *opdata,
				   uchar *buf, ULONG count,
				   LARGE_INTEGER pos, PULONG pcount)
{
  int cn;
  int len;

  if(!opdata->message) return -ENOMSG;

  if(opdata->data.pos+count>UL_BLK_HEAD(opdata->data.head_blk).len)
    count=UL_BLK_HEAD(opdata->data.head_blk).len-opdata->data.pos;

  cn=count;
  while(cn>0)
  {
    if(!ul_di_adjust(&opdata->data))
    {
      RtlZeroMemory(buf,cn);
      cn=0;
      break;
    };
    len=ul_di_atonce(&opdata->data);
    if(len>cn) len=cn;
    RtlCopyMemory(buf,ul_di_byte(&opdata->data),len);
    ul_di_add(&opdata->data,len);
    buf+=len;
    cn-=len;
  };

  *pcount=count-cn;

  return 0;
};


NTSTATUS ulan_write(ul_drv *udrv,ul_opdata *opdata,
				   const uchar *buf, ULONG count,
				   LARGE_INTEGER pos, PULONG pcount)
{
  int cn;
  int len;
  ul_mem_blk *blk;

  if(!opdata->message) return -ENOMSG;

  cn=count;
  while(cn>0)
  {
    while(!ul_di_adjust(&opdata->data))
    {
      if(!(blk=ul_alloc_blk(opdata->udrv))) 
        {count-=cn;cn=0;break;};
      RtlZeroMemory(UL_BLK_NDATA(blk),UL_BLK_SIZE);
      opdata->data.blk->next=blk;
    };
    len=ul_di_atonce(&opdata->data);
    if(len>cn) len=cn;
    RtlCopyMemory(ul_di_byte(&opdata->data),buf,len);
    ul_di_add(&opdata->data,len);
    buf+=len;
    cn-=len;
  };

  if(opdata->data.pos>UL_BLK_HEAD(opdata->data.head_blk).len)
   UL_BLK_HEAD(opdata->data.head_blk).len=opdata->data.pos;

  *pcount=count-cn;

  return 0;
};


//-------------------------------------------------------------------
//
//  deffered proccessing
//

/* notice about new message in udrv->proc_bll */
VOID ulan_bottom_dpc(IN PKDPC Dpc,IN PVOID contex,
		     IN PVOID arg1,IN PVOID arg2)
{
  ul_mem_blk *message;
  ul_drv *udrv=(ul_drv *)contex;

  if(udrv->magic!=UL_DRV_MAGIC)
  { UL_PRINTF("ulan_bottom_dpc : BAD udrv magic !!!");
    return;
  }

  #ifdef ENABLE_UL_IRQ_STALE_WDG
  ul_irq_stale_wdg_cnt=0;
  #endif /* ENABLE_UL_IRQ_STALE_WDG */

  LOG_MESSAGES("ulan_bottom_dpc : announcing messages\n");
  while(udrv->proc_bll.first||uld_test_dfl(udrv,ASK_BOTTOM))
  { /*  more Dpc can be started simultaneously on NT SMP box */
    if(uld_test_and_set_dfl(udrv,IN_BOTTOM))
    { LOG_MESSAGES("ulan_bottom_dpc : else is already in");
	  return;
    }
    uld_clear_dfl(udrv,ASK_BOTTOM);
    if(uld_test_dfl(udrv,CHECK_FILT)){
      uld_clear_dfl(udrv,CHECK_FILT);
      check_for_filtnew(udrv);
    }
    while((message=udrv->proc_bll.first))
    {
      int i;
      ul_bll_move_mes(&udrv->opan_bll,message);
      i=ulan_proc_arrived(udrv,message);
      LOG_MESSAGES("ulan_bottom_dpc : message sent to %d recchains\n",i);
    };
    uld_atomic_clear_dfl(udrv,IN_BOTTOM);
  }
};

PIRP xch_pending_irp(PIRP *irp_ptr, PIRP new_irp)
{
  PIRP old_irp;
  UL_DRV_LOCK_FINI
  UL_DRV_LOCK;
  old_irp=*irp_ptr;
  *irp_ptr=new_irp;
  UL_DRV_UNLOCK;
  return old_irp;
}

void del_pending_irp(PIRP *irp_ptr, PIRP del_irp)
{
  PIRP old_irp;
  UL_DRV_LOCK_FINI
  UL_DRV_LOCK;
  if(del_irp==*irp_ptr) *irp_ptr=NULL;
  UL_DRV_UNLOCK;
}

void wake_and_xch_pending_irp(PIRP *irp_ptr, PIRP new_irp, NTSTATUS status)
{
  PIRP old_irp;
  old_irp=xch_pending_irp(irp_ptr,new_irp);
  if(old_irp!=NULL){
    old_irp->IoStatus.Status=status;
    IoCompleteRequest(old_irp,IO_NO_INCREMENT);
  }
}

/*
  PDRIVER_CANCEL IoSetCancelRoutine(Irp,CancelRoutine);
*/
void ulan_cancel(IN PDEVICE_OBJECT DeviceObject, IN PIRP Irp)
{
  PIO_STACK_LOCATION  irpStack;
  ul_opdata	*opdata;
  PIRP old_irp;
  LOG_FILEIO("uLan: Irp Cancelled\n");
  irpStack = IoGetCurrentIrpStackLocation (Irp);
  opdata=irpStack->FileObject->FsContext;
  if(!opdata||(opdata->magic!=ULOP_MAGIC))
  { 
    uLan_DbgPrint ("uLan: corrupted opdata !!!!\n");
  }else if(Irp==opdata->wait_irp){
    del_pending_irp(&opdata->wait_irp,Irp);
  }
  Irp->IoStatus.Information = 0;
  Irp->IoStatus.Status=STATUS_CANCELLED;
  IoCompleteRequest(Irp,IO_NO_INCREMENT);
}

//-------------------------------------------------------------------
//
//  Begin FUNCTIONS
//

NTSTATUS
ulan_GetRegistryDword(
    IN      PWCHAR    RegPath,
    IN      PWCHAR    ValueName,
    IN OUT  PULONG    Value
    )

/*++

Routine Description:

	Obtain a Dword value from the registry


Arguments:

    RegPath  -- supplies absolute registry path
    ValueName    - Supplies the Value Name.
    Value      - receives the REG_DWORD value.

Return Value:

    TRUE if successfull, FALSE on fail.

--*/

{
  UNICODE_STRING path;
  RTL_QUERY_REGISTRY_TABLE paramTable[2];  //zero'd second table terminates parms
  ULONG lDef = *Value;                     // default
  NTSTATUS status;
  WCHAR wbuf[ MAXIMUM_FILENAME_LENGTH ];

  path.Length = 0;
  path.MaximumLength = MAXIMUM_FILENAME_LENGTH * sizeof( WCHAR );  // MAXIMUM_FILENAME_LENGTH defined in wdm.h
  path.Buffer = wbuf;

  RtlZeroMemory(path.Buffer, path.MaximumLength);
  RtlMoveMemory(path.Buffer, RegPath, wcslen( RegPath) * sizeof( WCHAR ));

  RtlZeroMemory(paramTable, sizeof(paramTable));

  paramTable[0].Flags = RTL_QUERY_REGISTRY_DIRECT;
  paramTable[0].Name = ValueName;
  paramTable[0].EntryContext = Value;
  paramTable[0].DefaultType = REG_DWORD;
  paramTable[0].DefaultData = &lDef;
  paramTable[0].DefaultLength = sizeof(ULONG);


  status = RtlQueryRegistryValues( RTL_REGISTRY_ABSOLUTE | RTL_REGISTRY_OPTIONAL,
                                  path.Buffer, paramTable, NULL, NULL);

  return status;
}


//-------------------------------------------------------------------
// ScanForPCICard for WinNT KMD
// WDM uses different technique to query VENDOR_ID and DEVICE_ID
// for example PCI\VEN_1415&DEV_950A&SUBSYS_00001415&REV_00

#define PCI_ANY_ID 0xffff
typedef struct pci_device_id {
  USHORT vendor;
  USHORT device;
  USHORT subvendor;
  USHORT subdevice;
  USHORT class_val,clask_mask;
  ULONG  driver_data;
} pci_device_id_t;

#ifdef UL_WITH_PCI

static struct pci_device_id ulan_pci_tbl[] /*__devinitdata*/ = {
	/* Advantech PCI-1602A, negated DTR */
	{ /* PCI_VENDOR_ID */ 0x13fe, /* PCI_DEVICE_ID */ 0x1600,
	  /* PCI_VENDOR_ID */ PCI_ANY_ID, /* PCI_SUBDEVICE_ID */ PCI_ANY_ID,
	  0, 0, 0x16954000|U950PCI_CHOPT_TXDTRNEG|U950PCI_CHOPT_HSPDOSC},
	/* Tedia PCI-1482 */
	{ /* PCI_VENDOR_ID */ 0x1415, /* PCI_DEVICE_ID */ 0x950A,
	  /* PCI_VENDOR_ID */ PCI_ANY_ID, /* PCI_SUBDEVICE_ID */ PCI_ANY_ID,
	  0, 0, 0x16954000|U950PCI_CHOPT_RXDONRI},
	/* Tedia PCI-1482 with PiKRON signature */
	{ /* PCI_VENDOR_ID */ 0x1760, /* PCI_DEVICE_ID */ 0x8004,
	  /* PCI_VENDOR_ID */ PCI_ANY_ID, /* PCI_SUBDEVICE_ID */ PCI_ANY_ID,
	  0, 0, 0x16954000|U950PCI_CHOPT_RXDONRI},
	{ 0, } /* terminate list */
};

#ifdef FOR_WIN_WDM

BOOLEAN FindPciHWID(PWCHAR HWID,
		    OUT pci_device_id_t **Ppci_device_id
	           )
{
  PWCHAR pwcH, pwcM;
  WCHAR  fldName;
  USHORT vendorID=0;
  USHORT deviceID=0;
  USHORT subvendorID=0;
  USHORT subdeviceID=0;
  ULONG  val, i, c;
  static WCHAR PciHWIDMask[]=L"PCI\\VEN_vvvv&DEV_xxxx&SUBSYS_sssswwww";
  pci_device_id_t *device_id;

  *Ppci_device_id=NULL;
  if(HWID==NULL) return FALSE;

  pwcH=HWID;
  pwcM=PciHWIDMask;
  while(*pwcH&&*pwcM) {
    if(*pwcH==*pwcM){
      pwcH++; pwcM++;
      continue;
    }
    fldName=*pwcM;
    if((fldName!='v')&&(fldName!='x')&&
       (fldName!='s')&&(fldName!='w')) break;
    val=0;
    for(i=4;i--;){
      c=*(pwcH++); pwcM++;
      if(c<='9') c-='0';
      else if(c<='F') c=c-'A'+10;
      else if(c<='f') c=c-'a'+10;
      if(c>15) {val=0;break;}
      val<<=4;
      val+=c;
    }
    switch (fldName) {
      case 'v':
        vendorID=(USHORT)val;
	break;
      case 'x':
        deviceID=(USHORT)val;
	break;
      case 'w':
        subvendorID=(USHORT)val;
	break;
      case 's':
        subdeviceID=(USHORT)val;
	break;
    }
  }
#if DBG
  uLan_DbgPrint("uLan: Ven %04X Dev %04X SubVen %04X SubDev %04X\n", 
                 vendorID,deviceID,subvendorID,subdeviceID);
#endif

  for(device_id=ulan_pci_tbl;device_id->vendor;device_id++){
    if((device_id->vendor!=PCI_ANY_ID)&&(device_id->vendor!=vendorID)) continue;
    if((device_id->device!=PCI_ANY_ID)&&(device_id->device!=deviceID)) continue;
    if((device_id->subvendor!=PCI_ANY_ID)&&(device_id->vendor!=subvendorID)) continue;
    if((device_id->subdevice!=PCI_ANY_ID)&&(device_id->device!=subdeviceID)) continue;
    *Ppci_device_id=device_id;
    return TRUE;
  }

  return FALSE;
}

#else /* !FOR_WIN_WDM */

BOOLEAN ScanForPCICard(IN OUT PULONG PBusNumber,
                       IN OUT PPCI_SLOT_NUMBER PSlotNumber,
                       IN BOOLEAN FindFirst,
		       OUT pci_device_id_t **Ppci_device_id
		      )
{
  ULONG BusNumber=*PBusNumber;
  PCI_SLOT_NUMBER SlotNumber=*PSlotNumber;
  UCHAR Buffer[2*sizeof(USHORT)];
  ULONG Readed=sizeof(Buffer);
  pci_device_id_t *device_id;

  while(1){
    if(FindFirst){
      BusNumber=0;
      SlotNumber.u.AsULONG=0;
      FindFirst=0;
    }else{
      if(Readed>2){
        SlotNumber.u.bits.FunctionNumber=
            (SlotNumber.u.bits.FunctionNumber+1)&7;
      }else{
        SlotNumber.u.bits.FunctionNumber=0;
      }
      if(!SlotNumber.u.bits.FunctionNumber){
        if(!Readed&&!SlotNumber.u.bits.DeviceNumber)
            break;
        SlotNumber.u.bits.DeviceNumber=
            (SlotNumber.u.bits.DeviceNumber+1)&0x1f;
        if(!SlotNumber.u.bits.DeviceNumber)
            BusNumber++;
      }
    }

    Readed=HalGetBusData(PCIConfiguration,BusNumber,SlotNumber.u.AsULONG,
                         Buffer,sizeof(Buffer));
    if(Readed==sizeof(Buffer)){
      for(device_id=&ulan_pci_tbl[0];device_id->vendor;device_id++){
        if((((PPCI_COMMON_CONFIG)Buffer)->VendorID==device_id->vendor)&&
           (((PPCI_COMMON_CONFIG)Buffer)->DeviceID==device_id->device)){
            *PBusNumber=BusNumber;
            *PSlotNumber=SlotNumber;
            if(Ppci_device_id)
                *Ppci_device_id=device_id;
            return TRUE;
        }
      }
    }
  }

  return FALSE;
}                    

#endif /* FOR_WIN_WDM */

#endif /*UL_WITH_PCI*/


#ifndef FOR_WIN_WDM
//---------------------------------------------------------------------------
// ReportUsage
//
// Description:
//  This routine registers (reports) the I/O and IRQ usage for this driver.
//
// Arguments:
//      DriverObject    - Pointer to the driver object
//      DeviceObject    - Pointer to the Device object
//      PortAddress     - Address of I/O port used
//      ConflictDetected - TRUE if a resource conflict was detected.
//
// Return Value:
//      TRUE    - If a Resource conflict was detected
//      FALSE   - If no conflict was detected
//
BOOLEAN ReportUsage(IN PDRIVER_OBJECT DriverObject,
		    IN PDEVICE_OBJECT DeviceObject,
		    IN INTERFACE_TYPE InterfaceType,
                    IN ULONG BusNumber,
                    IN PHYSICAL_ADDRESS PortAddress,
		    IN LONG PortRange,
		    IN KIRQL IRQLine,
                    IN BOOLEAN *ConflictDetected)
{
  ULONG sizeOfResourceList;
  PCM_RESOURCE_LIST resourceList;
  PCM_FULL_RESOURCE_DESCRIPTOR nextFrd;
  PCM_PARTIAL_RESOURCE_DESCRIPTOR partial;

  //
  // The size of the resource list is going to be one full descriptor
  // which already has one partial descriptor included, plus another
  // partial descriptor. One partial descriptor will be for the
  // interrupt, and the other for the port addresses.
  //

  sizeOfResourceList = sizeof(CM_FULL_RESOURCE_DESCRIPTOR);

  //
  // The full resource descriptor already contains one
  // partial.	Make room for one more.
  //
  // It will hold the irq "prd", and the port "prd".
  //	  ("prd" = partial resource descriptor)
  //

  sizeOfResourceList += sizeof(CM_PARTIAL_RESOURCE_DESCRIPTOR);

  //
  // Now we increment the length of the resource list by field offset
  // of the first frd.   This will give us the length of what preceeds
  // the first frd in the resource list.
  //	 (frd = full resource descriptor)
  //

  sizeOfResourceList += FIELD_OFFSET(CM_RESOURCE_LIST, List[0]);

  resourceList = ExAllocatePool(PagedPool, sizeOfResourceList);

  if (!resourceList) {
    return FALSE;
  }

  //
  // Zero out the list
  //

  RtlZeroMemory(resourceList, sizeOfResourceList);

  resourceList->Count = 1;
  nextFrd = &resourceList->List[0];

  nextFrd->InterfaceType = InterfaceType;
  nextFrd->BusNumber = BusNumber;

  //
  // We are going to report port addresses and interrupt
  //

  nextFrd->PartialResourceList.Count = 2;

  //
  // Now fill in the port data.  We don't wish to share
  // this port range with anyone.
  //
  // Note: the port address we pass in is the one we got
  // back from HalTranslateBusAddress.
  //
  // CmResourceShareDriverExclusive
  // CmResourceShareDeviceExclusive
  // CmResourceShareShared

  partial = &nextFrd->PartialResourceList.PartialDescriptors[0];

  partial->Type = CmResourceTypePort;
  partial->ShareDisposition = CmResourceShareDeviceExclusive;
  partial->Flags = CM_RESOURCE_PORT_IO;
  partial->u.Port.Start = PortAddress;
  partial->u.Port.Length = PortRange;

  partial++;

  //
  // Now fill in the irq stuff.
  //
  // Note: for IoReportResourceUsage, the Interrupt.Level and
  // Interrupt.Vector are bus-specific level and vector, just
  // as we passed in to HalGetInterruptVector, not the mapped
  // system vector we got back from HalGetInterruptVector.
  //

  partial->Type = CmResourceTypeInterrupt;
  partial->u.Interrupt.Level = IRQLine;
  partial->u.Interrupt.Vector = IRQLine;
  partial->ShareDisposition = InterfaceType==Isa? 
                                 CmResourceShareDeviceExclusive:
                                 CmResourceShareShared;
  partial->Flags = InterfaceType==Isa?
                     CM_RESOURCE_INTERRUPT_LATCHED:
		     CM_RESOURCE_INTERRUPT_LEVEL_SENSITIVE;

  /* Claim resources for Driver */
  /*IoReportResourceUsage(
      NULL,
      DriverObject,
      resourceList,
      sizeOfResourceList,
      NULL,
      NULL,
      0,
      FALSE,
      ConflictDetected);*/

  /* Claim resources for individual DeviceObject */
  IoReportResourceUsage(
      NULL,
      DriverObject,
      NULL,
      0,
      DeviceObject,
      resourceList,
      sizeOfResourceList,
      FALSE,
      ConflictDetected);

  //
  // The above routine sets the BOOLEAN parameter ConflictDetected
  // to TRUE if a conflict was detected.
  //

  ExFreePool(resourceList);

  return (*ConflictDetected);
}


//-------------------------------------------------------------------
// DriverEntry for WinNT style KMD
//
// Description:
//  NT device Driver Entry point
//
// Arguments:
//      DriverObject    - Pointer to this device's driver object
//      RegistryPath    - Pointer to the Unicode regsitry path name
//
// Return Value:
//      NTSTATUS
//
NTSTATUS DriverEntry(IN PDRIVER_OBJECT DriverObject, IN PUNICODE_STRING RegistryPath)
{
    PDEVICE_OBJECT deviceObject = NULL;
    NTSTATUS status, ioConnectStatus;
    UNICODE_STRING uniNtNameString;
    UNICODE_STRING uniWin32NameString;
    UNICODE_STRING uniRegPath;
    INTERFACE_TYPE InterfaceType;
    ULONG BusNumber;
    PCI_SLOT_NUMBER SlotNumber;
    KIRQL irql;
    KIRQL IRQLine;
    KIRQL IRQLevel;
    KIRQL OldIrql;
    KAFFINITY Affinity;
    LONG TmpLong;
    LONG PortRange;
    LONG BaudRate;
    LONG MyAddr;
    LONG BufferSize;
    LONG ScanForPCI;
    ULONG MappedVector=0, AddressSpace = 1;
    PULAN_DEVICE_EXTENSION extension;
    BOOLEAN ResourceConflict;
    PHYSICAL_ADDRESS InPortAddr, OutPortAddr;
    int ChipOptions=0;

    uLan_DbgPrint("uLan v" UL_DRV_VERSION " Enter the driver!\n");
    uLan_DbgPrint("uLan: " __DATE__ " " __TIME__ "\n");

    //
    // Create counted string version of our device name.
    //

    RtlInitUnicodeString(&uniNtNameString, NT_DEVICE_NAME);

    //
    // Default configuration
    //

    ScanForPCI=0;
    InterfaceType = Isa;
    BusNumber = 0;
    InPortAddr.LowPart = DEF_PORT_ADDRESS;
    InPortAddr.HighPart = 0;
    PortRange = DEF_PORT_RANGE;
    IRQLine = DEF_IRQ_LINE;
    BaudRate = DEF_BAUD_RATE;
    BufferSize=0x8000;
    MyAddr = DEF_MY_ADDR;

    //
    // Get the configuration information from the Registry
    //

    /* RtlInitUnicodeString(&uniRegPath,RegistryPath->Buffer); */
    /* RtlCopyUnicodeString(&uniRegPath,RegistryPath); */

    status=STATUS_SUCCESS;
    RtlInitUnicodeString(&uniRegPath, NULL);
    uniRegPath.MaximumLength = RegistryPath->Length + sizeof(L"\\Parameters")+1;
    uniRegPath.Buffer = ExAllocatePool(PagedPool, uniRegPath.MaximumLength*sizeof(WCHAR));

    if (!uniRegPath.Buffer) {
        uLan_DbgPrint("uLan: ExAllocatePool failed for Path in DriverEntry\n");
        status = STATUS_UNSUCCESSFUL;
    }
    else
    {
        RtlZeroMemory(uniRegPath.Buffer,uniRegPath.MaximumLength*sizeof(WCHAR));
        RtlAppendUnicodeStringToString(&uniRegPath, RegistryPath);
        RtlAppendUnicodeToString(&uniRegPath, L"\\Parameters");
    }

    if (NT_SUCCESS(status))
        status=ulan_GetRegistryDword(uniRegPath.Buffer,L"ScanForPCI",&ScanForPCI);
    if (NT_SUCCESS(status))
	status=ulan_GetRegistryDword(uniRegPath.Buffer,L"Port Address",&InPortAddr.LowPart);
    if (NT_SUCCESS(status))
	status=ulan_GetRegistryDword(uniRegPath.Buffer,L"Port Range",&PortRange);
    TmpLong=IRQLine;
    if (NT_SUCCESS(status))
	status=ulan_GetRegistryDword(uniRegPath.Buffer,L"IRQ Line",&TmpLong);
    IRQLine=(KIRQL)TmpLong;
    IRQLevel=(KIRQL)TmpLong;
    if (NT_SUCCESS(status))
        status=ulan_GetRegistryDword(uniRegPath.Buffer,L"Baud Rate",&BaudRate);
    if (NT_SUCCESS(status))
        status=ulan_GetRegistryDword(uniRegPath.Buffer,L"My Addr",&MyAddr);
    if (NT_SUCCESS(status))
        status=ulan_GetRegistryDword(uniRegPath.Buffer,L"Buffer Size",&BufferSize);
    TmpLong=uld_debug_flg;
    if (NT_SUCCESS(status))
        status=ulan_GetRegistryDword(uniRegPath.Buffer,L"Debug",&TmpLong);
    uld_debug_flg=TmpLong;
    if(uniRegPath.Buffer)
	ExFreePool(uniRegPath.Buffer);
    if (!NT_SUCCESS(status)) {
        uLan_DbgPrint("uLan: GetConfiguration failed\n");
        return status;
    }

   #ifdef UL_WITH_PCI
    if(ScanForPCI){
        pci_device_id_t *device_id;
        uLan_DbgPrint("uLan: Calling ScanForPCICard\n");
        if(ScanForPCICard(&BusNumber,&SlotNumber,TRUE,&device_id)){
	    /*PCI_COMMON_CONFIG PCIConfig;*/
            /*HalGetBusData(PCIConfiguration,BusNumber,SlotNumber.u.AsULONG,
                             PCIConfig,sizeof(PCIConfig));*/
            uLan_DbgPrint("uLan: ScanForPCICard found slot %02X:%02X.%1X\n",
                (int)BusNumber,(int)SlotNumber.u.bits.DeviceNumber,
                (int)SlotNumber.u.bits.FunctionNumber);
            InterfaceType=PCIBus;
            ChipOptions=device_id->driver_data; /*0x16954000*/
        }
    }
   #endif /*UL_WITH_PCI*/

    //
    // Create the device object, multi-thread access (FALSE)
    //

    status = IoCreateDevice(DriverObject, sizeof(ULAN_DEVICE_EXTENSION),
                            &uniNtNameString, FILE_DEVICE_UNKNOWN, 0,
                            /* TRUE */ FALSE, &deviceObject);

    if (!NT_SUCCESS (status) ) {
        uLan_DbgPrint("uLan: IoCreateDevice failed\n");
        return status;
    }

    //
    // Set the FLAGS field
    //

    deviceObject->Flags |= DO_BUFFERED_IO;

    extension = (PULAN_DEVICE_EXTENSION) deviceObject->DeviceExtension;
    extension->flag_CHIPOK=0;

    //
    // Claim PnP busses resources
    //

   #ifdef UL_WITH_PCI
    if(InterfaceType!=Isa){
        PCM_RESOURCE_LIST  AllocatedResources;
        PCM_PARTIAL_RESOURCE_DESCRIPTOR PartialDescriptor;
        int PartialCount, i;
        InPortAddr.LowPart=IRQLine=0;
        status = HalAssignSlotResources(RegistryPath,NULL,DriverObject,
                     deviceObject,InterfaceType,BusNumber,
                     SlotNumber.u.AsULONG,&AllocatedResources);
        
        if(!AllocatedResources->Count)
            status = STATUS_INSUFFICIENT_RESOURCES;
        if(NT_SUCCESS(status)){
            PartialCount=AllocatedResources->List[0].PartialResourceList.Count;
            PartialDescriptor=AllocatedResources->List[0].PartialResourceList.PartialDescriptors;
            for(i=0;i<PartialCount;i++){
                if((PartialDescriptor[i].Type==CmResourceTypePort)&&
                  (!InPortAddr.LowPart)){
                    InPortAddr=PartialDescriptor[i].u.Port.Start;
                    PortRange=PartialDescriptor[i].u.Port.Length;
                }
                if((PartialDescriptor[i].Type==CmResourceTypeInterrupt)&&
                  (!IRQLine)){
                    IRQLevel=(KIRQL)PartialDescriptor[i].u.Interrupt.Level;
                    IRQLine=(KIRQL)PartialDescriptor[i].u.Interrupt.Vector;
                }
            }
        }
        if(!NT_SUCCESS(status)){
            uLan_DbgPrint("uLan: HalAssignSlotResources error %X\n",status);
            ExFreePool(AllocatedResources);
            IoDeleteDevice(deviceObject);
            return status;
        }
        if(!InPortAddr.LowPart || !IRQLine)
            status = STATUS_INSUFFICIENT_RESOURCES;
        uLan_DbgPrint("uLan: Found PCI resources Port=%04X IRQLevel=%02X IRQVector=%02X\n",
                       InPortAddr.LowPart,IRQLevel,IRQLine);
        ExFreePool(AllocatedResources);
    }
   #endif /*UL_WITH_PCI*/

    //
    // This call will map our IRQ to a system vector. It will also fill
    // in the IRQL (the kernel-defined level at which our ISR will run),
    // and affinity mask (which processors our ISR can run on).
    //
    // We need to do this so that when we connect to the interrupt, we
    // can supply the kernel with this information.
    //


    irql=IRQLevel;
    MappedVector = HalGetInterruptVector(
            InterfaceType, // Interface type
            BusNumber,     // Bus number
            IRQLevel,	   // BusInterruptLevel
            IRQLine,       // BusInterruptVector
            &irql,         // IRQ level
            &Affinity      // Affinity mask
            );
    

    //
    // A little known Windows NT fact,
    // If MappedVector==0, then HalGetInterruptVector failed.
    //

    
    if (MappedVector == 0) {
        uLan_DbgPrint("uLan: HalGetInterruptVector failed\n");
        IoDeleteDevice(deviceObject);
        return (STATUS_INVALID_PARAMETER);
    }

    //
    // Save off the Irql
    //

    extension->Irql = irql;


    uLan_DbgPrint("uLan: driver mappedvector=%d irql=%d, dispatch irql=%d\n",
		   MappedVector,irql,DISPATCH_LEVEL);
    if(uL_SpinLock_Irql<irql)
  		uL_SpinLock_Irql=irql;
    uLan_DbgPrint("uLan: spin lock irql=%d\n",uL_SpinLock_Irql);

    //
    // Translate the base port address to a system mapped address.
    // This will be saved in the device extension after IoCreateDevice,
    // because we use the translated port address to access the ports.
    //

    if (!HalTranslateBusAddress(InterfaceType, BusNumber, InPortAddr,
                                &AddressSpace, &OutPortAddr)) {
        uLan_DbgPrint("uLan: HalTranslateBusAddress failed\n");
        return STATUS_SOME_NOT_MAPPED;
    }
 

    if ( NT_SUCCESS(status) ) {

	//
        // Create dispatch points for create/open, close, unload, and ioctl
        //

        DriverObject->MajorFunction[IRP_MJ_CREATE] = DispatchRoutine;
        DriverObject->MajorFunction[IRP_MJ_CLOSE] = DispatchRoutine;
        DriverObject->MajorFunction[IRP_MJ_READ] = DispatchRoutine;
        DriverObject->MajorFunction[IRP_MJ_WRITE] = DispatchRoutine;
        DriverObject->MajorFunction[IRP_MJ_DEVICE_CONTROL] = DispatchRoutine;
        DriverObject->MajorFunction[IRP_MJ_CLEANUP] = DispatchRoutine;
        DriverObject->DriverUnload = UnloadDriver;

        //
        // check if resources (ports and interrupt) are available
        //

		
       #ifdef UL_WITH_PCI
        if(InterfaceType==Isa)
       #endif /*UL_WITH_PCI*/
        {
	    ReportUsage(DriverObject,deviceObject,InterfaceType,BusNumber,
                        InPortAddr,PortRange,IRQLine,&ResourceConflict);

            if (ResourceConflict) {
                uLan_DbgPrint("uLan: Couldn't get resources\n");
                IoDeleteDevice(deviceObject);
                return STATUS_INSUFFICIENT_RESOURCES;
            }
        }
		

        //
        // fill in the device extension
        //
        extension = (PULAN_DEVICE_EXTENSION) deviceObject->DeviceExtension;
        extension->DeviceObject = deviceObject;
        extension->port = OutPortAddr.LowPart;
		

        status=ul_drv_init_ext(extension,OutPortAddr.LowPart,
				IRQLine,BaudRate,ChipOptions,BufferSize,MyAddr);

        if ( !NT_SUCCESS (status) ) {
            uLan_DbgPrint("uLan: Initialization failed\n");
            IoDeleteDevice(deviceObject);
            return status;
        }


        KeInitializeDpc(&extension->bottom_dpc,ulan_bottom_dpc,extension);
        KeInitializeDpc(&extension->wd_timer_dpc,ulan_wd_dpc,extension);
	KeInitializeTimer(&extension->wd_timer);

		//
        // connect the device driver to the IRQ
        //
        ioConnectStatus = IoConnectInterrupt(&extension->InterruptObject,
                                             uld_irq_handler,	// ServiceRoutine 
                                             extension,		// ServiceContext
                                             NULL,		// SpinLock
                                             MappedVector,	// Vector
                                             irql,		// Irql 
                                             irql,		// SynchronizeIrql
                                             InterfaceType==Isa? Latched:
					       LevelSensitive,	// InterruptMode
					     InterfaceType==Isa? FALSE:
                                               TRUE,		// ShareVector 
                                             Affinity,		// ProcessorEnableMask 
                                             FALSE);		// FloatingSave 

        if ( !NT_SUCCESS (ioConnectStatus) ) {
            uLan_DbgPrint("uLan: Couldn't connect interrupt\n");
            IoDeleteDevice(deviceObject);
            return ioConnectStatus;
        }


        uLan_DbgPrint("uLan: just about ready!\n");

        //
        // Create counted string version of our Win32 device name.
        //

        RtlInitUnicodeString( &uniWin32NameString, DOS_DEVICE_NAME);
    
        //
        // Create a link from our device name to a name in the Win32 namespace.
        //
        
        /* IoDeleteSymbolicLink (&uniWin32NameString); */
        status = IoCreateSymbolicLink( &uniWin32NameString, &uniNtNameString );

        if (!NT_SUCCESS(status)) {
            uLan_DbgPrint("uLan: Couldn't create the symbolic link\n");
            IoDeleteDevice (DriverObject->DeviceObject);
        } else {

            //
            // Setup the Dpc for ISR routine
            //
			/*

            IoInitializeDpcRequest (DriverObject->DeviceObject, uLan_Dpc_Routine);

            //
            // Initialize the device (enable IRQ's, hit the hardware)
            //

            Initialize_uLan (extension);
		    */

            uLan_DbgPrint("uLan: All initialized!\n");
            
            #ifdef UL_INT_TST
			geninfoblk(extension);
            printudrvbll(extension);
			extension->flag_NACTIV=1;
			KeRaiseIrql(DISPATCH_LEVEL,&OldIrql);
            uld_timeout(extension);
			KeLowerIrql(OldIrql);
            #endif  /* UL_INT_TST */

        }

    } else {
        uLan_DbgPrint("uLan: Couldn't create the device\n");
    }
    return status;
	
}



//---------------------------------------------------------------------------
// UnloadDriver
//
// Description:
//     Free all the allocated resources, etc.
//
// Arguments:
//     DriverObject - pointer to a driver object
// 
// Return Value:
//      None
// 
VOID UnloadDriver (IN PDRIVER_OBJECT DriverObject)
{
    WCHAR                  deviceLinkBuffer[]  = DOS_DEVICE_NAME;
    UNICODE_STRING         deviceLinkUnicodeString;
    PULAN_DEVICE_EXTENSION extension;

    extension = DriverObject->DeviceObject->DeviceExtension;

    #ifdef UL_INT_TST
    if(extension->flag_CHIPOK);
      printudrvbll(extension);
    #endif  /* UL_INT_TST */

    //
    // Deactivate uLan 
    //

    ul_drv_done_ext(extension);

    //
    // Disconnect IRQ and DPC sources
    //
	
    IoDisconnectInterrupt (extension->InterruptObject);
      
    KeCancelTimer(&extension->wd_timer);
    KeRemoveQueueDpc(&extension->wd_timer_dpc);
    KeRemoveQueueDpc(&extension->bottom_dpc);
		
    //
    // Delete the symbolic link
    //
	

    RtlInitUnicodeString (&deviceLinkUnicodeString, deviceLinkBuffer);

    IoDeleteSymbolicLink (&deviceLinkUnicodeString);

    //
    // Delete the device object
    //

    IoDeleteDevice (DriverObject->DeviceObject);

   #ifdef ENABLE_UL_MEM_CHECK
    UL_PRINTF("MEM_CHECK: malloc-free=%d\n",
		 (int)atomic_read(&ul_mem_check_counter));
   #endif /* ENABLE_UL_MEM_CHECK */
    uLan_DbgPrint ("uLan: Unloaded\n");
    return;
}

#endif /* !FOR_WIN_WDM */

//---------------------------------------------------------------------------
//
//
// Routine Description:
// 
//     Process the IRPs sent to this device.
// 
// Arguments:
// 
//     DeviceObject - pointer to a device object
// 
//     Irp          - pointer to an I/O Request Packet
// 
// Return Value:
// 
// 
NTSTATUS DispatchRoutine (IN PDEVICE_OBJECT DeviceObject, IN PIRP Irp)
{
    PIO_STACK_LOCATION  irpStack;
    PULAN_DEVICE_EXTENSION   extension;
    PVOID               ioBuffer;
    ULONG               inputBufferLength;
    ULONG               outputBufferLength;
    ULONG               ioControlCode;
    NTSTATUS            ntStatus=0;
    KIRQL		OldIrql;

    LARGE_INTEGER       CurrentSystemTime;
    LARGE_INTEGER       ElapsedTime;
    ul_opdata			*opdata;
	ul_msginfo msginfo;

    
    Irp->IoStatus.Status      = STATUS_SUCCESS;
    Irp->IoStatus.Information = 0;

    //
    // Get a pointer to the current location in the Irp. This is where
    //     the function codes and parameters are located.
    //

    irpStack = IoGetCurrentIrpStackLocation (Irp);

    //
    // Get a pointer to the device extension
    //

    extension = DeviceObject->DeviceExtension;

  #ifdef FOR_WIN_WDM
    if(!extension||(extension->State<=STATE_ALL_BELOW_FAIL)||!extension->flag_CHIPOK){
       uLan_DbgPrint ("uLan: DeviceObject initialization not finished\n");
	   Irp->IoStatus.Status=STATUS_UNSUCCESSFUL;
       IoCompleteRequest (Irp, IO_NO_INCREMENT);
	   return STATUS_UNSUCCESSFUL;
				
	}
  #endif /* FOR_WIN_WDM */

    //
    // Get the pointer to the input/output buffer and it's length
    //

    switch (irpStack->MajorFunction) {
        case IRP_MJ_CREATE:
        {    
            LOG_FILEIO("uLan: IRP_MJ_CREATE\n");
	    if(irpStack->FileObject==NULL)
	    {	uLan_DbgPrint ("uLan: Open called without FileObject\n");
		Irp->IoStatus.Status=STATUS_UNSUCCESSFUL;
		break;
	    }
	    if(irpStack->FileObject->FsContext)
	    {	uLan_DbgPrint ("uLan: FsContext already used\n");
		Irp->IoStatus.Status=STATUS_UNSUCCESSFUL;
		break;
	    }
            if(!(opdata=MALLOC(sizeof(ul_opdata))))
	    {	uLan_DbgPrint ("uLan: Cannot allocate op_data\n");
		Irp->IoStatus.Status=STATUS_NO_MEMORY;
		break;
	    }
	    irpStack->FileObject->FsContext=opdata;
            opdata->file=irpStack->FileObject;
	    if((ntStatus=ul_drv_open(extension,opdata)))
	    {	LOG_FILEIO ("uLan: ul_drv_open failed\n");
		Irp->IoStatus.Status=ntStatus;
		FREE(opdata);
	        irpStack->FileObject->FsContext=NULL;
		break;
	    }
            break;
        }

        case IRP_MJ_CLOSE:
        {
	    LOG_FILEIO("uLan: IRP_MJ_CLOSE\n");
	    opdata=irpStack->FileObject->FsContext;
	    if(!opdata||(opdata->magic!=ULOP_MAGIC))
	    {	uLan_DbgPrint ("uLan: corrupted opdata !!!!\n");
		Irp->IoStatus.Status=STATUS_UNSUCCESSFUL;
		break;
	    }
	    wake_and_xch_pending_irp(&opdata->wait_irp,NULL,STATUS_UNSUCCESSFUL);
	    if((ntStatus=ul_drv_close(extension,opdata)))
	    {	LOG_FILEIO("uLan: ul_drv_close failed\n");
		Irp->IoStatus.Status=ntStatus;
	    }
	    irpStack->FileObject->FsContext=NULL;
	    FREE(opdata);
            break;
        }

        case IRP_MJ_READ:
        {
	    LOG_FILEIO("uLan: IRP_MJ_READ\n");
	    opdata=irpStack->FileObject->FsContext;
	    if(!opdata||(opdata->magic!=ULOP_MAGIC))
	    {	uLan_DbgPrint ("uLan: corrupted opdata !!!!\n");
		Irp->IoStatus.Status=STATUS_UNSUCCESSFUL;
		break;
	    }
	    ntStatus=ulan_read(extension,opdata,
				Irp->AssociatedIrp.SystemBuffer,
				irpStack->Parameters.Read.Length,
				irpStack->Parameters.Read.ByteOffset,
				&Irp->IoStatus.Information);
	    if(ntStatus)
	    {	LOG_FILEIO("uLan: ul_drv_read failed\n");
		Irp->IoStatus.Status=ntStatus;
	    }
            break;
        }

        case IRP_MJ_WRITE:
        {
	    LOG_FILEIO("uLan: IRP_MJ_WRITE\n");
	    opdata=irpStack->FileObject->FsContext;
	    if(!opdata||(opdata->magic!=ULOP_MAGIC))
	    {	uLan_DbgPrint ("uLan: corrupted opdata !!!!\n");
		Irp->IoStatus.Status=STATUS_UNSUCCESSFUL;
		break;
	    }
	    ntStatus=ulan_write(extension,opdata,
				Irp->AssociatedIrp.SystemBuffer,
				irpStack->Parameters.Read.Length,
				irpStack->Parameters.Read.ByteOffset,
				&Irp->IoStatus.Information);
	    if(ntStatus)
	    {	LOG_FILEIO("uLan: ul_drv_write failed\n");
		Irp->IoStatus.Status=ntStatus;
	    }
            break;
        }

        case IRP_MJ_CLEANUP:
        {
	    LOG_FILEIO("uLan: IRP_MJ_CLEANUP\n");
	    opdata=irpStack->FileObject->FsContext;
	    if(!opdata||(opdata->magic!=ULOP_MAGIC))
	    {	uLan_DbgPrint ("uLan: corrupted opdata !!!!\n");
		Irp->IoStatus.Status=STATUS_UNSUCCESSFUL;
		break;
	    }
	    wake_and_xch_pending_irp(&opdata->wait_irp,NULL,STATUS_UNSUCCESSFUL);
            break;
        }

        case IRP_MJ_DEVICE_CONTROL:
        {
	    LOG_FILEIO("uLan: IRP_MJ_DEVICE_CONTROL - ");
	    opdata=irpStack->FileObject->FsContext;
	    if(!opdata||(opdata->magic!=ULOP_MAGIC))
	    {	uLan_DbgPrint ("uLan: corrupted opdata !!!!\n");
		Irp->IoStatus.Status=STATUS_UNSUCCESSFUL;
		break;
	    }
			    
            ioControlCode = irpStack->Parameters.DeviceIoControl.IoControlCode;
	    ioBuffer      = Irp->AssociatedIrp.SystemBuffer;
	    inputBufferLength  = irpStack->Parameters.DeviceIoControl.InputBufferLength;
	    outputBufferLength = irpStack->Parameters.DeviceIoControl.OutputBufferLength;

    
            switch (ioControlCode)
            {
		case UL_DRV_VER :
		    LOG_FILEIO("UL_DRV_VER\n");
                    if (outputBufferLength < sizeof(ULONG))
		    {
			Irp->IoStatus.Status = STATUS_UNSUCCESSFUL;
			break;
		    }
                    *(ULONG *)ioBuffer = UL_DRV_VERCODE;
                    Irp->IoStatus.Information = sizeof(ULONG);
		    break;
	    
		case UL_NEWMSG :
		    LOG_FILEIO("UL_NEWMSG\n");
                    if (inputBufferLength < sizeof(ul_msginfo)) 
		    {	LOG_FILEIO("uLan: newmsg shorten msginfo !!!!\n");
			Irp->IoStatus.Status = STATUS_UNSUCCESSFUL;
			break;
		    }
		    if((ntStatus=ulan_newmsg(opdata,ioBuffer)))
		    {	LOG_FILEIO("uLan: newmsg failed !!!!\n");
			Irp->IoStatus.Status = ntStatus;
		    }
		    break;
		case UL_TAILMSG :
		    LOG_FILEIO("UL_TAILMSG\n");
                    if (inputBufferLength < sizeof(ul_msginfo)) 
		    {	LOG_FILEIO("uLan: tailmsg shorten msginfo !!!!\n");
			Irp->IoStatus.Status = STATUS_UNSUCCESSFUL;
			break;
		    }
		    if((ntStatus=ulan_tailmsg(opdata,ioBuffer)))
		    {	LOG_FILEIO("uLan: tailmsg failed !!!!\n");
			Irp->IoStatus.Status = ntStatus;
		    }
		    break;
		case UL_FREEMSG :
		    LOG_FILEIO("UL_FREEMSG\n");
		    if((ntStatus=ulan_freemsg(opdata))&~0x7fffffff)
		    {	LOG_FILEIO("uLan: freemsg failed !!!!\n");
			Irp->IoStatus.Status = ntStatus;
		    }
                    if (outputBufferLength >= sizeof(ULONG)) {
                        *(ULONG *)ioBuffer = ntStatus;
                        Irp->IoStatus.Information = sizeof(ULONG);
                    }
					break;
		case UL_ACCEPTMSG :  	
		    LOG_FILEIO("UL_ACCEPTMSG\n");
                    if (outputBufferLength < sizeof(ul_msginfo)) 
		    {	LOG_FILEIO("uLan: acceptmsg shorten msginfo !!!!\n");
			Irp->IoStatus.Status = STATUS_UNSUCCESSFUL;
			break;
		    }
		    if((ntStatus=ulan_acceptmsg(opdata,ioBuffer)))
		    {	LOG_FILEIO("uLan: acceptmsg failed !!!!\n");
			Irp->IoStatus.Status = ntStatus;
		    }
		    else Irp->IoStatus.Information = sizeof(ul_msginfo);
		    break;
		case UL_ACTAILMSG :  	
		    LOG_FILEIO("UL_ACTAILMSG\n");
                    if (outputBufferLength < sizeof(ul_msginfo)) 
		    {	LOG_FILEIO("uLan: actailmsg shorten msginfo !!!!\n");
			Irp->IoStatus.Status = STATUS_UNSUCCESSFUL;
			break;
		    }
		    if((ntStatus=ulan_actailmsg(opdata,ioBuffer)))
		    {	LOG_FILEIO("uLan: actailmsg failed !!!!\n");
			Irp->IoStatus.Status = ntStatus;
		    }
		    else Irp->IoStatus.Information = sizeof(ul_msginfo);
		    break;
		case UL_ADDFILT :
		    LOG_FILEIO("UL_ADDFILT\n");
                    if (inputBufferLength < sizeof(ul_msginfo)) 
		    {	LOG_FILEIO("uLan: addfilt shorten msginfo !!!!\n");
			Irp->IoStatus.Status = STATUS_UNSUCCESSFUL;
			break;
		    }
		    if((ntStatus=ulan_addfilt(opdata,ioBuffer)))
		    {	LOG_FILEIO("uLan: addfilt failed !!!!\n");
			Irp->IoStatus.Status = ntStatus;
		    }
		    break;
		case UL_ABORTMSG :
		    LOG_FILEIO("UL_ABORTMSG\n");
		    if((ntStatus=ulan_abortmsg(opdata)))
		    {	LOG_FILEIO("uLan: abortmsg failed !!!!\n");
			Irp->IoStatus.Status = ntStatus;
		    }
		    break;
		case UL_REWMSG :
		    LOG_FILEIO("UL_REWMSG\n");
		    if((ntStatus=ulan_rewmsg(opdata)))
		    {	LOG_FILEIO("uLan: rewmsg failed !!!!\n");
			Irp->IoStatus.Status = ntStatus;
		    }
		    break;
		case UL_INEPOLL :
		    LOG_FILEIO("UL_INEPOLL\n");
                    if (outputBufferLength < sizeof(ULONG))
		    {
			Irp->IoStatus.Status = STATUS_UNSUCCESSFUL;
			break;
		    }
                    *(ULONG *)ioBuffer = ulan_inepoll(opdata);
                    Irp->IoStatus.Information = sizeof(ULONG);
		    break;
		case UL_WAITREC :
		    LOG_FILEIO("UL_WAITREC\n");
                    if(ulan_inepoll(opdata)) break;
		    {
			KIRQL OldIrql;
			IoAcquireCancelSpinLock(&OldIrql);
			IoSetCancelRoutine(Irp,&ulan_cancel);
			IoReleaseCancelSpinLock(OldIrql);
		    }
                    IoMarkIrpPending(Irp);
	            wake_and_xch_pending_irp(&opdata->wait_irp,Irp,STATUS_UNSUCCESSFUL);
		    return STATUS_PENDING ;
		case UL_KLOGBLL :
		    LOG_FILEIO("UL_KLOGBLL\n");
		    printudrvbll(extension);
		    #ifdef ENABLE_UL_MEM_CHECK
		     printudrvoperators(extension);
		    #endif /* ENABLE_UL_MEM_CHECK */
		    break;
		case UL_STROKE :
		    LOG_FILEIO("UL_STROKE\n");
		    ulan_stroke(extension);
		    break;
		case UL_DEBFLG :
		    LOG_FILEIO("UL_DEBFLG\n");
                    if (inputBufferLength >= sizeof(ULONG)) 
		    {	
			uld_debug_flg=*(ULONG *)ioBuffer;
			break;
		    }
		    break;
		case UL_HWTEST :
		    LOG_FILEIO("UL_HWTEST\n");
                    if ((outputBufferLength < sizeof(ULONG))||
                        (inputBufferLength < sizeof(ULONG)))
		    {	
			Irp->IoStatus.Status = STATUS_UNSUCCESSFUL;
			break;
		    }
                    *(LONG *)ioBuffer = ulan_hwtest(opdata,*(ULONG *)ioBuffer);
		    if(*(LONG *)ioBuffer<0)
		        Irp->IoStatus.Status = STATUS_UNSUCCESSFUL;
                    Irp->IoStatus.Information = sizeof(ULONG);
		    break;
                default:
                {
                    LOG_FILEIO("uLan: Unknown IRP_MJ_DEVICE_CONTROL\n");
                    Irp->IoStatus.Status = STATUS_INVALID_PARAMETER;
                    break;
                }

            }
    
            break;
        }
        default:
        {
            uLan_DbgPrint ("uLan: Unhandled IRP_MJ function\n");
            Irp->IoStatus.Status = STATUS_NOT_IMPLEMENTED;
            break;
        }
    }

    //
    // DON'T get cute and try to use the status field of
    // the irp in the return status.  That IRP IS GONE as
    // soon as you call IoCompleteRequest.
    //

    ntStatus = Irp->IoStatus.Status;

    IoCompleteRequest (Irp, IO_NO_INCREMENT);

    LOG_FILEIO("uLan: DisptachRoutine exit.\n");

    //
    // We never have pending operation so always return the status code.
    //

    return ntStatus;
}

#ifdef FOR_WIN_WDM

#include "ul_wdent.c"

#include "ul_wdpnp.c"

#endif /* FOR_WIN_WDM */


#endif /* _WIN32 */
